package nxm.ice.core;

import nxm.ice.lib.*;
import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;
import nxm.sys.libm.Fft;

/**
  Implements a Bank of Tuners

  @author Jeff Schoen
  @version $Id: BoT.java,v 1.8 2011/05/24 13:21:13 jgs Exp $
*/
public class BoT extends CoreFactory {

 // native implementation wrappers
 public static class VHS extends CoreNative { public native long alloc(); }
 public static class ICE extends CoreNative { public native long alloc(); }

 // java implementation
 public static class JVM extends Core { 

  static String flagsList = "Real,Pkt";
  static int REAL=0x1,PKT=0x2;
  boolean real;
  double foff,fdel,ratio,phase,phaso,phasd,phasr;
  int nfft,nout,nppt,nrft,nrpt,nper,gain,flags,pass,rfpass,pfpass;
  float[] rpbuf,ppbufe,ppbufo,fbe,fbo,fbx,fppf,pfir,qfir,rfir;
  Fft fft;

  /** set parameters */
  public void set (String key, Data value) {
         if (key.equals("NFFT"))  nfft  = value.toL();
    else if (key.equals("NOUT"))  nout  = value.toL();
    else if (key.equals("NPPT"))  nppt  = value.toL();
    else if (key.equals("NRPT"))  nrpt  = value.toL();
    else if (key.equals("GAIN"))  gain  = value.toL();
    else if (key.equals("FOFF"))  foff  = value.toD();
    else if (key.equals("FLAGS")) flags = value.toL();
    else if (key.equals("RATIO")) ratio = value.toD();
    else if (key.equals("FPPF"))  fppf  = value.castF(true);
    else if (key.equals("RFIR"))  rfir  = value.castF(true);
    else super.set(key,value);
  }

  /** get parameters */
  public Data get (String key, byte type) {
         if (key.equals("NFFT"))  return new Data(nfft);
    else if (key.equals("NOUT"))  return new Data(nout);
    else if (key.equals("NPPT"))  return new Data(nppt);
    else if (key.equals("NRPT"))  return new Data(nrpt);
    else if (key.equals("GAIN"))  return new Data(gain);
    else if (key.equals("FOFF"))  return new Data(foff);
    else if (key.equals("RATIO")) return new Data(ratio);
    else if (key.equals("FLAGS")) return new Data(flags);
    else if (key.equals("FLAGSLIST")) return new Data(flagsList);
    else return super.get(key,type);
  }

  /** initialize resources */
  public int open () {
    real = (flags&REAL)!=0;
    nrft = rfir.length;
    nper = nrft/nrpt;
    qfir = new float[nrpt];
    rpbuf = new float[nout*2*nrpt];
    ppbufe = new float[nfft*2*nppt];
    ppbufo = new float[nfft*2*nppt];
    fbe = new float[nfft*2];
    fbo = new float[nfft*2];
    fbx = new float[nfft*4];
    fft = new Fft(nfft,Fft.FORWARD|Fft.COMPLEX);
    fdel = 0.5/nfft;
    phase = phaso = 0;
    pass = rfpass = pfpass = 0;
    phasd = (real?1:2)/ratio;
    phasr = nrpt;
    return 0;
  }

  /** free resources */
  public int close () {
    fft.free();
    return 0;
  }

  /** process a buffer of data.  */
  public int process (byte[] a, int na, byte[] b, int nb) {
    int n=0;
    if (na<nfft*8) return 0;
    float[] af = Convert.castF(a,true);
    float[] bf = Convert.castF(b,false);
    phase = processTune(nfft,af,fbe,phase,foff);
    phaso = processTune(nfft,af,fbo,phaso,foff+fdel);
    n=processFPPF(nppt,nfft,fbe,fbe,ppbufe);
    n=processFPPF(nppt,nfft,fbo,fbo,ppbufo);
    if (n>0) {
      fft.work(fbe);
      fft.work(fbo);
      n=processRFIR(nout,fbe,fbo,bf);
    }
    Convert.uncast(af,a,false);
    Convert.uncast(bf,b,true);
    pass++;
    return n*4;
  }

  private double processTune (int nfft, float[] a, float[] b, double phase, double freq) {
    for (int i=0; i<nfft*2; ) {
      float cr=(float)Math.cos(phase);
      float ci=(float)Math.sin(phase);
      phase -= (TWOPI*freq);
      float ar=a[i], ai=a[i+1];
      b[i++] = ar*cr - ai*ci;
      b[i++] = ar*ci + ai*cr;
    }
    return phase;
  }

  private int processFPPF (int ntap, int nfft, float[] a, float[] b, float[] ppbuf) {
    int nfft2 = nfft*2;
    int ntap2 = ntap*nfft2;
    int n0 = (pass%ntap)*nfft2;
    int n1 = ((pass+ntap-1)%ntap)*nfft2;
    float sumr,sumi,factor=1;
    System.arraycopy(a,0, ppbuf,n1, nfft2);
    if (pass<ntap-1) return 0;
    for (int i=0,ii=0; i<nfft; i++) {
      sumr = sumi = 0;
      for (int n=0,i0=i,nx=n0+ii; n<ntap; n++,i0+=nfft,nx=(nx+nfft2)%ntap2) {
        sumr += (fppf[i0]*ppbuf[nx+0]);
        sumi += (fppf[i0]*ppbuf[nx+1]);
      }
      b[ii++] = factor*sumr;
      b[ii++] = factor*sumi;
    }
    return 1;
  }

  private int processRFIR (int nout, float[] ae, float[] ao, float[] b) {
    float sum,scale=(float)Math.pow(2,gain)/nfft;
    int nout2=nout*2, nrpt2=nrpt*nout2;
    int i,j,k,l,ii=(rfpass%nrpt)*nout2;
    // copy next row into sample array
    for (i=j=0,k=ii; i<nout; ) {
      rpbuf[k++] = ae[i++];
      rpbuf[k++] = ae[i++];
      rpbuf[k++] = ao[j++];
      rpbuf[k++] = ao[j++];
    }
    rfpass++;
    // compute next set of taps
    ii = (int)phasr;
    if (ii>rfpass) return 0;
    ii=(ii%nrpt)*nout2;
    // if real, adjust sign and IQ for each pass 0i,-1q,-2i,3q
    if (real) { if (((pfpass+1)&2)!=0) scale*=-1; if ((pfpass&1)!=0) ii++; }
    double phasrm = (phasr % 1.0) * nper;
    i=(int)(phasrm); phasrm-=i;
    double p0=phasrm, p1=(1.0-p0);
    for (j=0; j<nrpt; j++,i+=nper) {
      int ip1 = Math.min(i+1,nrft-1);
      qfir[nrpt-1-j] = (float)(rfir[i]*p1 + rfir[ip1]*p0);
    }
    // compute array of taps
    for (i=l=0; i<nout2; i++,ii++) {
      for (sum=0,j=0,k=ii; j<nrpt-1; j++) {
        k+=nout2; if (k>=nrpt2) k-=nrpt2;
        sum += qfir[j]*rpbuf[k];
      }
      b[l++] = sum*scale;
      if (real) { i++; ii++; }
    }
    pfpass++;
    phasr += phasd;
    return l;
  }

 }
}
