/**
  Implements various demodulators on a complex baseband input.

  @author Jeff Schoen
  @version $Id: Demod.c,v 1.6 2011/06/07 20:47:32 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME Demod
#define CORE_AREA nxm_ice_core

/* CORE defines */
#include "CoreDefs.h"

/* Local defines */
#define MODELIST "LUT"
#define LUT 1

#define DB 16           // bits in data type
#define QB 5            // addr bits per quadrant
#define TB (QB+1)       // addr bits total
#define LB (DB-TB)      // lower address bits
#define QW (1<<QB)      // divisions per quadrant
#define TW (1<<TB)      // divisions per table
#define LW (1<<LB)      // divisions per residual
#define tmask (TW-1)
#define lmask (LW-1)
#define umask (0xFFFF & ~lmask)
#define TS (QW*QW)

/* CORE Plan handle */
typedef struct {
  HALO halo;
  int_4 state;
  int_4 mode;
  int_4 flags;
  int_4 pass;
  double fullScale;
  int_4 valueLast;
  int_2 tab0[TS],tab1[TS],tab2[TS],tab3[TS];
} DemodPlan;

/* CORE prototypes */
#include "CoreProtos.h"

/* Local prototypes */
#if CORE_IS_CPU
static void  Demod_loadFPGA (DemodPlan *plan, int_2 *data, int_4 n);
static int_4 Demod_procFPGA (DemodPlan *plan, int_2 *bufin, int_4 nin, int_2 *bufout, int_4 nout);
#endif

/* CORE code */

void* Demod_alloc (char *config) {
  DemodPlan *plan = Demod_plan();
  HW_open(config);
  plan->state = 0;
  plan->mode = LUT;
  plan->valueLast = 0;
  plan->pass = 0;
  return (void*)plan;
}

int_4 Demod_setkey (DemodPlan *plan, char *key, void *data, int_4 len) {
  int_4 i, *ldata = (int_4*)data;
  if      (keyMatch("L:MODULATION")) { plan->mode = ldata[0]; }
  else if (keyMatch("D:FULLSCALE"))  { plan->fullScale = *(double*)data; }
  else if (keyMatch("L:LUT.FLAGS"))  { plan->flags = ldata[0]; }
  else if (keyMatch("I:LUT.DATA"))   { 
#if CORE_IS_FPGA
    HW_write(0x0,0x0000);
    HW_write(0x0,0x0809);
    for (i=0; i<len/4; i++) HW_write(0x20,ldata[i]);
    HW_write(0x20,plan->flags);
    HW_write(0x0,0x0000);
#elif CORE_IS_CPU
    Demod_loadFPGA(plan,(short*)data,len/4);
#endif
  }
  else return 0;
  return len;
}

int_4 Demod_getkey (DemodPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:MODULATION"))  { *(int_4*)data = plan->mode; return 4; }
  if (keyMatch("S:MODULATIONS")) { strcpy (data,MODELIST); return strlen(data); }
  if (keyMatch("D:FULLSCALE"))   { *(double*)data = plan->fullScale; return 8; }
  return 0;
}

int_4 Demod_init (DemodPlan *plan) {
#if CORE_IS_FPGA
  HW_write(0x0,0x0809);  // enable
#endif
  return 0;
}

int_4 Demod_work (DemodPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#if CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#elif CORE_IS_CPU
  n = Demod_procFPGA (plan,(int_2*)buf1,n1/4,(int_2*)buf2,n2/2) * 2;
#endif
  plan->pass++;
  return n;
}

int_4 Demod_free (DemodPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

/* Local Code */

#if CORE_IS_CPU
int_4 Demod_procFPGA (DemodPlan *plan, int_2 *ibi, int_4 nin, int_2 *ibo, int_4 nout) {
  int i,j,k,ar,ai,br,bi,dr,di,v,vv,qa;
  int diff=(plan->flags&0x1), quad=(plan->flags&0x2), vl=plan->valueLast;
  int_2 *iut0 = plan->tab0;
  int_2 *iut1 = plan->tab1;
  int_2 *iut2 = plan->tab2;
  for (i=j=0; i<nin;) {
    ar = ibi[j++]; ai = ibi[j++];
    if (ar<0) {
      if (ai<0) { br=-ar; bi=-ai; qa=0x8000; }
      else      { br=ai;  bi=-ar; qa=0x4000; }
    } else {
      if (ai<0) { br=-ai; bi=ar; qa=0xC000; }
      else      { br=ar;  bi=ai; qa=0x0000; }
    }
    k = (bi&umask)>>(LB-QB) | (br&umask)>>LB;         // lookup table
    dr = br&lmask; di = bi&lmask;                     // residuals
    v = iut0[k] + ((dr*iut1[k])>>(16-TB)) + ((di*iut2[k])>>(16-TB));
    if (quad) v = qa + (v>>1); else v <<= 1;
    vv = v; if (diff) vv -= vl;
    ibo[i++] = (short)vv;
    vl = v;
  }
  if (diff && plan->pass==0) ibo[0]=ibo[1];
  plan->valueLast = vl;
  return nin;
}

void Demod_loadFPGA (DemodPlan *plan, int_2 *data, int_4 n) {
  int i=0, fs=QW, x,y;
  for (y=0; y<fs; y++) {
    for (x=0; x<fs; x++) {
      plan->tab0[i] = data[0*fs*fs+i];
      plan->tab1[i] = data[1*fs*fs+i];
      plan->tab2[i] = data[2*fs*fs+i];
      plan->tab3[i] = data[3*fs*fs+i];
      i++;
    }
  }
}
#endif

