package nxm.ice.core;

import nxm.ice.lib.*;
import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;

/**
  Implements various demodulators on a complex baseband input.

  @author Jeff Schoen
  @version $Id: Demod.java,v 1.8 2011/05/24 13:21:13 jgs Exp $
*/
public class Demod extends CoreFactory {

 // native implementation wrappers
 public static class CPU extends CoreNative { public native long alloc(); }
 public static class VHS extends CoreNative { public native long alloc(); }
 public static class ICE extends CoreNative { public native long alloc(); }

 // java implementation
 public static class JVM extends Core {

  private int mode,pass,flags;
  private double fullScale;
  private boolean lut;

  public static final String keyList = "L:MODULATION,D:FULLSCALE";

  public static final String modeList = "AM,FM,PM,LSB,USB,CW,PSK,QPSK,LUT";
  public static final int AM=1,FM=2,PM=3,LSB=4,USB=5,CW=6,PSK=7,QPSK=8,LUT=9;

  /** set parameters */
  public void set (String key, Data value) {
         if (key.equals("MODULATION")) { setModulation(value.toL()); }
    else if (key.equals("FULLSCALE")) { fullScale = value.toD(); }
    else if (key.equals("LUT.DATA"))  { loadFPGA(value); }
    else if (key.equals("LUT.FLAGS")) { flags = value.toL(); }
    else super.set(key,value);
  }

  /** get parameters */
  public Data get (String key, byte type) {
         if (key.equals("KEYS"))        return new Data(keyList);
    else if (key.equals("MODULATION"))  return new Data(mode);
    else if (key.equals("MODULATIONS")) return new Data(modeList);
    else if (key.equals("FULLSCALE"))   return new Data(fullScale);
    else return super.get(key,type);
  }

  /** Process a buffer of data.
    If LUT, buffers are CI,SI
    else    buffers are CF,SF.
  */
  public int process (byte[] a, int na, byte[] b, int nb) {
    int n=0;
    if (lut) {
      short[] ai = Convert.castI(a,true);
      short[] bi = Convert.castI(b,false);
      n = processFPGA (ai, bi, na/4) * 2;
      Convert.uncast(ai,a,false);
      Convert.uncast(bi,b,true);
    }
    else {
      float[] af = Convert.castF(a,true);
      float[] bf = Convert.castF(b,false);
      n = processSTD (af, bf, na/8) * 4;
      Convert.uncast(af,a,false);
      Convert.uncast(bf,b,true);
    }
    pass++;
    return n;
  }

  private void setModulation (int mode) {
    this.mode = mode;
    flags = 0;
    if (mode==FM) flags |= 0x1; // diff
    if ((mode==FM) || (mode==PM)) flags |= 0x2; // quad
    lut = (mode==LUT);
  }

  /*
     Standard Java Code Engine
  */

  public int processSTD (float[] a, float[] b, int n) {
    switch (mode) {
      case AM:  processAM  (a,b,n); break;
      case FM:  processFM  (a,b,n); break;
      case PM:  processPM  (a,b,n); break;
      case PSK: processPSK (a,b,n); break;
      default: n = -1;
    }
    return n;
  }

  private float valueLast;
  private boolean firstPass=true;

  private void processAM (float[] fbi, float[] fbo, int n) {
    float fr,fi;
    for (int i=0,j=0; i<n; ) {
      fr=fbi[j++]; fi=fbi[j++];
      fbo[i++] = (float)Math.sqrt((double)(fr*fr+fi*fi));
    }
  }

  private void processFM (float[] fbi, float[] fbo, int n) {
    float fr,fi,fp,value,scale=(float)(fullScale/TWOPI);
    float bound2=(float)TWOPI, bound=bound2/2;;
    for (int i=0,j=0; i<n; ) {
      fr=fbi[j++]; fi=fbi[j++];
      value = getPhase(fr,fi);
      fp = value-valueLast; 
      // adjust for arctan phase wrap issues by checking for abs(fr) > pi
      if      (fp>=bound) fp -= bound2;
      else if (fp<-bound) fp += bound2;
      fbo[i++] = fp*scale;
      valueLast = value;
    }
    if (pass==0) fbo[0]=fbo[1]; 
  }

  private void processPM (float[] fbi, float[] fbo, int n) {
    float fr,fi,scale = (float)(fullScale/TWOPI);
    for (int i=0,j=0; i<n; ) {
      fr=fbi[j++]; fi=fbi[j++];
      fbo[i++] = scale*getPhase(fr,fi);
    }
  }

  private void processPSK (float[] fbi, float[] fbo, int n) {
  }

  private float getPhase (float dr, float di) {
    return (float)Math.atan2((double)di,(double)dr);
  }

  /*
     LookUpTable Java Code Engine
  */

  private static final int DB=16;				// bits in data type
  private static final int QB=5;				// addr bits per quadrant
  private static final int TB=QB+1;				// addr bits total
  private static final int LB=DB-TB;				// lower address bits
  private static final int QW=1<<QB;				// divisions per quadrant
  private static final int TW=1<<TB;				// divisions per table
  private static final int LW=1<<LB;				// divisions per residual
  private static final int tmask = TW-1;
  private static final int lmask = LW-1;
  private static final int umask = 0xFFFF & ~lmask;

  private int lut0[],lut1[],lut2[],lut3[];
  private int valueLastLUT=0;

  public int getLUTframe() { return QW; }

  public int processLUT (short[] ibi, short[] ibo, int n) {
    int i,j,k,ar,ai,dr,di,v,vv;
    boolean diff = (flags&0x1) != 0;
    boolean quad = (flags&0x2) != 0;
    for (i=j=0; i<n; ) {
      ar = ibi[j++]; ai = ibi[j++];
      k = (ai&umask)>>(LB-TB) | (ar&umask)>>LB;		// lookup table
      dr = ar&lmask; di = ai&lmask;			// residuals
      v = lut0[k] + dr*lut1[k] + di*lut2[k];
      v >>= 15;
      ibo[i++] = (short)(diff? v-valueLastLUT : v);
      valueLastLUT = v;
    }
    if (diff && pass==0) ibo[0]=ibo[1];
    return n;
  }

  private void initLUT() {
    int i,j,k=0;
    double h=.5;
    lut0 = new int[TW*TW];
    lut1 = new int[TW*TW];
    lut2 = new int[TW*TW];
    lut3 = new int[TW*TW];
    for (j=-QW; j<QW; j++) {
      for (i=-QW; i<QW; i++) {
        k = (i&tmask) | ((j&tmask)<<TB);
	lut0[k] = lutval(j+h,i+h);
	lut1[k] = lutval(j+h,i+1)-lutval(j+h,i+0);
	lut2[k] = lutval(j+1,i+h)-lutval(j+0,i+h);
        lut1[k] <<= 1; lut1[k] >>= 1;	// handle wrap
        lut2[k] <<= 1; lut2[k] >>= 1;	// handle wrap
        lut3[k] = lut0[k];
        lut0[k] -= (lut1[k]/2 + lut2[k]/2);
	lut1[k] >>= (16-TB);
	lut2[k] >>= (16-TB);
      }
    }
  }

  private short iut0[],iut1[],iut2[],iut3[];
  private int valueLastFPGA=0;


// FPGA uses only 1 quadrant of short LUT values
// This code emulates that process (for development) but is slower on CPUs
// Quadrant shift: 1=(x,y) 2=(y,-x) 3=(-x,-y) 4=(-y,x)

  public int processFPGA (short[] ibi, short[] ibo, int n) {
    int i,j,k,ar,ai,br,bi,dr,di,v,vv,qa;
    boolean diff = (flags&0x1) != 0;
    boolean quad = (flags&0x2) != 0;
    for (i=j=0; i<n;) {
      ar = ibi[j++]; ai = ibi[j++];
      if (ar<0) {
        if (ai<0) { br=-ar; bi=-ai; qa=0x8000; }
        else      { br=ai;  bi=-ar; qa=0x4000; }
      } else {
        if (ai<0) { br=-ai; bi=ar; qa=0xC000; }
        else      { br=ar;  bi=ai; qa=0x0000; }
      }
      k = (bi&umask)>>(LB-QB) | (br&umask)>>LB;		// lookup table
      dr = br&lmask; di = bi&lmask;			// residuals
      v = iut0[k] + ((dr*iut1[k])>>(16-TB)) + ((di*iut2[k])>>(16-TB));
      if (quad) v = qa + (v>>1); else v <<=1;
      vv = v; if (diff) vv -= valueLastFPGA;
      ibo[i++] = (short)vv;
      valueLastFPGA = v;
    }
    if (diff && pass==0) ibo[0]=ibo[1];
    return n;
  }

  private void initFPGA() {
    int i,j,k,s;
    double h=.5;
    boolean diff = (flags&0x1) != 0;
    boolean quad = (flags&0x2) != 0;
    iut0 = new short[QW*QW];
    iut1 = new short[QW*QW];
    iut2 = new short[QW*QW];
    iut3 = new short[QW*QW];
    for (j=0; j<QW; j++) {
      for (i=0; i<QW; i++) {
        k = (i&tmask) | ((j&tmask)<<QB);
	int i0 = lutval(j+h,i+h);
	int i1 = lutval(j+h,i+1)-lutval(j+h,i+0);
	int i2 = lutval(j+1,i+h)-lutval(j+0,i+h);
        int i3 = lutamp(j+h,i+h);
        i1 <<= 1; i1 >>= 1;	// handle wrap
        i2 <<= 1; i2 >>= 1;	// handle wrap
        i0 -= (i1/2 + i2/2);
        s = quad? 14 : 16;	// preserve more precision
        iut0[k] = (short)(i0>>s);
        iut1[k] = (short)(i1>>s);
        iut2[k] = (short)(i2>>s);
        iut3[k] = (short)(i3>>16);
      }
    }
  }

  private void loadFPGA (Data dd) {
    int i=0, fs=QW;
    iut0 = new short[QW*QW];
    iut1 = new short[QW*QW];
    iut2 = new short[QW*QW];
    iut3 = new short[QW*QW];
    for (int tn=0; tn<=3; tn++) {
      for (int y=0; y<fs; y++) {
        for (int x=0; x<fs; x++) {
          setLUTentry(tn,y*fs+x,dd.getI(i++));
        }
      }
    }
  }

  private static final double scaleAMP = 2147483648.0;
  // allow values > 1 in corners by scaling to 1/2 fullscale and shifting by 1 last
  private static final double scaleLUT = 1073741824.0;

  private int lutval (double imag, double real) {
    double value=0;
    if (imag==0 && real==0) return 0;
    switch (mode) {
    case AM:
      double ai = imag/QW; double ar = real/QW;
      value = Math.sqrt( ar*ar + ai*ai );
      break;
    case FM: case PM:
      double p = Math.atan2( imag, real );
      value = (p/Math.PI);
      break;
    }
    return (int)(value*scaleLUT);
  }

  private int lutamp (double imag, double real) {
    double ai = imag/QW; double ar = real/QW;
    double value = Math.sqrt( ar*ar + ai*ai );
    return (int)Math.min(2147483647.0, value*scaleAMP);
  }

  public short getLUTentry (int tn, int index) {
    if (iut0==null) initFPGA();
    short[] tab = (tn==1)? iut1 : (tn==2)? iut2 : (tn==3)? iut3 : iut0;
    return tab[index];
  }

  public void setLUTentry (int tn, int index, short value) {
    if (iut0==null) initFPGA();
    short[] tab = (tn==1)? iut1 : (tn==2)? iut2 : (tn==3)? iut3 : iut0;
    tab[index] = value;
  }

  public int getFlags() {
    return flags;
  }

 }
}
