/**
  Implements a Fast Fourier Transform.

  @author Jeff Schoen
  @version $Id: FFT.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME FFT
#define CORE_AREA nxm_ice_core

/* Generic defines */
#include "CoreDefs.h"

typedef struct {
  HALO halo;
  int_4 flags;
  int_4 nfft;
  int_4 nout;
  int_4 navg;
  int_4 nmax;
  int_4 nexp;
  int_4 over;
  int_4 shift;
  int_4 pass;
  int_4 pow2;
  int_4 iavg;
  int_4 imax;
  real_4 *wind,*fppf,*ppbuf,*temp;
#if CORE_IS_GPU
  cufftHandle cuhandle;
  cufftComplex *cubuf;
#endif
} FFTPlan;

#include "CoreProtos.h"

/* Local prototypes */
#if CORE_IS_CPU
static void  fftprep (FFTPlan *plan);
static void  fft     (FFTPlan *plan, real_4 *data);
#endif
#if CORE_IS_CPU || CORE_IS_GPU
static int_4 window  (FFTPlan *plan, real_4 *data1, real_4 *data2);
static int_4 fppf    (FFTPlan *plan, real_4 *data1, real_4 *data2);
static int_4 average (FFTPlan *plan, real_4 *data1, real_4 *data2);
#endif

/* CORE Code */
#define MTAP 4096
#define flagsList "Inv,AC,Mag,Exp,Max,PPF,Real"
#define INV  0x01
#define AC   0x02
#define MAG  0x04
#define EXP  0x08
#define MAXM 0x10
#define PPF  0x20
#define REAL 0x40

void* FFT_alloc (char *config) {
  FFTPlan *plan = FFT_plan();
  HW_open(config);
  plan->flags = 0;
  plan->nfft = 256;
  plan->over = 0;
  plan->navg = 1;
  plan->nmax = 1;
  plan->shift = 0;
  plan->iavg = 0;
  plan->pass = 0;
  plan->wind = plan->fppf = plan->ppbuf = plan->temp = 0;
  return (void*)plan;
}

int_4 FFT_setkey (FFTPlan *plan, char *key, void *data, int_4 len) {
  if      (keyMatch("L:NFFT"))  { plan->nfft  = *(int_4*)data; }
  else if (keyMatch("L:NOUT"))  { plan->nout  = *(int_4*)data; }
  else if (keyMatch("L:NAVG"))  { plan->navg  = *(int_4*)data; }
  else if (keyMatch("L:NMAX"))  { plan->nmax  = *(int_4*)data; }
  else if (keyMatch("L:NEXP"))  { plan->nexp  = *(int_4*)data; }
  else if (keyMatch("L:OVER"))  { plan->over  = *(int_4*)data; }
  else if (keyMatch("L:SHIFT")) { plan->shift = *(int_4*)data; }
  else if (keyMatch("L:FLAGS")) { plan->flags = *(int_4*)data; }
  else if (keyMatch("F:FPPF"))   { plan->fppf = (real_4*)data; HW_loadPPF(COREG_USER0,4,len/4,plan->fppf); }
  else if (keyMatch("F:WINDOW")) { plan->wind = (real_4*)data; HW_loadTaps(COREG_USER0,len/4,plan->wind); }
  else return 0;
  return len;
}

int_4 FFT_getkey (FFTPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:NFFT"))  { *(int_4*)data = plan->nfft; return 4; }
  if (keyMatch("L:NOUT"))  { *(int_4*)data = plan->nout; return 4; }
  if (keyMatch("L:NAVG"))  { *(int_4*)data = plan->navg; return 4; }
  if (keyMatch("L:NEXP"))  { *(int_4*)data = plan->navg; return 4; }
  if (keyMatch("L:OVER"))  { *(int_4*)data = plan->over; return 4; }
  if (keyMatch("L:SHIFT")) { *(int_4*)data = plan->shift; return 4; }
  if (keyMatch("L:FLAGS")) { *(int_4*)data = plan->flags; return 4; }
  if (keyMatch("S:FLAGSLIST")) { strcpy(data,flagsList); return strlen(flagsList); }
  return 0;
}

int_4 FFT_init (FFTPlan *plan) {
  int i, sys;
  plan->pow2 = pow2ge(plan->nfft);
#if CORE_IS_FPGA
  sys = 0x0000;
  sys |= (plan->flags&MAG)? 0x4000:0x8000;
  sys |= (plan->flags&REAL)?0x0000:0x0800;
  HW_write(0<<2,sys);	
  HW_write(1<<2,pow2sizem1(plan->navg));
  HW_write(2<<2,fftscale(plan->shift,plan->pow2));	
  HW_write(4<<2,plan->over-1);	
  HW_write(5<<2,pow2sizem1(plan->nfft));	
  HW_write(7<<2,plan->flags);	
  HW_write(9<<2,pow2sizem1(plan->nout));
  HW_write(10<<2,pow2sizem1(plan->nmax));
  HW_write(11<<2,pow2sizem1(plan->nexp));
  HW_write(0<<2,sys|0x1);	
#elif CORE_IS_GPU
  cufftPlan1d(&plan->cuhandle, plan->nfft, (plan->flags&REAL)? CUFFT_R2C:CUFFT_C2C, 1);
  cudaMalloc((void**)&plan->cubuf, sizeof(float2) * ((plan->flags&REAL)?  plan->nfft/2+1 : plan->nfft));
  if (plan->fppf!=0) { plan->ppbuf = malloc(plan->nfft*2*4*4); for(i=0; i<plan->nfft*2*4; i++) plan->ppbuf[i]=0.0; }
#elif CORE_IS_CPU
  plan->temp = malloc(plan->nfft*4*8);
  if (plan->fppf!=0) { plan->ppbuf = malloc(plan->nfft*2*4*4); for(i=0; i<plan->nfft*2*4; i++) plan->ppbuf[i]=0.0; }
  fftprep(plan);
#endif
  return 0;
}

int_4 FFT_work (FFTPlan *plan) {
  int_4 i, n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#if CORE_IS_FPGA
  int_4 bpei=(plan->flags&REAL)? 2:4, bpeo=(plan->flags&MAG)? 4:4;
  int_4 *lbuf=(int_4*)buf2; int_2 *ibuf=(int_2*)buf2; real_4 *fbuf=(real_4*)buf2;
  if ((plan->flags&MAG)==0) n2/=2; /* will expand by two later */
  n = HW_proc (buf1,n1, buf2,n2); n2=n/bpeo;
  if (plan->flags&MAG) for (i=0; i<n2; i++) fbuf[i]=lbuf[i]*(1.0/65536.0);
  else { n2*=2;  for (i=n2-1; i>=0; i--) fbuf[i]=ibuf[i]; }
  n = n2*4;
#elif CORE_IS_GPU
  int_4 xfer, bpei=(plan->flags&REAL)? 4:8, bpeo=(plan->flags&MAG)? 4:4;
  if (n1<plan->nfft*bpei) return 0;
  if (plan->flags&PPF) n=fppf (plan,(real_4*)buf1,(real_4*)buf1);
  else               n=window (plan,(real_4*)buf1,(real_4*)buf1);
  if (n>0) {
    xfer  = sizeof(float2) * ((plan->flags&REAL)?  plan->nfft/2 : plan->nfft);
    cudaMemcpy(plan->cubuf, (cufftComplex*)buf1, xfer, cudaMemcpyHostToDevice);
    if (plan->flags&REAL) cufftExecR2C(plan->cuhandle, (cufftReal *)plan->cubuf, (cufftComplex *)plan->cubuf);
    else                  cufftExecC2C(plan->cuhandle, (cufftComplex *)plan->cubuf, (cufftComplex *)plan->cubuf, CUFFT_FORWARD);
    cudaMemcpy((cufftComplex*)buf1, plan->cubuf, xfer, cudaMemcpyDeviceToHost);
    n = average (plan,(real_4*)buf1,(real_4*)buf2) * bpeo;
  }
#elif CORE_IS_CPU
  int_4 bpei=(plan->flags&REAL)? 4:8, bpeo=(plan->flags&MAG)? 4:4;
  if (n1<plan->nfft*bpei) return 0;
  if (plan->flags&PPF) n=fppf (plan,(real_4*)buf1,(real_4*)buf1);
  else               n=window (plan,(real_4*)buf1,(real_4*)buf1);
  if (n>0) {
    fft (plan,(real_4*)buf1);
    n = average (plan,(real_4*)buf1,(real_4*)buf2) * bpeo;
  }
#endif
  plan->pass++;
  return n;
}

int_4 FFT_free (FFTPlan *plan) {
#if CORE_IS_GPU
  cudaFree(plan->cubuf);
  cufftDestroy(plan->cuhandle);
  if (plan->ppbuf) free(plan->ppbuf);
  free(plan->temp);
#elif CORE_IS_CPU
  free(plan->temp);
#endif
  HW_close();
  free(plan);
  return 0;
}


/* Local Code - not optimized */
#if CORE_IS_CPU
int_4 bitrev (int_4 fn, int_4 bits) {
  int_4 i,rn=0; 
  for (i=0; i<bits; i++) if ((fn&(1<<i))!=0) rn |= (1<<(bits-1-i));
  return rn;
}

#define TWOPI  6.283185307179586
void fftprep (FFTPlan *plan) {
  int_4 i, m=plan->pow2, n=plan->nfft, *r=(int_4*)plan->temp+n*4; 
  real_4 *s=plan->temp+n*0, *c=plan->temp+n*1; 
  for (i=0; i<n; i++) {
    s[i]=(real_4)sin(TWOPI*i/n);
    c[i]=(real_4)cos(TWOPI*i/n);
    r[i]=bitrev(i,m);
  }
}

void fft (FFTPlan *plan, real_4 *cx) {
  int_4 i,j,k,n1,n2,a,ai,m=plan->pow2,n=(1<<m);
  real_4 c,s,e,vx,vy,tx,ty;
  real_4 *sin=plan->temp+n*0, *cos=plan->temp+n*1, *x=plan->temp+n*2, *y=plan->temp+n*3; 
  int_4  *rev=(int_4*)plan->temp+n*4;
  int_4  real=(plan->flags&REAL);

  // bit-reverse and complex unpack
  if (real) for (i=0; i<n; i++) {
    k = rev[i];
    x[k]=cx[i];
    y[k]=0;
  }
  else for (i=j=0; i<n; i++) {
    k = rev[i];
    x[k]=cx[j++];
    y[k]=cx[j++];
  }
  // radix-2 butterfly passes
  for (i=0; i<m; i++) {
    n1 = 1<<i;
    n2 = n1+n1;
    ai = 1<<(m-i-1);
    for (a=j=0; j<n1; a+=ai,j++) {
      c = cos[a];
      s = -sin[a];
      for (k=j; k<n; k=k+n2) {
        vx = x[k+n1];
        vy = y[k+n1];
        tx = c*vx - s*vy;
        ty = s*vx + c*vy;
        vx = x[k];
        vy = y[k];
        x[k+n1] = vx - tx;
        y[k+n1] = vy - ty;
        x[k]    = vx + tx;
        y[k]    = vy + ty;
      }
    }
  }
  // complex pack
  k = real? n/2 : n;
  for (i=j=0; i<k; i++) {
    cx[j++]=x[i];
    cx[j++]=y[i];
  }
}

#endif

#if CORE_IS_CPU || CORE_IS_GPU

int_4 window (FFTPlan *plan, real_4 *a, real_4 *b) {
  int_4 i,j,n=plan->nfft;
  real_4 *wind=plan->wind, factor=(real_4)pow(2,plan->shift-plan->pow2);
 if (plan->flags&REAL)
  for (factor*=2,i=0; i<n; i++) {
    b[i] = a[i]*wind[i]*factor;
  }
 else
  for (i=j=0; i<n; i++) {
    b[j] = a[j]*wind[i]*factor; j++;
    b[j] = a[j]*wind[i]*factor; j++;
  }
  return 1;
}

int_4 fppf (FFTPlan *plan, real_4 *a, real_4 *b) {
  int_4 i,ii,jj,n,i0,nx, nfft=plan->nfft,pass=plan->pass;
  int_4 ntap=4, nfft2=(plan->flags&REAL)? nfft:nfft*2, ntap2=ntap*nfft2;
  int_4 n0=(pass%ntap)*nfft2, n1=((pass+ntap-1)%ntap)*nfft2;
  real_4 *ppbuf=plan->ppbuf, *fppf=plan->fppf;
  real_4 sumr,sumi,factor=pow(2,plan->shift-plan->pow2);
  memcpy(ppbuf+n1,a,nfft2*4);
  if (pass<3) return 0;
 if (plan->flags&REAL) {
  for (factor*=2,i=0,ii=0,jj=n1; i<nfft; i++) {
    sumr = 0;
    for (n=0,i0=i,nx=n0+ii; n<ntap; n++,i0+=nfft,nx=(nx+nfft2)%ntap2) {
      sumr += (fppf[i0]*ppbuf[nx]);
    }
    b[ii++] = factor*sumr;
  }
 } else {
  for (i=0,ii=0,jj=n1; i<nfft; i++) {
    sumr = sumi = 0;
    for (n=0,i0=i,nx=n0+ii; n<ntap; n++,i0+=nfft,nx=(nx+nfft2)%ntap2) {
      sumr += (fppf[i0]*ppbuf[nx+0]);
      sumi += (fppf[i0]*ppbuf[nx+1]);
    }
    b[ii++] = factor*sumr;
    b[ii++] = factor*sumi;
  }
 }
  return 1;
}

int_4 average (FFTPlan *plan, real_4 *a, real_4 *b) {
  int_4 i,j, m=plan->nfft, n=plan->nout, max=(plan->flags&MAXM)!=0;
  real_4 ar,ai,scale, *psd=plan->temp+m*6;
  if ((plan->flags&MAG)!=0) {
    for (i=0,j=0; i<n; i++) {
      ar=a[j++]; ai=a[j++];
      b[i] = (real_4)sqrt(ar*ar+ai*ai);
    }
  }
  else {
    n *= 2;
    for (i=0; i<n; i++) b[i]=a[i];
  }
  if (plan->iavg==0)  for (i=0; i<n; i++) psd[i]=b[i];
  else if (max) for (i=0; i<n; i++) { if (b[i]>psd[i]) psd[i]=b[i]; }
  else          for (i=0; i<n; i++) psd[i]+=b[i];
  if (++plan->iavg==plan->navg) {
    plan->iavg=0;
    scale = (real_4)(1.0/(max?1:plan->navg));
    for (i=0; i<n; i++) b[i] = scale*psd[i];
    return n;
  }
  return 0;
}
#endif
