package nxm.ice.core;

import nxm.ice.lib.*;
import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;
import nxm.sys.libm.Fft;

/**
  Implements a Fast Fourier Transform function

  @author Jeff Schoen
  @version $Id: FFT.java,v 1.8 2011/05/24 13:21:13 jgs Exp $
*/
public class FFT extends CoreFactory {

 // native implementation wrappers
 public static class CPU extends CoreNative { public native long alloc(); }
 public static class GPU extends CoreNative { public native long alloc(); }
 public static class VHS extends CoreNative { public native long alloc(); }
 public static class ICE extends CoreNative { public native long alloc(); }

 // java implementation
 public static class JVM extends Core {

  static String flagsList = "Inv,AC,Mag,Exp,Max,PPF,Real";
  static int INV=0x1,AC=0x2,MAG=0x4,EXP=0x8,MAX=0x10,PPF=0x20,REAL=0x40;
  int nfft,nout,navg,iavg,nmax,imax,nexp,flags,over,shift,pass,pow2;
  boolean ppf,real;
  double scale=1;
  float[] wind,fppf,ppbuf,pavg,pmax,pexp;
  Fft fft;

  /** set parameters */
  public void set (String key, Data value) {
         if (key.equals("NFFT"))   nfft = value.toL();
    else if (key.equals("NOUT"))   nout = value.toL();
    else if (key.equals("NAVG"))   navg = value.toL();
    else if (key.equals("NMAX"))   nmax = value.toL();
    else if (key.equals("NEXP"))   nexp = value.toL();
    else if (key.equals("OVER"))   over = value.toL();
    else if (key.equals("SHIFT"))  shift = value.toL();
    else if (key.equals("FLAGS"))  flags = value.toL();
    else if (key.equals("SCALE"))  scale = value.toD();
    else if (key.equals("WINDOW")) wind = value.castF(true);
    else if (key.equals("FPPF"))   fppf = value.castF(true);
    else super.set(key,value);
  }

  /** get parameters */
  public Data get (String key, byte type) {
         if (key.equals("NFFT"))  return new Data(nfft);
    else if (key.equals("NAVG"))  return new Data(navg);
    else if (key.equals("NMAX"))  return new Data(nmax);
    else if (key.equals("NEXP"))  return new Data(nexp);
    else if (key.equals("OVER"))  return new Data(over);
    else if (key.equals("SHIFT")) return new Data(shift);
    else if (key.equals("SCALE")) return new Data(scale);
    else if (key.equals("FLAGS")) return new Data(flags);
    else if (key.equals("FLAGSLIST")) return new Data(flagsList);
    else return super.get(key,type);
  }

  /** initialize resources */
  public int open () {
    ppf = (flags&PPF) != 0;
    real = (flags&REAL) != 0;
    fft = new Fft(nfft, real? Fft.FORWARD|Fft.REAL|Fft.PACKED : Fft.FORWARD|Fft.COMPLEX);
    for (pow2=2; Math.pow(2,pow2)<nfft; pow2++);
    if (ppf) ppbuf = new float[nfft*2*4];
    iavg = imax = pass = 0;
    return 0;
  }

  /** free resources */
  public int close () {
    fft.free();
    return 0;
  }

  /** process a buffer of data */
  public int process (byte[] a, int na, byte[] b, int nb) {
    int n=0, bpei=real?4:8;
    if (na<nfft*bpei) return 0; // no padding
    float[] af = Convert.castF(a,true);
    float[] bf = Convert.castF(b,false);
    if (ppf) n=processFPPF(4,nfft,af,af);
    else     n=processWindow(nfft,af,af);
    if (n>0) {
      fft.work(af);
      if (real) af[1]=0;
      n = processBlock(nout,af,bf) * 4;
    }
    Convert.uncast(af,a,false);
    Convert.uncast(bf,b,true);
    pass++;
    return n;
  }

  private int processWindow (int nfft, float[] a, float[] b) {
    float factor = (float)(scale*Math.pow(2,shift-pow2));
    if (real) for (int i=0; i<nfft; i++) {
      b[i] = a[i]*wind[i]*factor;
    }
    else for (int i=0,j=0; i<nfft; i++) {
      b[j] = a[j]*wind[i]*factor; j++;
      b[j] = a[j]*wind[i]*factor; j++;
    }
    return 1;
  }

  private int processFPPF (int ntap, int nfft, float[] a, float[] b) {
    int nfft2 = real? nfft:nfft*2;
    int ntap2 = ntap*nfft2;
    int n0 = ((pass)%ntap)*nfft2;
    int n1 = ((pass+ntap-1)%ntap)*nfft2;
    float sumr,sumi, factor=(float)(scale*Math.pow(2,shift-pow2));
    System.arraycopy(a,0, ppbuf,n1, nfft2);
    if (pass<ntap-1) return 0;
    if (real) for (int i=0,ii=0; i<nfft; i++) {
      sumr = 0;
      for (int n=0,i0=i,nx=n0+ii; n<ntap; n++,i0+=nfft,nx=(nx+nfft2)%ntap2) {
        sumr += fppf[i0]*ppbuf[nx+0];
      }
      b[ii++] = factor*sumr;
    }
    else for (int i=0,ii=0; i<nfft; i++) {
      sumr = sumi = 0;
      for (int n=0,i0=i,nx=n0+ii; n<ntap; n++,i0+=nfft,nx=(nx+nfft2)%ntap2) {
        sumr += fppf[i0]*ppbuf[nx+0];
        sumi += fppf[i0]*ppbuf[nx+1];
      }
      b[ii++] = factor*sumr;
      b[ii++] = factor*sumi;
    }
    return 1;
  }

  private int processBlock (int n, float[] a, float[] b) {
    boolean mag = (flags&MAG)!=0;
    boolean max = (flags&MAX)!=0;
    if (mag) {
      for (int i=0,j=0; i<n; i++) {
        float ar=a[j++], ai=a[j++];
        b[i] = (float)Math.sqrt(ar*ar+ai*ai);
      }
    }
    else {
      n *= 2;
      for (int i=0; i<n; i++) b[i]=a[i];
    }
    if (pavg==null) pavg = new float[n];
    if (iavg==0)  for (int i=0; i<n; i++) pavg[i]=b[i];
    else if (max) for (int i=0; i<n; i++) { if (b[i]>pavg[i]) pavg[i]=b[i]; }
    else          for (int i=0; i<n; i++) pavg[i]+=b[i];
    if (++iavg==navg) {
      iavg=0;
      float scale = (float)(1.0/(max?1:navg));
      for (int i=0; i<n; i++) b[i] = scale*pavg[i];
      return n;
    } 
    return 0;
  }

 }
}
