/**
  Implements a Quick Tune/Filter/Decimate Function

  @author Jeff Schoen
  @version $Id: QTFD.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME QTFD
#define CORE_AREA nxm_ice_core

/* Generic defines */
#include "CoreDefs.h"

typedef struct {
  HALO halo;
  int_4 dec;
  int_4 gain;
  int_4 order;
  int_4 ndiv;
  int_4 ntap;
  int_4 flags;
  double freq;
  double phase;
  float *taps;
  float *data;
} QTFDPlan;

#include "CoreProtos.h"

/* CORE Code */

void* QTFD_alloc (char *config) {
  QTFDPlan *plan = QTFD_plan();
  HW_open(config);
  plan->dec  = 4;
  plan->gain = 0;
  plan->order = 5;
  plan->ndiv = 256;
  plan->ntap = 0;
  plan->freq = 0.0;
  plan->phase = 0.0;
  return (void*)plan;
}

int_4 QTFD_setkey (QTFDPlan *plan, char *key, void *data, int_4 len) {
  int_4  *ldata = (int_4*)data;
  real_4 *fdata = (real_4*)data;
  real_8 *ddata = (real_8*)data;
       if (keyMatch("L:DECIMATION")) { plan->dec = ldata[0]; }
  else if (keyMatch("D:FREQUENCY"))  { plan->freq = ddata[0]; }
  else if (keyMatch("L:GAIN"))       { plan->gain = ldata[0]; }
  else if (keyMatch("L:ORDER"))      { plan->order = ldata[0]; }
  else if (keyMatch("F:PFIR"))       { plan->ntap = len/4; plan->taps = fdata; }
  else return 0;
  return len;
}

int_4 QTFD_getkey (QTFDPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:DECIMATION")) { *(int_4*)data = plan->dec; return 4; }
  if (keyMatch("D:FREQUENCY"))  { *(double*)data = plan->freq; return 8; }
  if (keyMatch("L:GAIN"))       { *(int_4*)data = plan->gain; return 4; }
  if (keyMatch("L:ORDER"))      { *(int_4*)data = plan->order; return 4; }
  return 0;
}

int_4 QTFD_init (QTFDPlan *plan) {
#if CORE_IS_FPGA
  int_4 sysreg=0x8101;
  HW_write(COREG_SYS,0x0);				// disable
  HW_write(COREG_DEC,_max(0,plan->dec-1));		// deci
  HW_write(COREG_FREQ,R2L(-plan->freq)&0xFF000000);	// freq to 256 bins
  qfirloadtaps (plan);					// fill out rotate lut
  HW_write(COREG_SYS,sysreg);				// enable
#elif CORE_IS_CPU
  int_4 i;
  plan->data = malloc(plan->ntap*4);
  for (i=0; i<plan->ntap; i++) plan->data[i]=0;
#endif
  return 0;
}

int_4 QTFD_work (QTFDPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#if CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#elif CORE_IS_CPU
  n = CPU_proc (plan, buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 QTFD_free (QTFDPlan *plan) {
#if CORE_IS_FPGA
  HW_close();
#elif CORE_IS_CPU
  free(plan->data);
#endif
  free(plan);
  return 0;
}

/*
  Special lut for complex rotate with gain 
*/
int_4 qfirloadtaps (QTFDPlan *plan) {
  real_8 phase;
  int_4 i,tapsin,tapcos,tapcx;
  int_4 scale = _min(0x7FFF,(int_4)(0x2000 * pow(2.0,(double)plan->gain/6)));
//printf("scale=%08x ndec=%d\n",scale,plan->dec);
  // load sincos lut
  HW_write(COREG_SYS,0x1);				// disable
  for (i=0; i<plan->ndiv; i++) {
    phase  = (i*6.283185307179586/plan->ndiv);
    tapsin = (int_4)(scale*sin(phase));
    tapcos = (int_4)(scale*cos(phase));
    tapcx  = ((tapsin<<16)&0xFFFF0000) | (tapcos&0xFFFF);
//printf("tapcx=%08x\n",tapcx);
    HW_write(COREG_USER4,tapcx);
  }
  HW_write(COREG_SYS,0x0);				// disable
  // load filter 
  HW_loadTaps(COREG_USER5,plan->ntap,plan->taps);
  return 0;
}

int_4 CPU_proc (QTFDPlan *plan, int_1 *buf1, int_4 n1, int_2 *buf2, int_4 n2) {
  int_4 i,j,n=0, ndec=plan->dec, ntap=plan->ntap;
  float dr,di, *taps=plan->taps, *fbuf=plan->data;
  double freq=plan->freq/2,phase,dsin,dcos;
  double scale = 256*2 * pow(2.0,(double)plan->gain/6); // apply gain and SB to CI scaling
  for (i=0; i<n1;) {
    for (j=ntap-1; j>=ndec; j--) fbuf[j]=fbuf[j-ndec];
    for (j=ndec-1; j>=0;    j--) fbuf[j]=buf1[i++];
    dr = di = 0;
    phase = plan->phase;
    for (j=0; j<ntap; j++) { 
      dsin = scale*sin(6.283185307179586*phase);
      dcos = scale*cos(6.283185307179586*phase);
      dr += fbuf[j]*dcos*taps[j];
      di += fbuf[j]*dsin*taps[j];
      phase += freq;
    }
    buf2[n++]=dr;
    buf2[n++]=di;
    plan->phase -= freq*ndec;
  }
  return n*2;
}

