/**
  Implements a Frame Based Smoother functions.

  @author Jeff Schoen
  @version $Id: Smooth.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME Smooth
#define CORE_AREA nxm_ice_core

/* Generic defines */
#include "CoreDefs.h"

typedef struct {
  HALO halo;
  int_4 frame;
  int_4 nexp;
  int_4 iavg;
  int_4 flags;
  double scale;
  real_4 *temp;
  int_4 pass;
  int_4 aminv;
} SmoothPlan;

#include "CoreProtos.h"

/* Local prototypes */
#if CORE_IS_CPU
static int_4 smooth  (SmoothPlan *plan, int_4 *data1, real_4 *data2);
#endif

/* CORE Code */
#define flagsList "Mag,Log,Exp,Fix,Max"
#define MAG  0x01
#define LOG  0x02
#define EXP  0x04
#define FIX  0x08
#define MAXM 0x10

void* Smooth_alloc (char *config) {
  SmoothPlan *plan = Smooth_plan();
  HW_init(config);
  plan->flags = 0;
  plan->frame = 256;
  plan->nexp  = 1;
  plan->scale = 1.0;
  plan->temp  = 0;
  return (void*)plan;
}

int_4 Smooth_setkey (SmoothPlan *plan, char *key, void *data, int_4 len) {
       if (keyMatch("L:FLAGS")) { plan->flags = *(int_4*)data; }
  else if (keyMatch("L:FRAME")) { plan->frame = *(int_4*)data; }
  else if (keyMatch("L:NEXP"))  { plan->nexp  = *(int_4*)data; }
  else if (keyMatch("D:SCALE")) { plan->scale = *(double*)data; }
  else return 0;
  return len;
}

int_4 Smooth_getkey (SmoothPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:FRAME")) { *(int_4*)data = plan->frame; return 4; }
  if (keyMatch("L:NEXP"))  { *(int_4*)data = plan->nexp; return 4; }
  if (keyMatch("D:SCALE")) { *(double*)data = plan->scale; return 8; }
  if (keyMatch("L:FLAGS")) { *(int_4*)data = plan->flags; return 4; }
  if (keyMatch("S:FLAGSLIST")) { strcpy(data,flagsList); return strlen(flagsList); }
  return 0;
}

int_4 Smooth_init (SmoothPlan *plan) {
  int_4 i;
  plan->iavg = 0;
  plan->pass = 0;
  plan->aminv = 0;
#if CORE_IS_CPU
  plan->temp = malloc(plan->frame*4);
  for (i=0; i<plan->frame; i++) plan->temp[i]=0;
#endif
  return 0;
}

int_4 Smooth_work (SmoothPlan *plan) {
  int_4 i, n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#if CORE_IS_CPU
  n = smooth (plan,(int_4*)buf1,(real_4*)buf2);
#endif
  plan->pass++;
  return n*4;
}

int_4 Smooth_free (SmoothPlan *plan) {
  HW_close();
#if CORE_IS_CPU
  free(plan->temp);
#endif
  free(plan);
  return 0;
}


/* Local Code  */
#if CORE_IS_CPU

int_4 smooth (SmoothPlan *plan, int_4 *a, real_4 *b) {
  int_4 i,j,v,adj=0,minv,aminv=plan->aminv,n=plan->frame;
  real_4 *tank=plan->temp;
  double factor = (real_4)plan->scale;
  double p=.125, q=1.0-p;

  if (plan->flags&FIX) {
    v = a[0]+a[1]+a[2]+a[3]+a[4];
    minv = v;
    for (i=3; i<n-3; i++) {
      v = v+a[i+2]-a[i-3];
      if (v<minv) minv = v;
    }
    if (aminv==0) aminv=minv;
#if DEBUG
    printf("Min=%d Amin=%d Diff=%d\n",minv,aminv,minv-aminv);
#endif
    if (minv-aminv>8000) adj = (int_4)(-.2*(minv-aminv));
    aminv = (int_4)(p*minv+q*aminv);
    plan->aminv=aminv;
  }

  for (i=0; i<n; i++) b[i] = a[i]+adj;

  if (plan->nexp > 1) {
    p=1.0/plan->nexp; q=1.0-p;
    for (i=0; i<n; i++) { tank[i] = tank[i]*q + b[i]*p; b[i]=tank[i]; }
  }

  for (i=0; i<n; i++) {
    b[i] = (20/2.3)*log(factor*b[i]); 
  }
  return n;
}

#endif
