/**
  Implements a Tune/Filter/Decimate Function

  @author Jeff Schoen
  @version $Id: TFDD.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME TFDD
#define CORE_AREA nxm_ice_core

/* Generic defines */
#include "CoreDefs.h"

typedef struct {
  HALO halo;
  int_4 dec;
  int_4 gain;
  double freq;
  double ratio;
  int_4 ovsr;
  int_4 rmon;
  int_4 flags;
} TFDDPlan;

#include "CoreProtos.h"

#define flagsList "CXI,CXO,HTF,RESAMP"
#define CXI    0x01
#define CXO    0x02
#define HTF    0x04
#define RESAMP 0x08

#define GCFIR_GAIN 0.9899
#define GPFIR_GAIN 1.1494

#ifdef CORE_IS_FPGA
int_4 getScale (TFDDPlan *plan) {
  int_4 gain=plan->gain, decratio=max(1,plan->dec/2), ovsr=plan->ovsr;
  int_4 coarse,scale,gaincomp,cdec;
  double dgain,sgain,fgain;
  if (ovsr>1) gain += log2(ovsr*6);
  cdec = decratio*4;
  coarse = (gain>=0)? (gain/6) : (gain-5)/6;
  scale = (int_4)(5.0*log2((double)cdec/4)+0.98) - coarse;
  dgain = powi(cdec/4,5) * powi(2,-scale);
  scale = max(0,scale+5); /* from 20b to 18b data */
  sgain = 1.0;
  sgain *= GCFIR_GAIN;
  sgain *= GPFIR_GAIN;
  fgain = pow( 2.0, ((double)gain)/6.0 ) / (sgain*dgain);
  if (decratio>1 && fgain>=2.0 && scale>0) { fgain/=2; scale--; }
  gaincomp = (int_4)(0x8000 * fgain);
  if (gaincomp>=0x10000) gaincomp = 0xffff;
  return (scale<<16)|gaincomp;
}
#endif

/* CORE Code */

void* TFDD_alloc (char *config) {
  TFDDPlan *plan = TFDD_plan();
  HW_open(config);
  plan->dec = 16;
  plan->gain = 0;
  plan->ovsr = 1;
  plan->rmon = 0;
  plan->freq = 0.0;
  return (void*)plan;
}

int_4 TFDD_setkey (TFDDPlan *plan, char *key, void *data, int_4 len) {
  int_4  *ldata = (int_4*)data;
  real_4 *fdata = (real_4*)data;
  real_8 *ddata = (real_8*)data;
       if (keyMatch("L:DECIMATION")) { plan->dec = ldata[0]; }
  else if (keyMatch("D:FREQUENCY"))  { plan->freq = ddata[0]; HW_write(COREG_FREQ,R2L(-plan->freq)); }  // real-time freq
  else if (keyMatch("D:RATIO"))      { plan->ratio = ddata[0]; }
  else if (keyMatch("L:RMON"))       { plan->rmon = ldata[0]; }
  else if (keyMatch("L:OVSR"))       { plan->ovsr = ldata[0]; }
  else if (keyMatch("L:GAIN"))       { plan->gain = ldata[0]; HW_write(COREG_GAIN,getScale(plan)); } // real-time gain
  else if (keyMatch("L:FLAGS"))      { plan->flags = ldata[0]; }
#ifdef CORE_IS_FPGA
  else if (keyMatch("F:CFIR"))       { HW_loadFIR(COREG_CFIR,3,(plan->dec==1)?1:2,len/4,fdata); }
  else if (keyMatch("F:PFIR"))       { HW_loadFIR(COREG_PFIR,3,2,len/4,fdata); }
  else if (keyMatch("F:RFIR"))       { HW_loadRTaps(COREG_RFIR,len/4,fdata); }
#endif
  else return 0;
  return len;
}

int_4 TFDD_getkey (TFDDPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:DECIMATION")) { *(int_4*)data = plan->dec; return 4; }
  if (keyMatch("D:FREQUENCY"))  { *(double*)data = plan->freq; return 8; }
  if (keyMatch("D:RATIO"))      { *(double*)data = plan->ratio; return 8; }
  if (keyMatch("L:GAIN"))       { *(int_4*)data = plan->gain; return 4; }
  if (keyMatch("L:FLAGS"))      { *(int_4*)data = plan->flags; return 4; }
  if (keyMatch("S:FLAGSLIST"))  { strcpy(data,flagsList); return strlen(flagsList); }
  return 0;
}

int_4 TFDD_init (TFDDPlan *plan) {
#ifdef CORE_IS_FPGA
  int_4 sysreg=0x0011;
  if (plan->flags&CXI)    sysreg |= 0x0800;
  if (plan->flags&CXO)    sysreg |= 0x8000;
  if (plan->flags&RESAMP) sysreg |= 0x0040;
  HW_write(COREG_SYS,0x0);           // disable
  HW_write(COREG_DEC,max(0,plan->dec/2-1)); // deci
  HW_write(COREG_GAIN,getScale(plan)); // scaling
  HW_write(COREG_RATE,plan->rmon);  // ratio M/N
  HW_write(COREG_RATIO,(int_4)(0x10000000/plan->ratio)); // ratio
  HW_write(COREG_FRAME,plan->ovsr);  // ovsr
  HW_write(COREG_FREQ,R2L(-plan->freq)); // freq
  HW_write(COREG_FLAG,0);            // poff
  HW_write(COREG_SYS,sysreg);        // enable
#endif
  return 0;
}

int_4 TFDD_work (TFDDPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#ifdef CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 TFDD_free (TFDDPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

