/**
  Implements VRT packet functions 

  @author Jeff Schoen
  @version $Id: VRT.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME VRT
#define CORE_AREA nxm_ice_core

/* CORE defines */
#include "CoreDefs.h"

/* CORE Plan handle */
typedef struct {
  HALO halo;
  int_4 sys;
  int_4 fmts;
  int_4 mode;
  real_8 dsec;
  int_4 *pkthdr;
  int_4 flags;
  int_4 bits;
} VRTPlan;

#include "CoreProtos.h"

/* CORE Code */

void* VRT_alloc (char *config) {
  VRTPlan *plan = VRT_plan();
  HW_open(config);
  plan->sys  = 0;
  plan->pkthdr = NULL;
  plan->flags = 0;
  plan->bits = 0;
  plan->dsec = 0.0;
  return (void*)plan;
}

int_4 VRT_setkey (VRTPlan *plan, char *key, void *data, int_4 len) {
  int_4 *ldata = (int_4*)data;
  real_8 *ddata = (real_8*)data;
  if (keyMatch("L:FORMATS"))     { plan->fmts  = ldata[0]; }
  else if (keyMatch("PKTHDR"))   { plan->pkthdr= ldata; }
  else if (keyMatch("L:FLAGS"))  { plan->flags = ldata[0]; }
  else if (keyMatch("L:MODE"))   { plan->mode  = ldata[0]; }
  else if (keyMatch("L:BITS"))   { plan->bits  = ldata[0]; }
  else if (keyMatch("D:DSEC"))   { plan->dsec  = ddata[0]; }
  else return 0;
  return len;
}

int_4 VRT_getkey (VRTPlan *plan, char *key, void *data, int_4 len) {
  int_4 *ldata = (int_4*)data;
  if (keyMatch("L:FORMATS")) { ldata[0] = plan->fmts; return 4; }
  if (keyMatch("L:FLAGS"))   { ldata[0] = plan->flags; return 4; }
  if (keyMatch("L:MODE"))    { ldata[0] = plan->mode; return 4; }
  if (keyMatch("D:DSEC"))    { ldata[0] = plan->dsec; return 8; }
  return 0;
}

/*
  seq = 0x0000WVYX
  3:0   selX
  5:4   selY 0=hdr 1=byp 2=dmux 3=emux + 
  7:6   selZ 0=noswap 4=swap4 8=swap2 C=dual
  9:8   V 1=rd4 2=rd8 
  13:12 W 1=wr4 2=wr8  

  23:16 hAdr

  24 hrA read reg A, clr B
  25 hrB read reg B
  26 hwR write register
  27 hsV set value
  
  28 hsC set carry
  29 hrC register carry
  30 hsE set enable
  31 hrE register enable
  
*/
int_4 nibinc (int_4 *seq, int_4 i, int_4 adr, int_4 nibs) {
  int_4 k,func;
  for (k=0; k<nibs; k++) {
    func = (k==0)? 0x1900:0x0100;
    seq[i++] |= ( func |adr)<<16;
    seq[i++] |= (0x2400|adr)<<16;
    seq[i++] |= (0x0000|adr)<<16;
    adr++;
  }
  return i;
}

int_4 nibadd (int_4 *seq, int_4 i, int_4 adr, int_4 bdr, int_4 nibs) {
  int_4 k;
  seq[i++] |= (0x1000<<16); // clear carry
  for (k=0; k<nibs; k++) {
    seq[i++] |= (0x0100|adr)<<16; // readA
    seq[i++] |= (0x0200|bdr)<<16; // readB
    seq[i++] |= (0x0000|adr)<<16;
    seq[i++] |= (0x2400|adr)<<16; // writeR
    seq[i++] |= (0x0000|adr)<<16;
    adr++; bdr++;
  }
  return i;
}

int_4 VRT_init (VRTPlan *plan) {
  // note that size includes all header fields
  int_8 dpsec = (int_8)(plan->dsec*1.0e12+.5);
  int_8 dpdif = (int_8)(0x10000000000LL-1000000000000LL);
  int_4 bits = (plan->pkthdr[3]&0x3F)+1;
  int_4 size = (plan->pkthdr[0]&0x1FF);
  int_4 tlr  = (plan->pkthdr[0]&0x04000000)?1:0;
#if CORE_IS_FPGA || CORE_IS_ICELIB
  // program the sequencer engine
  int_4 i, j=0, k, s, sys, seq[512];
  int_4 ibps = getFormatBits(plan->fmts,1);
  int_4 obps = getFormatBits(plan->fmts,2);
  int_4 n = size - (tlr?8:7);
  if (plan->bits!=0) bits = plan->bits;
  // system registers
  //printf("VRT tlr=%d hdr=%08x ibps=%d obps=%d bits=%d\n",tlr,plan->pkthdr[0],ibps,obps,bits);
  sys = 0x1;
  if (plan->mode>0) sys |= 0x02;
  sys |= getFmts(plan->fmts);
  HW_write(0x00,0x0);
  HW_write(0x1C,plan->flags);
  // prep header address
  HW_write(0x20,0xFFF); // clear count
  for (i=0; i<16; i++) HW_write(0x28,i);
  // header field registers
  HW_write(0x20,0xFFF); // clear count
  for (i=0; i<8; i++) HW_write(0x24,plan->pkthdr[i]);
  // extra fields for counter math
  HW_write(0x24,(int_4)(dpsec&0xFFFFFFFF));
  HW_write(0x24,(int_4)(dpsec>>32));
  HW_write(0x24,(int_4)(dpdif&0xFFFFFFFF));
  HW_write(0x24,(int_4)(dpdif>>32));
  // sequence rams
  s = (bits==32)? 0x0040 : (bits==16)? 0x0080 : 0x0000;
  if (plan->mode>0) {
    for (i=0; i<5; i++) seq[j++] = 0x1040|i; 
    seq[j++] = 0x1040|6; // BE fractional TS swap
    seq[j++] = 0x1040|5;
    for (i=0; i<n; i++,j++) {
      if (bits==ibps)    { k=i%2; seq[j] = 0x1010|s; seq[j-1] |= 0x0100; }
      else if (bits==12) { k=i%3; seq[j] = 0x1060|k; seq[j-1] |= ((k<2)?0x0200:0); }
      else if (bits==8)  { k=i%1; seq[j] = 0x1063;   seq[j-1] |= 0x0200; }
      else if (bits==6)  { k=i%3; seq[j] = 0x1064|k; seq[j-1] |= ((k<2)?0x0200:0); }
      else if (bits==4)  { k=i%1; seq[j] = 0x1067;   seq[j-1] |= 0x0200; }
    }
    if (tlr) for (i=7; i<8; i++) seq[j++] = 0x1040|i;
  } else {
    for (i=0; i<7; i++) seq[j++] = 0x0100;
    seq[j++] = 0x0000;
    for (i=0; i<n; i++,j++) {
      if (bits==obps)    { k=i%2; seq[j] = 0x1010|s; seq[j-1] |= 0x0100; }
      else if (bits==12) { k=i%3; seq[j] = 0x00F0|k|((k>0)?0x2000:0); seq[j-1] |= 0x0100; }
      else if (bits==8)  { k=i%1; seq[j] = 0x20F3;   seq[j-1] |= 0x0100; }
      else if (bits==6)  { k=i%3; seq[j] = 0x00F4|k|((k>0)?0x2000:0); seq[j-1] |= 0x0100; }
      else if (bits==4)  { k=i%1; seq[j] = 0x20F7;   seq[j-1] |= 0x0100; }
    }
    if (tlr) for (i=7; i<8; i++) seq[j++] = 0x0100;
  }
  // overlay the header manipulation sequencer
  if (plan->mode>0) {
    i=8; 				// get past header
    seq[i++] |= (0x4800<<16);		// set Enable
    i=nibinc(seq,i,0x04,1); 		// sequence count (4 bits)
    i=nibadd(seq,i,0x28,0x40,10);	// fractional seconds increment (40 bits)
    seq[i++] |= (0x4000<<16);		// unset Enable
    i=nibadd(seq,i,0x28,0x50,10);	// fractional seconds overflow test (40 bits)
    seq[i++] |= (0x8000<<16);		// set Enable if carry
    i=nibadd(seq,i,0x28,0x50,10);	// fractional seconds overflow (40 bits)
    i=nibinc(seq,i,0x20,8);		// integer seconds (32 bits)
  } else {

  }
  if (i>j) printf("Packet size=%d too small. Need=%d to support TC increment.\n",j,i);

  // download sequencer
  HW_write(0x20,0xFFF); // clear count and set clast
  for (i=0; i<j; i++) HW_write(0x28,seq[i]);
//  for (i=0; i<j; i++) printf("i=%04x Seq=%08x\n",i,seq[i]);
  // enable 
  HW_write(0x20,j-2); // clear count and set clast-2
  HW_write(0x00,sys);
#endif
  return 0;
}

int_4 VRT_work (VRTPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#if CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 VRT_free (VRTPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

