/****************************************************************/
/*  ICE-PIC Device Driver for LINUX  2.4 kernels new PCI        */
/*                                                              */
/*  Author:  Jeff Schoen - Innovative Computer Engineering      */
/*                                                              */
/****************************************************************/
/*
   Bill Reece Mods:  
   membase must be unsigned so it's not sign extended when it's cast
    to an unsigned long for the call to ioremap().
   vfree to iounmap in detach
*/

#define ICE_DRIVER_VERSION 380

#define MAX_TEMP 70

#include "picdrv.h"

/*****************/
/* Include Files */
/*****************/
#include <linux/version.h> 
#include <linux/module.h> 
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/isdnif.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <asm/io.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
 #include <linux/ioctl.h>
#else
 #ifdef CONFIG_COMPAT
  #include <linux/ioctl32.h>
 #endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
#else
  #define PDE_DATA(A) PDE(A)->data
#endif

/**********************/
/* Module Definitions */
/**********************/
/* RedHat has internal inconsistencies for these 3 Macros comment out for now */
MODULE_AUTHOR("Jeff Schoen");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Driver for DSP Cards from Innovative Computer Engineering");

/****************************/
/* Debug/Status Definitions */
/****************************/
#define ESUCCESS 0
#define	SUCCESS(a)	(((a) & 1) == 1)
#define	FAILURE(a)	(((a) & 1) == 0)

#define vprintf printk
#define wprintf if (0!=0) printk

/***********************************/
/* Device Driver Macro Definitions */
/***********************************/
#define _RD(A,B)  {B = readl(ucb->loc+A);}
#define _WR(A,B)  {writel(B,(int *)(ucb->loc+A));}

/************************************/
/* Linux Special Access Enforcement */
/************************************/
#define is_rd() ((filp->f_flags&0x3)==0)
#define is_wr() ((filp->f_flags&0x3)==1)
#define is_su() ((filp->f_flags&0x3)==2)
#define vm_rd() ((vma->vm_flags&0xFF)==0xfa)
#define vm_wr() ((vma->vm_flags&0xFF)==0xd1)
#define vm_su() ((vma->vm_flags&0xFF)==0xfb)

/*********************************/
/* Generate info for /proc entry */
/********************************/
static int proc_icepic (char *buf, char **a, off_t b, int c);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
#include <linux/seq_file.h>
static int icepic_info_seq_show (struct seq_file *seq, void *offset) {
  char p[1020];
  proc_icepic(p,0,0,0);
  seq_printf(seq,"%s",p);
  return 0;
}
static int icepic_info_open_fs (struct inode *inode, struct file *file) {
  return single_open (file,icepic_info_seq_show,PDE_DATA(inode));
}
static const struct file_operations icepic_info_fops = {
  .owner = THIS_MODULE,
  .open = icepic_info_open_fs,
  .read = seq_read,
  .llseek = seq_lseek,
  .release = single_release,
};
void create_proc_entry_withInfo (const char *name, mode_t mode, 
				 struct proc_dir_entry *base, void *get_info) {
  struct proc_dir_entry *res = proc_create_data (name,S_IRUGO,base,&icepic_info_fops,NULL);
  if (!res) vprintf("icepic unable to register /proc entry\n");
}
#else
void create_proc_entry_withInfo (const char *name, mode_t mode, 
				 struct proc_dir_entry *base, get_info_t *get_info) {
  struct proc_dir_entry *res = create_proc_entry (name,mode,base);
  if (res) res->get_info = get_info;
}
#endif
/*
 Ioctl support for 32 bit apps in 64 bit Linux kernels is implemented either
 with an entry in the 'fops' table (if supported) or with registering a
 function directly with the kernel. This next macro, if defined, implies
 that support exists in the 'fops' table. It probably should be set as part
 of 'picdrv.h' definition. 
*/

#if defined(CONFIG_COMPAT) && !defined(HAVE_COMPAT_IOCTL)
#define ICEIOCCMD32   (_IOWR('f',0x1, PICIOCTL32))
#endif

/*******************************/
/* Driver Specific Definitions */
/*******************************/

typedef int64_t int_8;
typedef uint64_t int_u8;
typedef int32_t int_4;
typedef uint32_t int_u4;

#include "../../inc/iceioctl.h"
#include "../../inc/iceflex.h"
#include "../../inc/iceconv.h"

#define B1G   0x0040000000LL	// 1G
#define B4G   0x0100000000LL	// 4G
#define B16G  0x0400000000LL	// 16G
#define B64G  0x1000000000LL	// 64G
#define B256G 0x4000000000LL	// 256G

#define MAXPIC 12
#define MAXIOBUF 512
#define MAXWCNT B256G	

#define VENDOR_ID 0x1172
#define DEVICE_ID 0x7777
#define DRV_NAME "icepic"

typedef struct {
  int_u8	paddr;
  int_u8	psize;
  int		device;
  pid_t 	pid,gid;
} DMACHANNEL;

typedef struct {
  int	attached;
  int	opened;
  unsigned int membase;
  int   lastoffset;
  void	*loc;
  int   csr,lat,rev,rid; 
  struct pci_dev *pdev;
  int time, temp, temp1, temp2, pktime, pktemp;
} UCB;

static UCB ucbs[MAXPIC];
static DMACHANNEL dma[MAXIOBUF];
static unsigned int icepic_major=0;
static unsigned int icepic_minor=0;
static int ndma=0,  icepic_count=0;
static int_u8 offset_min=0, offset_max=0;
static int_u8 rstart, rsized, rsizem;
static int cstat;
static struct semaphore icepic_mutex;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
static DEFINE_SEMAPHORE (icepic_mutex);
#else
static DECLARE_MUTEX (icepic_mutex);
#endif

#define ICEPIC_LOCK down(&icepic_mutex)
#define ICEPIC_UNLOCK up(&icepic_mutex)

#if _NEW
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
#define mygid task_pgrp_nr(current) 
#else
#define mygid process_group(current) 
#endif
#define mypid current->pid
#else
#define mygid current->pgrp
#define mypid current->pid
#endif

/*************************************/
/* Internal Function Prototypes      */
/*************************************/
static int icepic_attach (struct pci_dev *pdev, const struct pci_device_id *ent);
static void icepic_detach (struct pci_dev *pdev);

/************************************/
/* Prototypes for External Routines */
/************************************/
#ifdef __cplusplus
#define EXTERN_CDEFINE extern "C"
#else
#define EXTERN_CDEFINE extern
#endif

EXTERN_CDEFINE int icepic_open (struct inode *inode, struct file *filp);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
EXTERN_CDEFINE long icepic_ioctl (struct file *filp, unsigned int cmd, unsigned long arg);
#else
EXTERN_CDEFINE int icepic_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);
#endif
EXTERN_CDEFINE int icepic_close (struct inode *inode, struct file *filp);
EXTERN_CDEFINE int icepic_mmap (struct file *filp, struct vm_area_struct *vma);
EXTERN_CDEFINE int icepic_cancel (struct inode *inode, struct file *filp, int reason);
EXTERN_CDEFINE int icepic_mapio (UCB *ucb, int function, int *channel, int_u8 psize);
EXTERN_CDEFINE int icepic_reattach (int index);
EXTERN_CDEFINE void icepic_setram (int_4 start, int_4 sized, int_4 sizem, int_4 verbose);
EXTERN_CDEFINE void reverse (void* pBuf, int pLen);
EXTERN_CDEFINE long icepic_compat_ioctl (struct file *filp, unsigned int cmd, unsigned long arg);
EXTERN_CDEFINE int icepic_ioctl32 (unsigned int fd, unsigned int cmd, unsigned long arg, struct file *filp);


struct file_operations icepic_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
  unlocked_ioctl: icepic_ioctl,
#else
  ioctl:	icepic_ioctl, 
#endif
#if defined(CONFIG_COMPAT) && defined(HAVE_COMPAT_IOCTL)
  compat_ioctl:	icepic_compat_ioctl, 
#endif
  open:		icepic_open, 
  release:	icepic_close,
  mmap:		icepic_mmap
};

static struct pci_device_id icepic_pci_tbl[] = {
  {
    vendor:VENDOR_ID,
    device:DEVICE_ID,
    subvendor:PCI_ANY_ID,
    subdevice:PCI_ANY_ID,
    class: 0, class_mask:0,
  },
  {}
};
MODULE_DEVICE_TABLE(pci, icepic_pci_tbl);

static struct pci_driver icepic_driver = {
  name:       DRV_NAME,
  id_table:   icepic_pci_tbl,
  probe:      icepic_attach,
  remove:     icepic_detach,
  suspend:    NULL,
  resume:     NULL
};

/*******************************/
/* Module Parameters           */
/*******************************/
static int ice_ra = 0;
module_param(ice_ra, int, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(ice_ra, "DMA RAM allocation method (0=Manual,1=Kernel,2=PreKernel)");

static int ice_rs = RAM_START;
module_param(ice_rs, int, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(ice_rs, "DMA start address in MB (1024*1024 B)");

static int ice_rd = RAM_SIZED;
module_param(ice_rd, int, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(ice_rd, "DMA ramdisk size in MB (1024*1024 B)");

static int ice_rm = RAM_SIZEM;
module_param(ice_rm, int, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(ice_rm, "DMA mapable RAM size in MB (1024*1024 B)");

/************************************/
/* picdrv kernel memory routines    */
/************************************/
#ifdef USE_KNR_MEM
#include "picknr.c"
#endif

/************************************/
/* picdrv init/exit routine s       */
/************************************/
static int __init icepic_init (void) {
  int i,j; for (i=0; i<MAXPIC; i++) ucbs[i].attached=0;
  icepic_major = register_chrdev (0, DRV_NAME, &icepic_fops);
  icepic_count = pci_register_driver (&icepic_driver); /* probe and init */
  for (i=j=0; i<MAXPIC; i++) { if (ucbs[i].attached > 0) j++; }
  if (j>icepic_count) icepic_count=j; /* pci_register_driver count is unreliable */
  vprintf("icepic init major=%d count=%d\n",icepic_major,icepic_count);
#if defined(CONFIG_COMPAT) && !defined(HAVE_COMPAT_IOCTL)
  if (register_ioctl32_conversion(ICEIOCCMD32, icepic_ioctl32) != 0)
    vprintf("icepic unable to register ioctl32_conversion entry\n");
#endif
  if (ice_ra>0) {
#ifdef USE_KNR_MEM
    dma_parm_init();
    dma_pgent_alloc();
    icepic_setram (pdp.ramstart,pdp.ramdisk,pdp.rammap,1);
#else
    vprintf("ICE RamAlloc mode>0 not supported with this kernel\n");
#endif
  } else {
    icepic_setram (ice_rs,ice_rd,ice_rm,1); 
  }
  create_proc_entry_withInfo ("driver/icepic", 0, NULL, proc_icepic);
  return 0; /* keep loaded for hot plug even if no devices found */
}

static void __exit icepic_cleanup (void) {
  vprintf("icepic exit major=%d count=%d\n",icepic_major,icepic_count);
  remove_proc_entry("driver/icepic",NULL);
#ifdef USE_KNR_MEM
  if (ice_ra>0) dma_pgent_free((u64_t)0,(s32_t)1);
#endif
#if defined(CONFIG_COMPAT) && !defined(HAVE_COMPAT_IOCTL)
  unregister_ioctl32_conversion(ICEIOCCMD32);
#endif
  unregister_chrdev (icepic_major, DRV_NAME);
  pci_unregister_driver (&icepic_driver);
  icepic_major=icepic_minor=icepic_count=0;
}

module_init(icepic_init);
module_exit(icepic_cleanup);

/************************************/
/* pic attach/detach routines       */
/************************************/
static int icepic_attach (struct pci_dev *pdev, const struct pci_device_id *ent)
{
  int status, index=icepic_minor++;
  UCB *ucb = &ucbs[index];
  ucb->attached = 0;
  pci_set_master(pdev);
  status = pci_read_config_dword(pdev,0x10,&ucb->membase); ucb->membase &= 0xFFFFFFF0;
  status = pci_read_config_dword(pdev,0x04,&ucb->csr);
  status = pci_read_config_dword(pdev,0x08,&ucb->rev);
  status = pci_read_config_dword(pdev,0x0C,&ucb->lat);
  vprintf ("icepic attach - unit %d base %x csr %x lat %d\n",
	index,ucb->membase,ucb->csr,(ucb->lat>>8)&0xFF);
  ucb->loc = (void *)ioremap( (unsigned long)ucb->membase, 4096);
  ucb->pdev = pdev;
  ucb->attached = 1;
  ucb->lastoffset = 0;
  ucb->time = ucb->pktime = 0;
  ucb->temp = ucb->temp1 = ucb->temp2 = ucb->pktemp = -2;
  status = pci_enable_device(pdev);
  return 0; 
}

static void icepic_detach (struct pci_dev *pdev)
{
  UCB *ucb; int index; 
  for (index=0; index<MAXPIC && ucbs[index].pdev!=pdev; index++);
  if (index==MAXPIC) return;
  icepic_minor--;
  ucb = &ucbs[index];
  if (ucb->attached==0) return;
  vprintf ("icepic detach - unit %d baseaddr %x\n",index,ucb->membase);
  iounmap (ucb->loc);
  ucb->attached = 0;
  return;  
}

/************************************/
/* pic open routine                 */
/************************************/
int icepic_open (struct inode *inode, struct file *filp)
{
  UCB *ucb; int index = MINOR(inode->i_rdev); 
  wprintf ("pic_open - dev %#x fp.flags %#x\n",index,filp->f_flags);
  ucb = &ucbs[index];
  if (ucb->attached == 0) {
    wprintf("pic_open - device unit not attached\n"); /* normal detection case - dont spam kernel log */
    return(-ENXIO);
  }
  if ((filp->f_flags&O_NONBLOCK)==0) /* not mmap */
  ucb->opened++;
  return(ESUCCESS); 
}
 


/************************************/
/* pic close routine                */
/************************************/
int icepic_close (struct inode *inode, struct file *filp)
{
  int index = MINOR(inode->i_rdev);
  UCB *ucb = &ucbs[index];;
  wprintf ("pic_close - dev %#x fp.flags %#x\n",index,filp->f_flags);
  if (ucb->attached == 0) {
    vprintf("pic_close - device unit not attached\n");
  }
  /* make sure all DMA's from this PID have been terminated */
  icepic_cancel (inode, filp, ucb->opened);
  if ((filp->f_flags&O_NONBLOCK)==0) /* not mmap */
  ucb->opened--;
  return 0;
}
    
/************************************/
/* pic ioctl routine                */
/************************************/
#define PIC_FUNCTION	p->function	/* user function code */
#define PIC_OFFSET	p->offset	/* register offset */
#define PIC_BUFADDR	p->bufaddr	/* user buffer addr */
#define PIC_BYTES	p->bytes	/* bytes to transfer */
#define PIC_STATUS	p->status	/* return status */

int icepic_ioctl_work (int index, struct file *filp, PICIOCTL *p)
{
  UCB 		*ucb  = &ucbs[index];

  int		i,j,k;		/* Local counters */
  int		status;		/* Local function return status */
  volatile int	stat;		/* bus read return status */
  int		chan;		/* Current dma channel */
  volatile int	wtmp[6];	/* kernel space temp data */
  int		worda,wordb;	/* load IOC temps */
  int_u8	msize,maddr;	/* word temps */

  if (PIC_FUNCTION==IOCTL_MAP||PIC_FUNCTION==IOCTL_UNMAP) ICEPIC_LOCK;

  i = PIC_BYTES;			/* normal|default completion status */
  status = ESUCCESS; 			/* assume success */

  switch (PIC_FUNCTION) {

  case IOCTL_READ:			/* read a device register */

    if (PIC_OFFSET == REG_FIFO) {
      for (i=0; i<PIC_BYTES; ) {
        _RD (REG_MCSR,stat);
             if ((stat&MCSR_IFIFO_FULL) && (PIC_BYTES-i>=24)) j=6;
        else if ((stat&MCSR_IFIFO_AFULL) && (PIC_BYTES-i>=16)) j=4;
        else if ((stat&MCSR_IFIFO_HALF) && (PIC_BYTES-i>=8)) j=2;
        else if (!(stat&MCSR_IFIFO_EMPTY)) j=1;
        else break;
        for (k=0; k<j; k++) _RD(REG_FIFO, wtmp[k]);  
	cstat=copy_to_user(((void *)PIC_BUFADDR)+i, (void *)wtmp, j*4);
	i += (j*4);
      }
    }
    else if ( (PIC_OFFSET&0XFFFFF000) != 0) 	/* invalid register ? */
      status = EINVAL;
    else {
      _RD (PIC_OFFSET, wtmp[0]);		
      cstat=copy_to_user((void *)PIC_BUFADDR, (void *)wtmp, 4);
    }
    break;

  case IOCTL_WRITE:			/* write a device register */

    if (PIC_OFFSET == REG_FIFO) {
      for (i=0; i<PIC_BYTES; ) {
        _RD (REG_MCSR,stat);
             if ((stat&MCSR_OFIFO_EMPTY) && (PIC_BYTES-i>=24)) j=6;
        else if (!(stat&MCSR_OFIFO_HALF) && (PIC_BYTES-i>=16)) j=4;
        else if (!(stat&MCSR_OFIFO_AFULL) && (PIC_BYTES-i>=8)) j=2;
        else if (!(stat&MCSR_OFIFO_FULL)) j=1;
        else break;
	cstat=copy_from_user((void *)wtmp, ((void *)PIC_BUFADDR)+i, j*4);
        for (k=0; k<j; k++) _WR(REG_FIFO, wtmp[k]);  
	i += (j*4);
      }
    }
    else if ( (PIC_OFFSET&0XFFFFF000) != 0) 	/* invalid register ? */
      status = EINVAL;
    else {
      cstat=copy_from_user((void *)wtmp, (void *)PIC_BUFADDR, 4);
      if (PIC_OFFSET==ucb->lastoffset+4) _WR(REG_PBAR+8,0);
      _WR(PIC_OFFSET,wtmp[0]);
      ucb->lastoffset=PIC_OFFSET;
    }
    break;

  case IOCTL_ACMD: 			/* command to SHARC processor */

    cstat=copy_from_user((void *)wtmp, (void *)PIC_BUFADDR, 20);
    for (k=4; k>=0; k--) _WR(REG_MBX+k*4,wtmp[k]);
    for (i=0,j=0; j<32; j++) {  /* cant wait forever in kernel mode */
      _RD (REG_MCSR,stat);
      if ( (stat&MCSR_IMB1F) != 0x0 ) {
        _RD (REG_IMB1, wtmp[0]);		
        cstat=copy_to_user((void *)PIC_BUFADDR, (void *)wtmp, 4);
	i=4; break;
      }
      if (j>4) udelay(100);
    }
    break;

  case IOCTL_MAP:			/* map user buffer to PCI */

    i=0;	/* assume failure */
    /* find an open dma handler slot */
    if (ndma==0 || ndma+1>=MAXIOBUF) status=EINVAL; 

    msize = phys2long(PIC_BYTES);

    /* Validate user buffer */
    if ( (msize == 0)  || (msize > MAXWCNT) ) status = EINVAL;
    if (status!=ESUCCESS) break;

    /* map buffer into PCI bus space */
    status = icepic_mapio (ucb, IOCTL_MAP, &chan, msize);
    if (status==ESUCCESS) i=long2phys(dma[chan].paddr); 
    if (status==ESUCCESS) dma[chan].device=index; 
    break;
    
  case IOCTL_UNMAP: 			/* unmap user buffer from PCI */

    msize = phys2long(PIC_BYTES);
    maddr = phys2long(PIC_OFFSET);

    /* find the matching dma handler slot */
    for (chan=1; chan<ndma; chan++) {
      if (maddr >= dma[chan].paddr && 
	  maddr+msize <= dma[chan+1].paddr) break;
    }
    if (chan>=ndma) { status = EINVAL; break; }

    /* unmap user buffer from PCI bus space */
    status = icepic_mapio (ucb, IOCTL_UNMAP, &chan, 0);
    break;

  case IOCTL_QMAP: 			/* query user buffer map */

    msize = phys2long(PIC_BYTES);
    maddr = phys2long(PIC_OFFSET);

    /* is it in the table of driver mapped PCI memory */
    for (chan=0; chan<ndma; chan++) {
      if (maddr >= dma[chan].paddr && 
	  maddr+msize <= dma[chan].paddr+dma[chan].psize) break;
    }
    if (chan==ndma) i=0;
    else if (PIC_BYTES!=0) i=PIC_BYTES;
    else i=long2phys(dma[chan].paddr+dma[chan].psize-maddr);  
//vprintf("icepic qmap index=%d s=%x o=%x chan=%d ndma=%d i=%d %llx %llx %llx\n",index,PIC_BYTES,PIC_OFFSET,chan,ndma,i,dma[0].paddr,dma[0].psize,maddr);
    break;

  case IOCTL_RALLOC: 			/* allocate contiguous RAM */

    icepic_setram ( (int_4)PIC_OFFSET, (int_4)(long)PIC_BUFADDR, (int_4)PIC_BYTES, 1);
    break;

  case IOCTL_QALLOC: 			/* report allocated RAM */

    i = long2phys(rstart);
    break;

  case IOCTL_LIOC:			/* load IOC words */

    for (i=0; i<PIC_BYTES; i+=4) {
      k = PIC_BYTES-i; if (k>4) k=4;
      cstat=copy_from_user((void *)(&worda), ((void *)PIC_BUFADDR)+i, k);
      reverse(&worda,4);
      for (j=0; j<(k*8); j++) {
        wordb = (worda>>j)&0x1;
        _WR(REG_IOC,wordb);
      }
    }
    break;

  case IOCTL_REATTACH:			/* reattach test device */

    i = icepic_reattach (index);
    break;

  case IOCTL_VERSION:			/* return version */

    wtmp[0] = ICE_DRIVER_VERSION;
    cstat=copy_to_user((void *)PIC_BUFADDR, (void *)wtmp, 4);
    break;

  case IOCTL_LOG:			/* log status */

    cstat=copy_from_user((void *)wtmp, (void *)PIC_BUFADDR, PIC_BYTES);
    ucb->time = wtmp[0];
    ucb->temp = wtmp[1];
    ucb->temp1 = (PIC_BYTES>8)? wtmp[2]:-1;
    ucb->temp2 = (PIC_BYTES>12)? wtmp[3]:-1;
    k = ucb->temp; 
    if (ucb->temp1 > k) k = ucb->temp1;
    if (ucb->temp2 > k) k = ucb->temp2;
    if (k > ucb->pktemp) { ucb->pktemp = k;  ucb->pktime = ucb->time; }
    if (k>MAX_TEMP) printk("ICEPIC Overtemp pic=%d pm1=%d pm2=%d\n",ucb->temp,ucb->temp1,ucb->temp2);
    break;

  case IOCTL_QLOG:			/* log status */

    wtmp[0] = ucb->time;
    wtmp[1] = ucb->temp;
    wtmp[2] = ucb->pktime;
    wtmp[3] = ucb->pktemp;
    cstat=copy_to_user((void *)PIC_BUFADDR, (void *)wtmp, 16);
    break;

  default: 				/* handle unknown function codes */
    status = EINVAL;
    break;
  }

  if (PIC_FUNCTION==IOCTL_MAP||PIC_FUNCTION==IOCTL_UNMAP) ICEPIC_UNLOCK;

  p->status = i;
  return(status);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
long icepic_ioctl (struct file *filp, unsigned int cmd, unsigned long arg)
{
  long		status;
  int		index = MINOR(filp->f_dentry->d_inode->i_rdev);
#else
int icepic_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
  int		status;
  int 		index = MINOR(inode->i_rdev);
#endif
  int		psize = _IOC_SIZE(cmd);
  PICIOCTL	ploc;	/* pic ioctl handle structure */
  wprintf("pic_ioctl - dev %#x, cmd %#x, arg %#lx\n", index,cmd,arg);
  cstat=copy_from_user ((void *)(&ploc), (void *)arg, psize);
  status = icepic_ioctl_work (index,filp,&ploc);
  cstat=copy_to_user((void *)&((PICIOCTL *)arg)->status, (void *)&ploc.status, 4);
  return status;
}

long icepic_compat_ioctl (struct file *filp, unsigned int cmd, unsigned long arg)
{
  int 		index = MINOR(filp->f_dentry->d_inode->i_rdev);
  int		psize = _IOC_SIZE(cmd);
  int		status;
  PICIOCTL32	ploc32;	/* pic ioctl handle structure */
  PICIOCTL	ploc;	/* pic ioctl handle structure */
  wprintf("pic_compat_ioctl - dev %#x, cmd %#x, arg %#lx\n", index,cmd,arg);
  cstat=copy_from_user ((void *)(&ploc32), (void *)arg, psize);
  ploc.function = ploc32.function;
  ploc.offset   = ploc32.offset;
  ploc.bufaddr  = (void *)(long)ploc32.bufaddr;
  ploc.bytes    = ploc32.bytes;
  ploc.status   = ploc32.status;
  status = icepic_ioctl_work (index,filp,&ploc);
  cstat=copy_to_user((void *)&((PICIOCTL32 *)arg)->status, (void *)&ploc.status, 4);
  return (long)status;
}

int icepic_ioctl32 (unsigned int fd, unsigned int cmd, unsigned long arg, struct file *filp)
{
  return (int)icepic_compat_ioctl (filp, cmd, arg);
}

void icepic_setram (int_4 start, int_4 sized, int_4 sizem, int_4 verbose) {
  if (start!=-1) rstart = (int_u8)start * 0x100000; 	/* RAM start in bytes */
  if (sized!=-1) rsized = (int_u8)sized * 0x100000; 	/* RAMDISK size in bytes */
  if (sizem!=-1) rsizem = (int_u8)sizem * 0x100000; 	/* DYNRAM size in bytes */

  dma[0].paddr=rstart;
  dma[0].psize=rsized;
  dma[0].pid=0;
  dma[1].paddr=rstart+rsized+rsizem;
  dma[1].psize=0;
  dma[1].pid=0;
  ndma=1;
  offset_min = rstart;
  offset_max = rstart+rsized+rsizem;
  if (verbose) printk("PIC SetRam start=%d %llx sized=%d %llx sizem=%d %llx  off_min=%llx off_max=%llx\n",
	start,rstart,sized,rsized,sizem,rsizem,offset_min,offset_max);
}


/************************************/
/* pic cancel routine               */
/************************************/
int icepic_cancel (struct inode *inode, struct file *filp, int reason)
{
  int reset,status,chan;
  int index = MINOR(inode->i_rdev);
  UCB *ucb = &ucbs[index];
  reset = MCSR_RESET;

  wprintf ("pic_cancel - dev %#x reason %x flags %x\n",index,reason,filp->f_flags);

  /* the mmap opens from icelib.c will have the O_NONBLOCK flag */
  /* these closes should not perform cleanup tasks */
  if (reason==1 && (filp->f_flags&O_NONBLOCK)==0) {	/* close */
    ICEPIC_LOCK;
    for (chan=ndma-1; chan>=1; chan--) {
      if (dma[chan].device == index
       && dma[chan].gid == mygid 
       && dma[chan].paddr != 0) {
        vprintf ("pic_cancel - stray DMA - resetting card\n");
        /* unexpected termination - reset card to avoid stray DMAs */
	_WR (REG_MCSR, reset);
	/* deallocate/unmap stray buffer */
	status = icepic_mapio (ucb, IOCTL_UNMAP, &chan, 0L);
      }
    }
    ICEPIC_UNLOCK;
  }
  return(ESUCCESS); 
}



/************************************/
/* pic mapio routine                */
/************************************/
int icepic_mapio (UCB *ucb, int function, int *channel, int_u8 psize)
{
  int i,n,status,chan;

  status = ESUCCESS;	/* assume success */

  switch (function) {

  case IOCTL_MAP:

    for (i=1; i<=ndma; i++) if (dma[i].paddr-
	(dma[i-1].paddr+dma[i-1].psize) >= psize) goto ALLOCATE;

    return ENOMEM; /* out of room */

    ALLOCATE:
    n=i; ndma++;
    for (i=ndma; i>n; i--) dma[i] = dma[i-1];
    dma[n].paddr = dma[n-1].paddr+dma[n-1].psize;
    dma[n].psize = psize;
    dma[n].pid = mypid;
    dma[n].gid = mygid;
    wprintf("DMA map psize=%llx n=%d paddr=%llx ndma=%d\n",psize,n,dma[n].paddr,ndma);
    *channel = n;
    break;

  case IOCTL_UNMAP:

    chan = *channel;
    if (chan>=0) psize = dma[chan].psize;
    wprintf("DMA unmap psize=%llx %d\n",psize,ndma);
    for (i=chan; i<ndma; i++) dma[i] = dma[i+1];
    ndma--;
    break;

  default:
    status = EINVAL;
  }
#ifdef DEBUG_MAPIO
  for (i=0; i<=ndma; i++) wprintf("DMA table i=%d paddr=%llx psize=%llx\n",i,dma[i].paddr,dma[i].psize);
#endif
  return (status);
}

/************************************/
/* pic reattach routine             */
/************************************/
int icepic_reattach (int index)
{
  UCB *ucb = &ucbs[index];
  int status;
  ucb->attached = 0;
  if (!ucb->pdev) return 0; 
  status = pci_write_config_dword(ucb->pdev,0x10,ucb->membase);
  status = pci_write_config_dword(ucb->pdev,0x04,ucb->csr);
  status = pci_write_config_dword(ucb->pdev,0x0C,ucb->lat);
  vprintf ("pic_reattach - unit %d base %x csr %x\n",index,ucb->membase,ucb->csr);
  ucb->attached = 1;
  return (1); 
}

void reverse (void* pBuf, int pLen) {
#if defined(BYTE_ORDER) && BYTE_ORDER == BIG_ENDIAN
  int i,j;
  char temp, *tPtr;
  tPtr = (char *)pBuf;
  j=pLen; pLen >>= 1;
  for (i=0; i<pLen; i++) {
    temp = tPtr[--j]; tPtr[j]=tPtr[i]; tPtr[i]=temp;
  }
#endif
}

/************************************/
/* pic mmap routine                 */
/************************************/
int icepic_mmap (struct file *filp, struct vm_area_struct *vma)
{
  int_u8 offset = ((int_u8)vma->vm_pgoff) << PAGE_SHIFT;
  int_u8 length = vma->vm_end - vma->vm_start;

  wprintf("PIC mmap off=%llx end=%llx off_min=%llx, off_max=%llx flags=%x prot=%x\n",
	offset,offset+length,offset_min,offset_max,filp->f_flags,(int_u4)vma->vm_flags);
  if (offset<offset_min || offset+length>offset_max) return -EAGAIN;

  /* special mode for kernel mapped memory */
  if (ice_ra>0) {
#ifdef USE_KNR_MEM
   return icemapknr (vma, 1);
#else
   return -EAGAIN;
#endif
  }

  vma->vm_flags |= VM_IO;
#if _VMAMODE == 0
  if (remap_page_range(vma->vm_start, (vma->vm_pgoff<<PAGE_SHIFT), 
	vma->vm_end-vma->vm_start, vma->vm_page_prot)) return -EAGAIN;
#elif _VMAMODE == 2
  if (remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff, 
	vma->vm_end-vma->vm_start, vma->vm_page_prot)) return -EAGAIN;
#else
  if (remap_page_range(vma, vma->vm_start, (vma->vm_pgoff<<PAGE_SHIFT), 
	vma->vm_end-vma->vm_start, vma->vm_page_prot)) return -EAGAIN;
#endif
  return 0;
}

static int proc_icepic (char *buf, char **a, off_t b, int c)
{
    char*   p = buf;
    int     i,tc,t0,t1,t2;

    p += sprintf (p, "RamStart  : %lld Mby\n", rstart>>20);
    p += sprintf (p, "RamDisk   : %lld Mby\n", rsized>>20);
    p += sprintf (p, "RamMapped : %lld Mby\n", rsizem>>20);
    p += sprintf (p, "RamPage   : %d Kby\n", (int)(PAGE_SIZE>>10));
    p += sprintf (p, "RamAlloc  : %d \n", ice_ra);
    p += sprintf (p, "Ram Addr  : Start=0x%llx End=0x%llx\n", offset_min,offset_max);
    p += sprintf (p, "Buffers   : %d\n",ndma);
    for (i=0; i < ndma; i++) {
      p += sprintf (p, "Buffer #%d : Addr=0x%llx Size=0x%llx \n", i, dma[i].paddr, dma[i].psize);
    }
    p += sprintf (p, "Devices   : %d\n",icepic_count);
    for (i=0; i<MAXPIC; i++) if (ucbs[i].attached > 0) {
      t0 = ucbs[i].temp; t1 = ucbs[i].temp1; t2 = ucbs[i].temp2;
      tc = t0; if (t1>tc) tc=t1; if (t2>tc) tc=t2;
      p += sprintf (p, "Device #%d : Revision=0x%x MemBase=0x%x   Card Temp=%d MJS=%d   Peak Temp=%d MJS=%d   Comp Temp=(%d,%d,%d,%d,%d)\n",
                       i,ucbs[i].rev&0xFF,ucbs[i].membase,tc,ucbs[i].time,ucbs[i].pktemp,ucbs[i].pktime,t0,t1,t2,-1,-1);
    }
    p += sprintf (p, "Version   : %d\n", ICE_DRIVER_VERSION);
    return  (p - buf);
}

