/**
  Common defines for processing COREs

  CORE_AREA is the java native package location
  CORE_NAME is the Library Function Name
  CORE_TYPE is the Processor Type Java,CPU,GPU,VHS,MLS,ICE

  @author Jeff Schoen
  @version $Id: CoreDefs.h,v 1.1 2011/04/27 13:09:47 jgs Exp $
*/

#ifndef CORE_IS_ICELIB

#include "icedefs.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define CORE_TYPE_JDK 0
#define CORE_TYPE_CPU 1
#define CORE_TYPE_PHI 2
#define CORE_TYPE_GPU 3
#define CORE_TYPE_MLS 4
#define CORE_TYPE_VHS 5
#define CORE_TYPE_ICE 6

#ifdef CORE_TYPE
#elif INNERCLASS_CPU
#define CORE_TYPE CORE_TYPE_CPU
#elif INNERCLASS_GPU
#define CORE_TYPE CORE_TYPE_GPU
#elif INNERCLASS_MLS
#define CORE_TYPE CORE_TYPE_MLS
#elif INNERCLASS_VHS
#define CORE_TYPE CORE_TYPE_VHS
#elif INNERCLASS_ICE
#define CORE_TYPE CORE_TYPE_ICE
#else
#define CORE_TYPE CORE_TYPE_JDK
#endif

#if CORE_TYPE == CORE_TYPE_JDK
#define CORE_SIG JDK
#define CORE_IS_JDK 1
#elif CORE_TYPE == CORE_TYPE_CPU
#define CORE_SIG CPU
#define CORE_IS_CPU 1
#elif CORE_TYPE == CORE_TYPE_GPU
#define CORE_SIG GPU
#define CORE_IS_GPU defined(_CUDA)
#elif CORE_TYPE == CORE_TYPE_MLS
#define CORE_SIG MLS
#define CORE_IS_MLS 1
#define CORE_IS_FPGA 1
#elif CORE_TYPE == CORE_TYPE_VHS
#define CORE_SIG VHS
#define CORE_IS_VHS 1
#define CORE_IS_FPGA 1
#elif CORE_TYPE == CORE_TYPE_ICE
#define CORE_SIG ICE
#define CORE_IS_ICE 1
#define CORE_IS_FPGA 1
#endif

/* Common Helper Functions */
double log2 (double value) {
  double val;
  val = log(value) * 1.44269504088896;
  return val;
}
double powi (int base, int power) {
  double val;
  if (power==2) val = (double)base*(double)base;
  else val = pow( (double)base, (double)power );
  return val;
}
int_4 pow2ge (int size) {
  int p2; for (p2=0;(1<<p2)<size;p2++);
  return p2;
}

#if CORE_IS_GPU
#include <cuda.h>
#include <cufft.h>
#include <driver_types.h>
#endif

#if CORE_IS_VHS
#include "VCore.h"
#endif

#ifndef CORE_IS_ICELIB
#include "icecores.h"
#endif

#endif

/* CORE Hardware Abstraction Layer Object fields */
typedef struct {

  int_4 (*init)();
  int_4 (*config)();
  int_4 (*setkey)();
  int_4 (*getkey)();
  int_4 (*work)();
  int_4 (*free)();

  int_4 func;
  int_4 fmts;

  int_4 ni1,ni2;
  int_4 no1,no2;
  int_1 *bi1,*bi2;
  int_1 *bo1,*bo2;

  int_4 status;
  int_4 psize;

} HALO;

#define keyMatch(A) keyCheck(key,A)
#define keyMatchIndexed(A) keyCheckIndexed(key,A)

#define FMTS plan->halo.fmts

#define VBUFI(A,B) void *A=plan->halo.bi1; int_4 B=plan->halo.ni1
#define VBUFO(A,B) void *A=plan->halo.bo1; int_4 B=plan->halo.no1
#define BBUFI(A,B) int_1 *A=plan->halo.bi1; int_4 B=plan->halo.ni1
#define BBUFO(A,B) int_1 *A=plan->halo.bo1; int_4 B=plan->halo.no1
#define IBUFI(A,B) int_2 *A=(int_2*)plan->halo.bi1; int_4 B=plan->halo.ni1<<1
#define IBUFO(A,B) int_2 *A=(int_2*)plan->halo.bo1; int_4 B=plan->halo.no1<<1
#define LBUFI(A,B) int_4 *A=(int_4*)plan->halo.bi1; int_4 B=plan->halo.ni1<<2
#define LBUFO(A,B) int_4 *A=(int_4*)plan->halo.bo1; int_4 B=plan->halo.no1<<2
#define FBUFI(A,B) real_4 *A=(real_4*)plan->halo.bi1; int_4 B=plan->halo.ni1<<2
#define FBUFO(A,B) ireal4 *A=(real_4*)plan->halo.bo1; int_4 B=plan->halo.no1<<2

int_4 keyCheck (char *keyinp, char *keycfg) {
  int ki = (keyinp[1]==':')? 2:0;
  int kc = (keycfg[1]==':')? 2:0;
  if (strcmp(keyinp+ki,keycfg+kc)!=0) return 0;
  if (ki==0 || kc==0 || keyinp[0]==keycfg[0]) return 1;
  return 0; 
}

int_4 keyCheckIndexed (char *keyinp, char *keycfg) {
  int ki = (keyinp[1]==':')? 2:0;
  int kc = (keycfg[1]==':')? 2:0;
  int n  = strlen(keycfg)-kc;
  int index = -1;
  if (strncmp(keyinp+ki,keycfg+kc,n)!=0) return index;
  if (keyinp[ki+n]<'0' || keyinp[ki+n]>'9') return index;
  index = atoi(keyinp+ki+n);
  if (ki==0 || kc==0 || keyinp[0]==keycfg[0]) return index;
  printf("KeyType=%c does not match configured Key=%s\n",keyinp[0],keycfg);
  return -1;
}

int_4 getFormatBits (int_4 fmts, int_4 n) {
  int_4 fmt = (fmts>>((n-1)*8)) & 0xFF;
  int_4 bps = 1;
  if ((fmt&0x8)!=0) { bps=2; fmt=fmt&0x7; } // complex
       if (fmt==0) bps *= 16; // int
  else if (fmt==1) bps *= 8;  // byte
  else if (fmt==2) bps *= 4;  // nibble
  else if (fmt==3) bps *= 1;  // bit
  else if (fmt==4) bps *= 32; // long
  else if (fmt==5) bps *= 64; // xlong
  else if (fmt==6) bps *= 32; // float
  else if (fmt==7) bps *= 64; // double
  return bps;
}

int_4 getFormatBytes (int_4 fmts, int_4 n) {
  int_4 byps = getFormatBits(fmts,n)>>3;
  return (byps<=0)? 1:byps;
}

int_4 getFmts (int_4 fmts) {
  int reg = 0;
  reg |= (fmts&0x000F)<<8; // input buffer  to sysreg bit 8
  reg |= (fmts&0x0F00)<<4; // output buffer to sysreg bit 12
  return reg;
}

int_4 getFormat (const char *fmts) {
  char fm=fmts[0], ft=fmts[1];
  int fmt = 0;
       if (ft=='I') fmt = 0x0;
  else if (ft=='B') fmt = 0x1;
  else if (ft=='N') fmt = 0x2;
  else if (ft=='P') fmt = 0x3;
  else if (ft=='L') fmt = 0x4;
  else if (ft=='X') fmt = 0x5;
  else if (ft=='F') fmt = 0x6;
  else if (ft=='D') fmt = 0x7;
  if      (fm=='C') fmt |= 0x8;
  return fmt;
}

int_4 getListEntry (const char *list, const char *entry) {
  int ll = strlen(list);
  int le = strlen(entry);
  int i,j,k=ll-le;
  for (i=0,j=1; i<=k; i++) {
    if (i==0 || list[i-1]==',') {
      if (strncmp(list+i,entry,le)==0 && (i==k || list[i+le]==',')) return j;
    }
    if (list[i]==',') j++;
  }
  printf("Entry=%s not found in list=%s\n",entry,list);
  return 0;
}


