/*
  Common defines for FPGA Core libraries in the ICE hardware environment.

*/

/* Common Core register offsets */
#define COREG_SYS   0x00
#define COREG_DEC   0x04
#define COREG_GAIN  0x08
#define COREG_RATE  0x0C
#define COREG_RATIO 0x10
#define COREG_FRAME 0x14
#define COREG_FREQ  0x18
#define COREG_FLAG  0x1C

#define COREG_COEF  0x20
#define COREG_CFIR  0x28
#define COREG_PFIR  0x30
#define COREG_RFIR  0x38

#define COREG_USER0 0x20
#define COREG_USER1 0x24
#define COREG_USER2 0x28
#define COREG_USER3 0x2C
#define COREG_USER4 0x30
#define COREG_USER5 0x34
#define COREG_USER6 0x38
#define COREG_USER7 0x3C

#define COREG_PLAN 0x100

/* Common Core sysreg bit register */
#define CORE_ENA  0x0001
#define CORE_PLY  0x0002
#define CORE_DEC  0x0004
#define CORE_OPT  0x0008

#define CORE_I1B  0x0300
#define CORE_I4B  0x0200
#define CORE_I8B  0x0100
#define CORE_I16B 0x0000
#define CORE_I32B 0x0400
#define CORE_ICX  0x0800

#define CORE_O1B  0x3000
#define CORE_O4B  0x2000
#define CORE_O8B  0x1000
#define CORE_O16B 0x0000
#define CORE_O32B 0x4000
#define CORE_OCX  0x8000

/* Common helper macros */
#define I2L(A) ((int_4)A<<15)
#define IX2L(A) ((int_4)A<<16)
#define R2L(A) (int_4)rint((A)*(2.147483648E9))
#define R2LC24(A) r2lc((float)(A),24)
#define R2LC20(A) r2lc((float)(A),20)
#define R2LC18(A) r2lc((float)(A),18)

int_4 r2lc (real_4 coef, int_4 bits) {
  int_4 lscl = (bits>=32)? 0x7FFFFFFF : (1<<(bits-1));
  int_4 l = (int_4)rint(coef*lscl);
  if (l>= lscl) l=lscl-1;
  if (l<=-lscl) l=1-lscl;
  l = l<<(32-bits);
  return l;
}

#ifndef CORE_IS_ICELIB
typedef struct {
  int_4 type;                   /* card type */
  int_1 rev;                    /* card revision */
  int_1 ptype;                  /* IO port type */
  int_1 pindex;                 /* IO port index */
  int_1 pmi;                    /* Processor module index */
} PICSTRUCT;
#endif

PICSTRUCT *ph;                  /* handle to minimalist structure */

typedef struct {
  int_4 sys;	/* system register */
  int_4 dec;	/* decimation register */
  int_4 gain;	/* gain register */
  int_4 rate;	/* sample rate register */
  int_4 ratio;	/* resampler ratio register */
  int_4 frame;	/* frame register */
  int_4 freq;	/* frequency register */
  int_4 flag;	/* flags register */
  int_4 user0;	/* user0 register */
  int_4 user1;	/* user1 register */
  int_4 user2;	/* user2 register */
  int_4 user3;	/* user3 register */
  int_4 user4;	/* user4 register */
  int_4 user5;	/* user5 register */
  int_4 user6;	/* user6 register */
  int_4 user7;	/* user7 register */
} HWPARAM;

#ifdef CORE_IS_ICE
/* Code for ICE Hardware mode */
int_4 FPGA_open (PICSTRUCT **p, char *config);
int_4 FPGA_syswr (PICSTRUCT *p, int_4 addr, int_4 value);
int_4 FPGA_sysrd (PICSTRUCT *p, int_4 addr);
int_4 FPGA_process (PICSTRUCT *p, int_4 *in, int_4 nin, int_4 *out, int_4 nout);
int_4 FPGA_close (PICSTRUCT *p);
int_4 FPGA_wpb (PICSTRUCT *p, int_4 node, int_4 addr, int_4 data);
#ifdef CORE_IS_ICELIB
int_4 HW_open  (char *config) { return 0; }
#define HW_read(A)  FPGA_sysrd(ph,A)
#define HW_write(A,B) FPGA_syswr(ph,A,B)
#define HW_wpb(A,B,C) FPGA_syswr(ph,A,B,C)
int_4 HW_close () { return 0; }
#else
#define HW_open(A) FPGA_open(&ph,A)
#define HW_close() FPGA_close(ph)
#define HW_read(A)  FPGA_sysrd(ph,A)
#define HW_write(A,B) FPGA_syswr(ph,A,B)
#define HW_wpb(A,B,C) FPGA_syswr(ph,A,B,C)
#define HW_proc(A,B,C,D) FPGA_process(ph,A,B,C,D)
#endif

#elif CORE_IS_VHS
/* Code for Verilog Hardware Simulation mode */
#include "VCore.h"
#define HW_open(A) VCore_open()
#define HW_close() VCore_close()
#define HW_read(A)  VCore_sysrd(A)
#define HW_write(A,B) VCore_syswr(A,B)
#define HW_wpb(A,B,C) VCore_syswr(B,C)
#define HW_proc(A,B,C,D) VCore_process(A,B,C,D,64)
#define HW_procw(A,B,C,D) VCore_process(A,B,C,D,128)
#define FPGA_wpb(P,A,B,C) VCore_syswr(B,C)

#else
/* Code pass-thru for non-Hardware mode */
int_4 HW_open  (char *config) { return 0; }
int_4 HW_read  (int_4 addr) { return 0; }
int_4 HW_write (int_4 addr, int_4 value) { return 0; }
int_4 HW_proc  (int_4 *in, int_4 nin, int_4 *out, int_4 nout) { return 0; }
int_4 HW_close () { return 0; }
int_4 FPGA_loadTaps () { return 0; }
int_4 FPGA_loadRTaps () { return 0; }
int_4 FPGA_loadFIR () { return 0; }
int_4 FPGA_loadPPF () { return 0; }
#endif

#define HW_loadFIR(A,B,C,D,E) FPGA_loadFIR(ph,A,B,C,D,E)
#define HW_loadPPF(A,B,C,D) FPGA_loadPPF(ph,A,B,C,D)
#define HW_loadRTaps(A,B,C) FPGA_loadRTaps(ph,A,B,C)
#define HW_loadTaps(A,B,C) FPGA_loadTaps(ph,A,B,C)

/* Common FPGA implementation helpers (icelib.c) */
int_4 pow2sizem1 (int_4 size) {
  return (pow2ge(size)<<16)|(size-1);
}
int_4 getFmt (int bits) {
  int_4 fmt = 0;
  if (bits<0) { fmt |= 0x8; bits=-bits; }
  if (bits==8)  fmt |= 0x1;
  if (bits==4)  fmt |= 0x2;
  if (bits==1)  fmt |= 0x3;
  if (bits==32) fmt |= 0x4;
  return fmt;
}
int_4 bits2fmt (int bits) {
  int_4 fmt = 0;
  if (bits<0) { fmt |= 0x8; bits=-bits; }
  if (bits==8)  fmt |= 0x1;
  if (bits==4)  fmt |= 0x2;
  if (bits==1)  fmt |= 0x3;
  if (bits==32) fmt |= 0x4;
  return fmt;
}
int_4 form2fmt (char *form) {
  int_4 fmt = 0;
  if (form[1]=='I') fmt |= 0x0;
  if (form[1]=='B') fmt |= 0x1;
  if (form[1]=='N') fmt |= 0x2;
  if (form[1]=='P') fmt |= 0x3;
  if (form[1]=='L') fmt |= 0x4;
  if (form[0]=='C') fmt |= 0x8;
  return fmt;
}

int_4 scltbl[13] = { 0,0,0,0,0,0, 0x2A, 0x6A, 0xAA, 0x1AA, 0x2AA, 0x6AA, 0xAAA };

int_4 fftscale (int_4 shift, int_4 pow2) {
  int_4 i,j,k,s=shift;
#if CORE_IS_ICE
  int_4 scale = scltbl[pow2];
  for (i=10; i>0 && s>0; i-=2) {
    j = (scale>>i)&0x3;
    k = min(j,s);
    if (k>0) { j-=k; s-=k; scale = (scale&~(0x3<<i)) | (j<<i); }
  }
  printf("shift=%d pow=%d scltbl=%04x scale=%04x\n",shift,pow2,scltbl[pow2],scale);
  return scale;
#else
  return shift;
#endif
}

#ifdef CORE_IS_ICE
#define GETNODE p->pmi
#else
#define GETNODE 0
#endif

#ifdef CORE_IS_FPGA

#define min _min
#define max _max
#define v2print printf

int_4 setup_fpga_filter (PICSTRUCT *p, int_4 port, int_4 pmod, int_4 addrs, int_4 order, int_4 dec, int_4 taps, real_4 *coef)
{
  int_4 i,odd,sym,anti,staps,otap,ttaps,tpc,cps,odec,otpc,aload,bload,acc,aco,seq,htaps,otaps;
  int_4 aadr,ashft,badr,bshft,cadr,cshft,radr,rshft,idec,tcnt,ccnt, addrt;
  int_4 sequence[32],o1,side;

  odd  = (taps%2 == 1);
  sym  = 1;
  anti = 1;

  /* parse taps */
  for (i=0; i<taps/2; i++) {
    if (fabs(coef[i]-coef[taps-1-i])>0.001) sym = 0;
    if (fabs(coef[i]+coef[taps-1-i])>0.001) anti = 0;
  }
  if (anti) sym=1;
  if (!sym) odd=0;
  staps = taps;
  if (sym) staps/=2;
  if (odd) staps++;
  otap = order;
  for (i=1; i<dec; i*=2) otap++;  /* disallow degenerate case of short decimating filters */
  ttaps = 1<<otap;
  for (; ttaps<staps; ttaps+=ttaps,otap++);
  tpc = 1<<(otap-order);
  cps = (tpc+dec-1)/dec;
  htaps =  sym? ttaps : ttaps/2;
  otaps = htaps / dec;
  o1 = (order==1)? 1:0;

  v2print("Filter order=%d dec=%d tpc=%d cps=%d ttaps=%d staps=%d sym=%d odd=%d anti=%d\n",
                                        order,dec,tpc,cps,ttaps,staps,sym,odd,anti);
  odec = dec;
  otpc = tpc;
  tpc--;
  cps--;
  dec--;
  /* form sequence */
  aadr=tpc, ashft=0;
  badr=dec, bshft=0;
  cadr=tpc, cshft=0;
  radr=1+(1-o1)*tpc, rshft=0;
  idec=0;
  if (odd) radr--;
  for (tcnt=0; tcnt<=tpc; tcnt++) {
    ccnt = tcnt&cps;
    aload = (ccnt==0);
    bload = (ccnt==cps);
    acc = (tcnt!=0);
    aco = (tcnt==tpc);
    if (bload) {
      if (idec==0) rshft=tpc-dec-o1;
      else rshft+=2;
      if (++idec>dec) idec=0;
    }
    seq = ((radr+rshft)<<18) | ((cadr+cshft)<<12) | ((badr+bshft)<<6) | ((aadr+ashft));
    if (aload) seq |= 0x01000000;
    if (bload) seq |= 0x02000000;
    if (acc)   seq |= 0x04000000;
    if (aco)   seq |= 0x08000000;
    v2print("Filter Sequence %d = %08x\n",tcnt,seq);
    sequence[tcnt] = seq;
    if (aload) { ashft++; radr++; }
    if (bload) { bshft++; radr--; }
    aadr-=odec; if (aadr<0)   aadr+=tpc;
    badr+=odec; if (badr>tpc) badr-=(tpc+2);
    cadr-=odec; if (cadr<0)   cadr+=tpc;
  }
  side = (port&0x1)? 1:2;
  v2print("FPGA filter load port=%d taps=%d pmod=%d tadr=%x\n",port,taps,pmod,addrs);
  addrt = addrs+4;

  /* load sequence */
  seq = 0x80000000;
  if (anti)  seq |= 0x10000000;
  if (!sym)  seq |= 0x20000000;
  FPGA_wpb (p,pmod,addrs,seq|(tpc<<16)|otaps);
  for (tcnt=tpc; tcnt>=0; tcnt--) FPGA_wpb (p,pmod,addrs,sequence[tcnt]);

  /* load taps */
  if (odd) coef[staps-1] /= 2;
  if (sym) {
    for (i=staps-1; i>=0; i--) FPGA_wpb (p,pmod,addrt,R2LC20(coef[i])|i);
    for (i=staps; i<ttaps; i++) FPGA_wpb (p,pmod,addrt,0);
  } else {
    for (i=staps; i<ttaps; i++) FPGA_wpb (p,pmod,addrt,0);
    for (i=taps-1; i>=0; i--) FPGA_wpb (p,pmod,addrt,R2LC20(coef[i])|i);
  }
  if (sym) ttaps *= 2;
  if (odd) ttaps -= 1;
  return ttaps;
}

int_4 FPGA_loadFIR (PICSTRUCT *p, int_4 addrs, int_4 order, int_4 dec, int_4 taps, real_4 *coef) {
  int_4 node=GETNODE;
  return setup_fpga_filter (p, 0, node, addrs, order, dec, taps, coef);
}
int_4 FPGA_loadTaps (PICSTRUCT *p, int_4 addrt, int_4 taps, real_4 *coef) {
  int_4 i,node=p->pmi;
  for (i=0; i<taps; i++) FPGA_wpb (p,node,addrt,R2LC20(coef[i])|i);
  return taps;
}
int_4 FPGA_loadRTaps (PICSTRUCT *p, int_4 addrt, int_4 taps, real_4 *coef) {
  int_4 i,node=GETNODE;
  FPGA_wpb (p,node,addrt,16); addrt+=4;
  for (i=0; i<taps/2; i++) FPGA_wpb (p,node,addrt,R2LC24(coef[i]));
  return taps;
}
int_4 FPGA_loadPPF (PICSTRUCT *p, int_4 addrt, int_4 rank, int_4 taps, real_4 *coef) {
  int_4 i,frame=taps/rank; /* banks of taps are pushed through the stack */
  for (i=0; i<rank; i++) FPGA_loadTaps(p,addrt,frame,coef+(i*frame));
  return taps;
}
int_4 FPGA_loadTwiddle (PICSTRUCT *p, int_4 addrt, int_4 taps) {
  int_4 i,node=GETNODE;
  for (i=0; i<taps; i++) {
    FPGA_wpb (p,node,addrt,R2LC20(cos(6.283185307179586*i/taps))|0x400|i);
    FPGA_wpb (p,node,addrt,R2LC20(sin(6.283185307179586*i/taps))|0x800|i);
  }
  return taps;
}
#endif
