package nxm.ice.lib;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import nxm.ice.prim.sourcepic;
import nxm.ice.lib.ICEPacket;

/**
  Extends DataFile to handle ICE Archiver functions.
  Although it acts as one file to the caller, it may split the data stream
  into multiple output files if setMulti() or setFunc() are non-zero.

  A TableOfContents is created with the name <archive>_toc.
  Individual channel files are <archive>_N where N is the SUFF field of the TOC entry.

  @author Jeff Schoen
  @version %I%, %G%
*/
public class Archiver extends DataFile {

  public final static String propertyList = "Func,Multi,MaxLines,MaxLineSize,Channel,Channels";

  public final static String functionList = "Normal,ICEUnpack,SDDSUnpack,ICEMulti,SDDSMulti";
  public final static int NORMAL=0, ICEUNPACK=1,SDDSUNPACK=2, ICEMULTI=3,SDDSMULTI=4, NIO=3;

  public final static int SDDSBUFSZ=1088*1024, ICEBUFSZ=1024*64, MAXNIOBUF=8192+64;

  private sourcepic sp;
  private DevIce pic;
  private NetIO nio;
  private DataFile htoc;
  private KeyVector channels;
  private Channel channel;
  private int multi=0,nchan=0,func=0,sock=0;
  private double tocOffset=0;
  private double maxLineSize=Constants.B1G;
  private int maxLines=0;
  private DataFile dfSuper;
  private String aaux,aaux1,aaux2;
  private Parser auxes;
  private int naux=0;
  private int packetLength=0;
  private boolean netio;
  private long niobufp;
  private SDDSPacket spkh;
  private ICEPacket ipkh;
  private Time stime;
  private int seqNext=0;
  private double niodelayfactor=0;

  public Archiver (sourcepic sp) {
    this.sp = sp;   // to get specs
    this.pic = sp.getDevIce();
    dfSuper = this; // for reference in Channel class
  }

  public int getMulti() { return multi; }
  public String getFunc() { return Parser.get(functionList,func,-1); }
  public int getMaxLines() { return maxLines; }
  public double getMaxLineSize() { return maxLineSize; }

  public void setMulti (int channels) { multi = channels; }
  public void setNChan (int channels) { nchan = channels; }
  public void setFunc (String function) { func = Parser.find(functionList,function,func,-1); }
  public void setMaxLines (int lines) { maxLines = lines; }
  public void setMaxLineSize (double bytes) { maxLineSize = bytes; }

  public void setThrottle (double rate) {
    double drate = (1.0/getDelta()) * getBPE();
    if (multi>0) drate *= multi;
    if (nchan>0) drate *= nchan;
    drate *= 1e-6; // in MBy/s
    if (rate>0) drate = rate;
    else drate *= -rate;
    niodelayfactor = (rate==0 || drate==0)? 0 : 1.0/drate;
  }

  public boolean open() {

    // is this a network output
    String fname = fn.getFullName();
    netio = fname.startsWith("udp:") || fname.startsWith("stp:");
    fname = netio? "network_toc" : fn.getRoot()+"_toc";
    if (netio) flags &= ~NATIVE;

    Object qfunc = (String)this.getQualifier("FUNC");
    if (qfunc!=null) setFunc((String)qfunc);
    Object qmulti = this.getQualifier("MULTI");
    if (qmulti!=null) setMulti(Convert.o2l(qmulti));
    Object qthrottle = this.getQualifier("THROTTLE");
    if (qthrottle!=null) setThrottle(Convert.o2d(qthrottle));

//   System.out.println("Archiver:Open url="+fname+" multi="+multi+" func="+getFunc());

    // open the table of contents
    Object ref=cmd; if (ref==null) ref=M;
    htoc = new DataFile();
    htoc.init (ref,fname,"3000","NH",(flags&~WRAP)); // get same flags OUTPUT|APPEND as archive file
    htoc.setSubRecords("SUFF|1A,STAT|1A,KEY|SI,PORT|SI,ADDR|4B,FORM|1A,RATE|SD,FREQ|SD,TIME|SD,LEN|SD");
    htoc.open();
    tocOffset = htoc.seek();
    double curOffset = htoc.keywords.getD("CUROFFSET",-1);
    if (curOffset>=0 && curOffset!=tocOffset) { 
      tocOffset=curOffset; 
      htoc.seek(tocOffset);
    }

    if (func>=NIO) {		// network IO open
      nio = new NetIO(true);
      sock = nio.nioOpen(fn.getFullName(),multi,0);
      niobufp = nio.nioAlloc(MAXNIOBUF);
      spkh = new SDDSPacket( getFormat(), SDDSPacket.ICE );
      ipkh = (ICEPacket)pkh;
      stime = new Time();
      isOpen = (sock>0);
    }
    else if (multi==0) {	// single channel
      channel = new Channel(this,0);
      channel.setFreq(sp.getFreq());
      channel.setTime(sp.getTime().getSec());
      channel.open();
      super.open();		// this is the actual archive file
    }
    else {			// multi mode file segments are opened on demand
      channels = new KeyVector(multi);
      for (int i=0; i<multi; i++) {
        Channel c = new Channel(this,i);
        channels.put(i,"CHAN_"+i,c);
      }
      super.open();		// this is now a file collection that access all the others
    }
    return isOpen;
  }

  public void flush() {
    if (func>=NIO);
    else if (multi==0) channel.flush();
    else for (int i=0; i<multi; i++) getChannel(i).flush();
    super.flush();
    htoc.flush();
  }

  public void close() {
    if (func>=NIO) { nio.nioClose(sock); nio.nioFree(niobufp); }
    else if (multi==0) channel.close();
    else for (int i=0; i<multi; i++) getChannel(i).close();
    if (func<NIO) super.close();
    htoc.close();
  }

  public int write (long lbuf, int boff, int bytes, int chan) {
    int stat=0;
    if (func>=NIO) {
      int niodelay = (int)(niodelayfactor*bytes);
      if (func==SDDSMULTI) {
        double delta = getDelta();
	ipkh.getTC(stime,0.0,delta); // get TC of first sample of ICE packet
	spkh.setTC(0,delta,stime.getWSec(),stime.getFSec()); // set TC of first SDDS packet
	if (chan==0) seqNext=spkh.getCount();
        spkh.setCount(seqNext);
	for (int i=0; i<bytes; i+=1024) {
          Native.ja2p(spkh.buf,0, niobufp,0, 64);
          Native.p2p(lbuf,boff+i, niobufp,64, 1024);
          nio.nioSend(sock,chan,niobufp+8,1080, (i==0)?niodelay:0);
	  if (i==0) spkh.setTC(0,0.0,0.0,0.0); // no valid SDDS TC on the rest
	  spkh.upCount();
        }
      }
      else if (boff>=64) {
        // OK to overwrite old data buffer with packet header for speed
        Native.ja2p(ipkh.buf,0, lbuf+boff,-64, 64);
        nio.nioSend(sock,chan,lbuf+boff-64,bytes+64,niodelay);
      } else {
	// copy header and data to contiguous buffer
        Native.ja2p(ipkh.buf,0, niobufp,0, 64);
        Native.p2p(lbuf,boff, niobufp,64, bytes);
        nio.nioSend(sock,chan,niobufp,bytes+64,niodelay);
      }
    }
    else if (func==SDDSUNPACK && chan<0) {
      int ndo=Math.min(bytes,1088), bend=boff+bytes;
      for (; boff<bend; boff+=ndo) {
	chan = findSDDSChannel (lbuf,boff,bytes);
	if (chan<0) M.error("Out of archive channels");
	stat += write (lbuf,boff,ndo,chan);
      }
    }
    else if (multi==0) {
      stat = write(lbuf,boff,bytes);
      channel.bytes += stat;
    }
    else {
      if (chan<0 && multi==1) chan=0;
      Channel c = getChannel(chan);
      if (!c.isOpen) {
        Time t = sp.getTime();
        if (func==SDDSUNPACK) t = SDDSPacket.getTC(null,boff,lbuf);
	c.setFreq(sp.getFreq());
	c.setTime(t);
	c.open();
	if (func==SDDSUNPACK) c.df.keywords.putMain("PACKET","SDDS/ICE");
      }
      if (c.buf!=null) {	// cache small writes
	Native.p2ja(lbuf,boff, c.buf,c.inbuf, bytes); c.inbuf += bytes;
	if (c.inbuf>=c.buf.length) { c.df.write(c.buf,0,c.inbuf); c.inbuf=0; }
	stat = bytes;
      } else {
        stat = c.df.write(lbuf,boff,bytes);
      }
      c.bytes += stat;
      if (maxLineSize>0 && c.bytes >= maxLineSize) c.close();
      c.hasNewData = true;
    }
    return stat;
  }

  public void setAAux (String aux) { 
    naux = 0;
    if (aux==null || aux.length()==0) return;
    int ic = aux.indexOf("<>");
    if (ic>0) {  // parse i1:i2 syntax
      int i1 = Convert.s2l(aux.substring(0,ic));
      int i2 = Convert.s2l(aux.substring(ic+2));
      aux = ""+i1; for (int i=i1+1; i<=i2; i++) aux += "|"+i;
    }
    if (aux.indexOf('|')>0) { 
      auxes = new Parser(aux,'|');
      naux = auxes.elements();
    }
    else {
      naux=1; 
      aaux = aux;
    }
  }
  public void setAAux1 (String aux) { 
    if (aux==null || aux.length()==0) return;
    M.deprecate("AAUX1=x,AAUX2=y changed to AAUX=x|y");
    aaux1 = aux; if (aaux2!=null) setAAux (aaux1+"|"+aaux2);
  }
  public void setAAux2 (String aux) { 
    if (aux==null || aux.length()==0) return;
    M.deprecate("AAUX1=x,AAUX2=y changed to AAUX=x|y");
    aaux2 = aux; if (aaux1!=null) setAAux (aaux1+"|"+aaux2);
  }

  public Channel getChannel() { return channel; }
  public Channel getChannel (int i) { return (Channel)channels.get(i); }
  public KeyVector getChannels() { return channels; }

  public class Channel {
    short port;
    int addr;
    double offset,bytes;
    DataFile df;
    Data dtoc;
    boolean isOpen;
    boolean hasNewData;
    Time time = new Time();
    byte[] buf;
    int inbuf=0;

    public Channel (DataFile df, int index) {
      this.df = df;
      dtoc = htoc.getDataBuffer(1);
      setSuffix("");
      setStatus("OPEN");
      setFormat(sp.getFormat());
      setPort((short)(index+1));
      setAddress(0);
      setFreq(0.0);
      setTime(0.0);
      if (func==SDDSUNPACK) buf = new byte[SDDSBUFSZ];
    }
    public boolean open() {
      if (isOpen) close();
      setStatus("OPEN");
      if (maxLines>0 && tocOffset>=maxLines) tocOffset=0;
      offset = tocOffset++;
      bytes = 0;
      boolean stat=true;
      if (multi>0) {
	int suffix = (int)(offset+1);
	setSuffix(""+suffix);
	String fname = fn.getRoot()+"_"+suffix;
	Object ref=cmd; if (ref==null) ref=M;
	df = new DataFile();
	df.init (ref,fname,dfSuper,(flags&~APPEND));
	df.fn.setQualifiers(dfSuper.fn.getQualifiers());
	int entry = (int)offset;
        if (naux==1) df.setAux(aaux);
        else if (naux>1) df.setAux(auxes.get((entry%naux)+1));
	if (time!=null) df.setTime(time);
	stat = df.open();
      }
      htoc.keywords.put("CUROFFSET",Convert.d2o(tocOffset));
      setRate(1.0/df.getXDelta());
      setLength(0.0);
      isOpen = stat;
      return stat;
    }
    public void flush() {
      if (!isOpen) return;
      if (inbuf>0) { df.write(buf,0,inbuf); inbuf=0; }
      if (multi>0) df.flush();
      double databytes = bytes;
      if (func==SDDSUNPACK) databytes = bytes*1024/1088;
      setLength( (databytes/df.dbpe)*df.getDelta() );
      htoc.write (dtoc,offset,1);
    }
    public void close() {
      setStatus ("CLOSED");
      flush();
      if (multi>0 && df!=dfSuper) df.close();
      isOpen = false;
    }

    public void setSuffix (String suffix) { dtoc.packS(0,8,suffix); }
    public void setStatus (String state) { dtoc.packS(8,8,state); }
    public void setKey (short key) { dtoc.packI(16,key); }
    public void setPort (short port) { dtoc.packI(18,port); this.port=port; }
    public void setAddress (int addr) { dtoc.packL(20,addr); this.addr=addr; }
    public void setFormat (String format) { dtoc.packS(24,8,format); }
    public void setRate (double rate) { dtoc.packD(32,rate); }
    public void setFreq (double freq) { dtoc.packD(40,freq); }
    public void setTime (double time) { dtoc.packD(48,time); }
    public void setLength (double sec) { dtoc.packD(56,sec); }
    public void setTime (Time time) { this.time=time; if (time!=null) setTime(time.getSec()); }

    public String getSuffix() { return dtoc.unpackS(0,8); }
    public String getStatus() { return dtoc.unpackS(8,8); }
    public short  getKey() { return dtoc.unpackI(16); }
    public short  getPort() { return dtoc.unpackI(18); }
    public int    getAddress() { return dtoc.unpackL(20); }
    public String getFormat() { return dtoc.unpackS(24,8); }
    public double getRate() { return dtoc.unpackD(32); }
    public double getFreq() { return dtoc.unpackD(40); }
    public double getTime() { return dtoc.unpackD(48); }
    public double getLength() { return dtoc.unpackD(56); }

  }

  public int findSDDSChannel (long lbuf, int boff, int bytes) {
    byte[] buffer = new byte[8];
    Native.p2ja(lbuf,boff,buffer,0,8);
    Convert.swap4(buffer,4,1);
    int addr = Convert.unpackL(buffer,4);
    Convert.swap2(buffer,2,1);
    short port = Convert.unpackI(buffer,2);
    int iopen = -1;
    for (int i=0; i<multi; i++) {
      Channel c = (Channel)channels.get(i);
      if (c.port==port && c.addr==addr) return i;
      else if (c.isOpen);
      else if (iopen<0) iopen=i;
    }
    if (iopen>=0) {
      Channel c = (Channel)channels.get(iopen);
      c.setPort(port);
      c.setAddress(addr);
    }
    return iopen;
  }

  public void checkChannels() {
    for (int i=0; i<multi; i++) {
      Channel c = (Channel)channels.get(i);
      if (c.isOpen && !c.hasNewData) {
	//System.out.println("Check Closing "+c+" "+Time.tag());
        c.close();
      }
      c.hasNewData = false;
    }
  }

}
