package nxm.ice.lib;

import nxm.sys.inc.Constants;
import nxm.sys.inc.DataTypes;

import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;
import nxm.sys.lib.Args;
import nxm.sys.lib.Shell;
import nxm.sys.lib.Parser;
import nxm.sys.lib.MidasException;

/**
  Structure for a Code-Once-Run-Everywhere library function
  Structure for a Hardware Abstraction Library Object 

  This function may have implementations in more than one language.

  Core.init	- instantiate, allocate resources
  Core.set	- set key=value named parameters
  Core.get	- get key=value named parameters
  Core.open	- ready for processing
  Core.process	- process a buffer of data
  Core.close	- finish up processing

  @author Jeff Schoen
  @version $Id: Core.java,v 1.2 2011/05/24 13:20:30 jgs Exp $
*/
public class Core extends Object implements Constants,DataTypes {

  public static int FORMATS=1;
  public int implementation;

  public int func,fmts;
  public int ni1,ni2;
  public int no1,no2;
  public byte[] bi1,bi2;
  public byte[] bo1,bo2;

  /** Create an instance of the function specified by name and parsing of the Args class */
  public static Core forName (String name, Args args) {
    return CoreFactory.getCore(name,args);
  }

  public Core () {
  }

  /** Initialize this engine with given config */
  public void allocNative (String config) {
    System.out.println("Err: Should not be calling allocNative from this implementation");
  }

  /** Ready this engine with current parameters */
  public int open() {
    return 0;
  }

  /** Process a buffer of data */
  public int process() {
    return 0;
  }
  public int process(byte[] buf1, int n1) {
    bi1=buf1; ni1=n1;
    return process();
  }
  public int process(byte[] buf1, int n1, byte[] buf2, int n2) {
    bi1=buf1; ni1=n1;
    bo1=buf2; no1=n2;
    return process();
  }
  public int process(byte[] buf1, int n1, byte[] buf2, int n2, byte[] buf3, int n3) {
    bi1=buf1; ni1=n1;
    bi2=buf2; ni2=n2;
    bo1=buf3; no1=n3;
    return process();
  }
  public int process(Data buf1, int n1) {
    return process(buf1.buf,n1*buf1.bpe) / buf1.bpe;
  }
  public int process(Data buf1, int n1, Data buf2, int n2) {
    return process(buf1.buf,n1*buf1.bpe,buf2.buf,n2*buf2.bpe) / buf2.bpe;
  }
  public int process(Data buf1, int n1, Data buf2, int n2, Data buf3, int n3) {
    return process(buf1.buf,n1*buf1.bpe,buf2.buf,n2*buf2.bpe,buf3.buf,n3*buf3.bpe) / buf3.bpe;
  }

  /** Finalize this engine */
  public int close() {
    return 0;
  }


  /** SetKey for generic parameter. Override this method and call super if not handled  */
  public void set (String key, Data value) {
    if (key.startsWith("BOGUS"));
    else if (key.equals("ICE.CONFIG")) System.out.println("Setting the key="+key+" to "+value);
    else throw new MidasException ("Problem setting key="+key+" to value="+value);
  }
  /** SetKey helper for String parameter */
  public void set (String key, String value) {
    set (key, new Data(value));
  }
  /** SetKey helper for double parameter */
  public void set (String key, double value) {
    set (key, new Data(value));
  }
  /** SetKey helper for float parameter */
  public void set (String key, float value) {
    set (key, new Data(value));
  }
  /** SetKey helper for int parameter */
  public void set (String key, int value) {
    set (key, new Data(value));
  }
  /** SetKey helper for state parameter */
  public void set (String key, boolean value) {
    set (key, new Data(value?1:0));
  }
  /** SetKey helper for float array parameter */
  public void set (String key, float[] value) {
    set (key, new Data(value));
  }
  /** SetKey helper for int array parameter */
  public void set (String key, int[] value) {
    set (key, new Data(value));
  }
  /** SetKey helper for int array parameter */
  public void set (String key, short[] value) {
    set (key, new Data(value));
  }
  /** SetKey helper for int array parameter */
  public void set (String key, byte[] value) {
    set (key, new Data(value));
  }


  /** GetKey for generic parameter. Override this method and call super if not handled  */
  public Data get (String key, byte type) {
    Data value = null;
    if (key.startsWith("BOGUS"));
    else throw new MidasException ("Problem getting key="+key);
    return value;
  }
  /** GetKey helper for String parameter */
  public String getS (String key) {
    Data value = get(key,DataTypes.STRING);
    return value.toString();
  }
  /** GetKey helper for double parameter */
  public double getD (String key) {
    Data value = get(key,DataTypes.DOUBLE);
    return value.toD();
  }
  /** GetKey helper for float parameter */
  public float getF (String key) {
    Data value = get(key,DataTypes.FLOAT);
    return value.toF();
  }
  /** GetKey helper for int parameter */
  public int getL (String key) {
    Data value = get(key,DataTypes.LONG);
    return value.toL();
  }
  /** GetKey helper for state parameter */
  public boolean getState (String key) {
    Data value = get(key,DataTypes.LONG);
    return value.toL()>0;
  }
  /** Helper for testing options bits */
  public boolean isSet (String list, String key, int mask) {
    int kmask = Parser.mask(list,key,0);
    return ((mask&kmask) != 0);
  }

  /** get the natural data format type */
  public byte getDataType() { 
    return CoreFactory.getDataType(implementation);
  }

  public static int getFormats (String fmts) {
    char fm=fmts.charAt(0), ft=fmts.charAt(1);
    int fmt = 0;
         if (ft=='I') fmt = 0x0; 
    else if (ft=='B') fmt = 0x1;
    else if (ft=='N') fmt = 0x2;
    else if (ft=='P') fmt = 0x3;
    else if (ft=='L') fmt = 0x4;
    else if (ft=='X') fmt = 0x5;
    else if (ft=='F') fmt = 0x6; 
    else if (ft=='D') fmt = 0x7; 
    if      (fm=='C') fmt |= 0x8;
    return fmt;
  }

  public static int getFormats (String fmt1, String fmt2) {
    return (getFormats(fmt1)<<0) | (getFormats(fmt2)<<8);
  }
  public static int getFormats (String fmt1, String fmt2, String fmt3) {
    return (getFormats(fmt1)<<0) | (getFormats(fmt2)<<8) | (getFormats(fmt3)<<16);
  }

  public int getFormatBits (int fmts, int n) {
    int fmt = (fmts>>((n-1)*8)) & 0xFF;
    int bps = 1;
    if ((fmt&0x8)!=0) { bps=2; fmt=fmt&0x7; } // complex
         if (fmt==0) bps *= 16; // int
    else if (fmt==1) bps *= 8;  // byte
    else if (fmt==2) bps *= 4;  // nibble
    else if (fmt==3) bps *= 1;  // bit
    else if (fmt==4) bps *= 32; // long
    else if (fmt==5) bps *= 64; // xlong
    else if (fmt==6) bps *= 32; // float
    else if (fmt==7) bps *= 64; // double
    return bps;
  }
  public int getFormatBytes (int fmts, int n) {
    int byps = getFormatBits(fmts,n)>>3;
    return (byps<=0)? 1:byps;
  }
  public int getFormatBytes (int n) {
    return getFormatBytes(fmts,n);
  }
  public int getFormatBits (int n) {
    return getFormatBits(fmts,n);
  }


  public int setFormats (String fmt1, String fmt2) {
    fmts = getFormats(fmt1,fmt2);
    return fmts;
  }

}
