package nxm.ice.lib;

import nxm.sys.inc.DataTypes;
import nxm.sys.lib.Args;
import nxm.sys.lib.Parser;
import nxm.sys.lib.Convert;
import nxm.sys.lib.Shell;

/**
  Structure for a Code-Once-Run-Everywhere library function
  Structure for a Hardware Abstraction Library Object 

  This function may have implementations in more than one language.

  The NeXtMidas policy is to code CORE functions in C or Fortran to allow
  code sharing with other versions of Midas.  All other framework and GUI
  code is in Java for portability and deployment options.

  Typically, math functions needed by GUIs are included in the SYS tree.
  Some functions are coded here in Java and in C for performance comparisons.

  The Java code for each function library exposes methods for setting and
  getting each of the library's properties.  If native code is used, the
  Java wrapper loads the native code and calls the standard library interface
  routines as appropriate.

  This class implements a factory for instantiating alternate implementations.
  The innerclasses of this factory comprise the supported algorithm implementations.

  The executable names will adhere to the following naming conventions when compiled by NeXtMidas:

   Foo$JVM.class   - the java implementation of the function Foo 
   Foo$CPU_xxx.so  - the CPU native implementation of Foo (C)
   Foo$PHI_xxx.so  - the Intel Xeon Phi accelerator card
   Foo$GPU_xxx.so  - the GPU implementation of Foo (CUDA)
   Foo$VHS_xxx.so  - the Verilator Hardware Simulator implementation of Foo (C++)
   Foo$MLS_xxx.so  - the MatLab Simulink implementation of Foo (C++)
   Foo$ICE_xxx.so  - the ICE FPGA implementation of Foo (FPGA)

  Where xxx is an architecture dependent tag such as lin32 or arm32.

  These will typically be conditionally compiled from common source code Foo.c and Foo.java
  in NeXtMidas but may be compiled differently with custom build scripts.

  @author Jeff Schoen
  @version $Id: CoreFactory.java,v 1.2 2011/05/24 13:20:30 jgs Exp $
*/
public class CoreFactory {

  public static String factoryList = "JVM,CPU,PHI,GPU,VHS,MLS,ICE";
  public static int JVM=0, CPU=1, PHI=2, GPU=3, VHS=4, MLS=5, ICE=6;

  /** Create an instance of the function specified by name and parsing of the Args class */
  public static Core getCore (String name, Args args) {
    String config = null;
    String path = "nxm."+args.option.toLowerCase()+".core";
    int impl=ICE, coreio=1;
    if (args.getS("/CORE").equals("ICEX")) coreio=0;
    else impl = args.getSelectionIndex("/CORE",factoryList,0,-1);
    if (impl==VHS && args.isPresent("/COREDEV")) impl=ICE;
    if (impl==ICE) {
      String alias  = args.getS("/COREDEV","PIC1AUTO:11");
      int ic = alias.indexOf(':');
      int cindex = 11;
      if (ic>0) {
	String cport = alias.substring(ic+1);
	if (cport.startsWith("CORE")) cport = cport.substring(4);
        cindex=Convert.s2l(cport);
        alias=alias.substring(0,ic);
      }
      int pmi    = cindex/10;
      int ci     = cindex%10;
      int side   = (ci+1)%2+1;
      config = MDevIce.getConfig(args,alias);
      if (config.indexOf(",IPORT=")<0) config += ",IPORT=PM0CORE"+side;
      if (config.indexOf(",OPORT=")<0) config += ",OPORT=PM"+pmi+"CORE"+ci;
      config += ",COREIO="+coreio;
    }
    Core core = getCore(path,name,impl);
    if (impl!=JVM) core.allocNative(config);
    return core;
  }

  /** Create an instance of the function specified by path, name, and implementation */
  public static Core getCore (String path, String name, int impl) {
    name = path+"."+name+"$"+Parser.get(factoryList,impl,-1); 	// add inner class $xxx suffix
    Core core = (Core)Shell.getInstanceForClassName(name);
    core.implementation = impl;
    return core;
  }

  /** get the natural data format type */
  public static byte getDataType (int implementation) {
    if (implementation==VHS) return DataTypes.INT;
    if (implementation==ICE) return DataTypes.INT;
    return DataTypes.FLOAT;
  }

}
