package nxm.ice.lib;

import nxm.sys.inc.DataTypes;

import nxm.sys.lib.Data;
import nxm.sys.lib.Shell;
import nxm.sys.lib.Convert;
import nxm.sys.lib.MidasException;

/**
  Structure for a Native Core 

  Native code libraries should have the following methods:
    alloc()  -  allocates the handle to an instance of this library
    setkey() -  sets a parameter
    getkey() -  gets a parameter
    init()   -  initialize the library for the current set of parameters
    work()   -  perform the desired function on a segment of data
    free()   -  release all resources

  @author Jeff Schoen
  @version $Id: CoreNative.java,v 1.3 2011/08/12 12:47:37 jgs Exp $
*/
public class CoreNative extends Core {

  protected long handle;
  protected String config;

  public native long alloc (); 
  public native int config (long handle, int key, int value);
  public native int setkey (long handle, String key, byte[] value, int len);
  public native int getkey (long handle, String key, byte[] value, int len);
  public native int init   (long handle);
  public native int work   (long handle, byte[] bi1, int ni1, byte[] bi2, int ni2, byte[] bo1, int no1, byte[] bo2, int no2);
  public native int free   (long handle);

  public static boolean useNative = false;

  /** Create an instance of this function with native implementation */
  public CoreNative () {
    useNative |= loadNative("nxm.ice.lib.CoreNative");
    loadNative(getClass().getName());
  }

  /** load the native libraries */
  private static boolean loadNative (String libName) {
    boolean ok = Shell.loadLibrary(libName);
    if (!ok) throw new MidasException ("Problem with load of native library: "+libName);
    return ok;
  }

  public int setFormats (String fmt1, String fmt2) {
    return config(handle,FORMATS,getFormats(fmt1,fmt2) );
  }

  public void allocNative (String config) {
    this.config = config;
    handle = alloc();
    if (handle==0) throw new MidasException ("Problem with alloc of native config="+config);
  }

  public void set (String key, Data value) {
    char ctype = (char)value.type; if (ctype=='A') ctype='S';
    int len = setkey (handle,ctype+":"+key,value.buf,value.size*value.bpe);
    if (len<=0) super.set(key,value);
  }

  public Data get (String key, byte type) {
    byte[] buffer = new byte[256];
    char ctype = (char)type; 
    int len = getkey (handle,ctype+":"+key,buffer,buffer.length);
    if (type==DataTypes.STRING) {
      type = DataTypes.ASCII;
      for (;len%8!=0;) buffer[len++]=0;
    }
    if (len>0) return new Data(buffer,0,len,type,1);
    else return super.get(key,type);
  }

  public int open() {
    return init(handle);
  }

  public int process() {
    return work(handle, null,0,null,0, null,0,null,0);
  }
  public int process(byte[] buf1, int n1) {
    return work(handle, buf1,n1,null,0, null,0,null,0);
  }
  public int process(byte[] buf1, int n1, byte[] buf2, int n2) {
    return work(handle, buf1,n1,null,0, buf2,n2,null,0);
  }
  public int process(byte[] buf1, int n1, byte[] buf2, int n2, byte[] buf3, int n3) {
    return work(handle, buf1,n1,buf2,n2, buf3,n3,null,0);
  }
  public int process(Data buf1, int n1) {
    return work(handle, buf1.buf,n1*buf1.bpe, null,0, null,0, null,0) / buf1.bpe;
  }
  public int process(Data buf1, int n1, Data buf2, int n2) {
    return work(handle, buf1.buf,n1*buf1.bpe, null,0, buf2.buf,n2*buf2.bpe, null,0) / buf2.bpe;
  }
  public int process(Data buf1, int n1, Data buf2, int n2, Data buf3, int n3) {
    return work(handle, buf1.buf,n1*buf1.bpe, buf2.buf,n2*buf2.bpe, buf3.buf,n3*buf3.bpe,null,0) / buf3.bpe;
  }

  public int close() {
    return free(handle);
  }

}
