package nxm.ice.lib;

import nxm.ice.lib.IceHW;

/* 
            WR    RD
  saddr[16] SYS / Status
  saddr[17] IRB / Ram
  saddr[18] VCO

  Status = 0x0CBBBBBA - C=currentTenths B=10MHzTickLoc A=160MHzTickLoc
  VCO    = 0x0000AAAA - A=VoltageOffset 0 to 0xFFFF

 */

/* IceGPS Algorithm

  TFOM Levels

  1 < 10^1 ns
  2 < 10^2 ns
  3 < 10^3 ns
  ...
  8 < 10^8 ns
  9 > 10^8 ns

  Clock = 160 MHz
  tstep = 6.25 ns

 */
public class IceGPS implements IceHW {

  // special static vars recognized by Ice JVM loader
  public static int saddr,sdata; // system bus
  public static int maddr,mdata; // memory bus
  public static int ioadr,ioctl; // IO bus

  // normal static vars
  public static int lvco,loff,resync;
  /** dilution of position */
  public static int dop;
  /** speed over ground */
  public static int sog;
  /** number of satellites */
  public static int nos;
  /** last second */
  public static int lsec;
  /** counter */
  public static int count;

  /** Gross frequency tuning */
  public static int Fg;
  /** Fine frequency tuning */
  public static int Ff;
  /** Gross time tuning */
  public static int Tg;
  /** Fine time tuning */
  public static int Tf;
  /** long-term average count */
  public static int acnt;
  /** long-term offset average */
  public static int aoff;

  public static void main() {
    run();
  }

  public static void run() {
    lsec = 0;
    dop = 10; 
    loff = 0x0;
    // 
    count = nos = 0;
    // initialize averages
    acnt = aoff = 0;
    Fg = Ff = Tg = Tf = 0;

    // try and find the last value
    lvco = arMem(0x00010);
    if ( lvco < 0x2000 ) {
      // value not found, init DAC to midpoint    
      lvco = 0x9000;
      resync = 3;
      // initial sync
      awBus(0x10000, 0x1);
    }
    else {
      resync = 2;      
    }
    awBus(0x40000,lvco);
    awMem(0x00010,lvco);

    int status;
    for (int beat=1;;beat++) {
      awMem(HBT,beat);					// heartbeat monitor
      status = arBus(0x10000);			// poll for counter to advance 
      if ((status&0x70000000)!=0x30000000) continue;    // wait for .3 sec in
      processTick (status);				// update VCO register
      processIRB (status);				// update IRIG register
      while ((arBus(0x10000)&0x70000000)==0x30000000) udelay(10); // wait for > .3 sec in
    }

  }

  public static void processTick (int status) {
    int offin = (status&0x0FFFFFFF);
    // center about zero (account for delay)
    int off = ( offin > 80000000 ) ? offin-160000018 : offin-18;
    awMem(0x0000C, status);
    // get current time
    int sec = arMem(0x00008);
    if ( resync > 0 ) {
      // re-aligning PPS
      resync--;
      awBus(0x10000, 0x0);
      loff = off;
      Tg = Tf = 0;
      return;
    }

    // get delta's (from last process)
    int dsec = sec - lsec; 
    int doff = off - loff;
    // update for next pass
    loff = off;
    lsec = sec;   
    // validate time delta
    if ( dsec != 1 ) {
      return;
    }
    count++;
    // long-term counter average
    int ccnt = ( doff << 20 ) - acnt;
    acnt += ( (ccnt+128)>>8 ); 
    ccnt = ( (acnt+512) >> 10 );
    // long-term offset average
    int coff = ( off << 20 ) - aoff;
    aoff += ( (coff+128)>>8 );
    coff = ( (aoff+512) >> 10 );

    // Quality of Signal Calculation
    int weight;
    if ( nos > 6 ) {
      weight = ( dop < 23 ) ? 26 : 25;
    }
    else if ( nos > 4 ) {
      weight = 24;
    }
    else {
      weight = 0;
    }

    //  calculate TFOM
    int tfom;
    int sacnt = ( acnt < 0 ) ? -acnt : acnt;
    int saoff = ( aoff < 0 ) ? -aoff : aoff;
    if ( weight>0 && sacnt<1048576 ) { 
      if ( sacnt < 16384 ) {
        if ( saoff < 524288 ) {
          tfom = 4;
        }
        else if ( saoff < 2097152 ) {
          tfom = 5;
        }
        else {
          tfom = 6;
        }
      }
      else if ( sacnt < 65536 ) {
        tfom = 7;
      }
      else if ( sacnt < 131072 ) {
        tfom = 8;
      }
      else {
        tfom = 9;
      }
    }
    else {
      tfom = 0;
    }
    int lock;
    if ( tfom > 0 ) {
      if ( off<100 && off>-100 ) {
        lock = 1;
      }
      else {
        lock = off = 0;
        // PPS not worth correcting, gross adjustment
        resync = 3;
        awBus(0x10000,0x1);
        lsec -= 2;
        // prevent getting stuck in a cycle of reset's
        acnt = ( acnt < 0 ) ? acnt-524288 : acnt+524288;
      }
    }
    else { 
      lock = off = 0;
    }
    // averaging function 
    stableAverage(off, doff, weight);

    // update registers    
    awBus(0x40000, lvco);
    awMem(0x00010, lvco);
    awMem(0x00014, ((0x001F&weight)<<17)|((0x1&lock)<<16)|(0xFFFF&tfom));
    awMem(0x00018, ccnt);
    awMem(0x0001C, coff);
  }

  public static void stableAverage ( int ioff, int doff, int weight ) { 
    // Gross Frequency Correction
    int cfg = ( doff<<weight ) - Fg;
    Fg += ( (cfg+16)>>5 );
    int dfg = ((Fg+16777216)>>25)<<6;

    // limit for time corrections
    int soff = ioff;
    if ( ioff < -4 ) {
      soff = -4;
      if ( ioff < -49 ) {
        ioff = -49;
      }
    }
    else if ( ioff > 4 ) {
      soff = 4;
      if ( ioff > 48 ) { 
        ioff = 48;
      }      
    }
    // Gross Time Correction
    ioff = ( (ioff+8)>>4 );
    int ctg = ( ioff<<weight ) - Tg;
    Tg += ( (ctg+8)>>4 );
    int dtg = ((Tg+33554432)>>26);

    // limit offset for 'finer' corrections
    if ( doff < -3 ) {
      doff = -3;
    }
    else if ( doff > 3 ) {
      doff = 3;
    }
    
    // Fine Time Correction
    int ctf = ( soff<<weight ) - Tf;
    Tf += ( (ctf+1024)>>11 );
    int dtf = ((Tf+33554432)>>26);

    // Fine Frequency Correction
    int cft = ( doff<<weight ) - Ff;
    Ff += ( (cft+128)>>8 );
    int dff = ((Ff+262144)>>19);

    // VCO adjustment
    int dvco = dfg + dff + dtg + dtf;
    // check for update
    if ( count>19 && dvco!=0 ) {
      lvco -= dvco;
      count = 0;
      // zap gross estimation
      Fg = ( dfg == 0 ) ? Fg : ( (Fg+2)>>2 );
      // scaled down fine freq by 3/4 if active
      Ff = ( dff == 0 ) ? Ff : ( ((Ff<<1)+Ff+2)>>2 );
      // half fine time if active
      Tf = ( dtf == 0 ) ? Tf : ( (Tf+1)>>1 );
    }    
  }

  public static void processIRB (int status) {
    int iadr = 0x20000;

    // NMEA returns 10 lines, GPRMC is the 2nd & POLYP is the 9th
    iadr = checkLine(iadr);  //  1
    iadr = checkLine(iadr);  //  2
    iadr = checkLine(iadr);  //  3
    iadr = checkLine(iadr);  //  4
    iadr = checkLine(iadr);  //  5
    iadr = checkLine(iadr);  //  6
    iadr = checkLine(iadr);  //  7
    iadr = checkLine(iadr);  //  8
    iadr = checkLine(iadr);  //  9
  }

  public static int checkLine ( int iadr ) {
    // get the relevant characters
    int a = arBus(iadr+3);
    int b = arBus(iadr+4);
    int c = arBus(iadr+5);
    int nadr;

    // check for GPRMC (get seconds, day, month, year & SOG)
    if ( a==82 & b==77 & c==67 ) {
      // $GPRMC,175151.012,A,3851.740204,N,07718.137225,W,0.001,353.40,150513,,,A*49
      int sec=0, day=0, mon=0, year=0;
      // field 1 is HHMMSS
      iadr = nextField(iadr, 1);
      nadr = nextField(iadr, 1);
      // check for a valid field length
      if ( nadr-iadr > 5 ) {
        sec = arBCD(iadr, 5);
        sec += 10*arBCD(iadr, 4);
        sec += 60*arBCD(iadr, 3);
        sec += 600*arBCD(iadr, 2);
        sec += 3600*arBCD(iadr, 1);
        sec += 36000*arBCD(iadr, 0);

        // get speed over ground in knots (5 fields away from nadr)
        iadr = nextField(nadr, 5);
        nadr = nextField(iadr, 1);
        sog  = scaleDecimal(iadr, nadr);

        // get DDMMYY (1 field away from nadr)
        iadr = nextField(nadr, 1);
        nadr = nextField(iadr, 1);
        // make sure field length is valid
        if ( nadr-iadr > 5 ) {
          // day 
          day = 10*arBCD(iadr,0) + arBCD(iadr,1);
          // month
          mon = 10*arBCD(iadr,2) + arBCD(iadr,3);
          // year
          year = 10*arBCD(iadr,4) + arBCD(iadr,5);
          // update the IRIG-B values
          updateIrigB(year, mon, day, sec);
        }
        iadr = nadr;
      }
    }
    // POLYP (get DOP & NOS)
    else if ( a==76 && b==89 && c==80 ) {
      // $POLYP,175151.012,3851.740204,N,07718.137225,W,92.612,G3,002.1,005.5,0.002,353.40,0.000,,0.82,1.35,1.58,1.74,0.71,9,,0*2F
      // 3-D DOP 
      iadr = nextField(iadr, 16);
      nadr = nextField(iadr, 1);
      dop = scaleDecimal(iadr, nadr);
      // Satellite information is 3 fields away (or 2 from nadr)
      iadr = nextField(nadr, 2);
      nadr = nextField(iadr, 1);
      nos = readSatellites(iadr, nadr);
    }
    return nextLine(iadr, 200);
  }

  public static void updateIrigB ( int year, int mon, int day, int sec ) {
    int irb0=0,irb1=0,irb2=0;
    int cn,dn;
    int doy = getDoY(day,mon,year);

    // set readback fields for host monitor
    awMem(0x00000,year);
    awMem(0x00004,doy);
    awMem(0x00008,sec);

    sec = sec + 1; // increment for next tic 
    if (sec>=86400) {
      doy = doy+1;
      sec = sec-86400;
    }

    // set 96 bit IRIG-B shift register
    irb2 |= (sec<<8);     // binary seconds field

    dn=36000; for (cn=0; sec>=dn; cn++,sec-=dn); irb0 |= (cn<<23); // tens hours
    dn=3600;  for (cn=0; sec>=dn; cn++,sec-=dn); irb0 |= (cn<<18); // ones hours
    dn=600;   for (cn=0; sec>=dn; cn++,sec-=dn); irb0 |= (cn<<14); // tens minutes
    dn=60;    for (cn=0; sec>=dn; cn++,sec-=dn); irb0 |= (cn<<9);  // ones minutes
    dn=10;    for (cn=0; sec>=dn; cn++,sec-=dn); irb0 |= (cn<<6);  // tens seconds
    cn = sec;          irb0 |= (cn<<1);  // ones seconds

    dn=100;   for (cn=0; doy>=dn; cn++,doy-=dn); irb1 |= (cn<<4);  // hund days
    dn=10;    for (cn=0; doy>=dn; cn++,doy-=dn); irb1 |= (cn<<0);  // tens days
    cn = doy;          irb0 |= (cn<<27); // ones days

    dn=10;  for (cn=0; year>=dn; cn++,year-=dn); irb1 |= (cn<<18); // tens year
    cn = year;           irb1 |= (cn<<13); // ones year

    awBus(0x20000,irb0);
    awBus(0x20004,irb1);
    awBus(0x20008,irb2);
  }
  /** Parse bus for the NMEA start character ($)

      @param adr starting address
      @param len maximum length of characters
      @return starting address of NMEA string
   */
  public static int nextLine ( int adr, int len ) {
    // increment adr (looking for the next line, not this one) and 
    // check for the characters: '$' and '#'
    for ( adr++; (arBus(adr)!=36) && (arBus(adr)!=35) && len>0; adr++, len-- );
    return adr;
  }

  /** Read the Number of Satellites field. 
      The number of satellites can be 1 or 2 digits

      @param iadr starting address (of the field)
      @param nadr ending address (of the field)
      @return number of satellites
   */
  public static int readSatellites ( int iadr, int nadr ) {
    // check if # of satellites is greater than or equal to 10      
    if ( nadr-iadr == 3 ) {
      return 10*arBCD(iadr, 0) + arBCD(iadr, 1);
    }
    if ( nadr-iadr == 2 ) {
      return arBCD(iadr, 0);
    }
    return 0;
  }

  /** Read an NMEA decimal field

      @param iadr starting address (of the field)
      @param nadr ending address (of the field)
      @return integer portion of the decimal
   */
  public static int scaleDecimal ( int iadr, int nadr ) {
    // find the '.' character (char 46)
    int per = findChar(iadr, nadr, 46);
    // determine number of characters
    int nc = per - iadr;

    int deci_;
    if ( per < nadr ) {
      deci_ = arBCD(per, 1);
    }
    else {
      deci_ = 0;
    }

    // return the integer portion of the decimal
    if ( nc > 2 || nc == 0)
      return 1024;
    if ( nc == 2 )
      return 100*arBCD(iadr, 0) + 10*arBCD(iadr, 1) + deci_;
    return 10*arBCD(iadr, 0) + deci_;
  }


  /** Look for a specific character in a stream 

      @param iadr starting address
      @param eind ending address
      @param ac character
      @return address of chracter
   */
  public static int findChar ( int iadr, int eind, int ac ) {
    for ( ; (eind>iadr) && (arBus(iadr)!=ac); iadr++ );
    return iadr;
  }

  public static int arBCD (int adr, int off) {
    int bcd = arBus(adr+off)-48;	// char 48 is a 0
    return bcd;
  }
  public static int nextField (int adr, int cc) {
    for (;cc>0;) {
      if (arBus(adr++)==44) cc--;	// char 44 is a ,
    }
    return adr;
  }

  public static int getDoY (int day, int mon, int year) {
    if (mon>1) day += 31; // add Jan
    if (mon>2) day += 28;
    if (mon>3) day += 31;
    if (mon>4) day += 30;
    if (mon>5) day += 31;
    if (mon>6) day += 30;
    if (mon>7) day += 31;
    if (mon>8) day += 31;
    if (mon>9) day += 30;
    if (mon>10) day += 31;
    if (mon>11) day += 30; // add Nov
    if (mon>2 && (year&0x3)==0) day += 1; // leap year
    return day;
  }

  public static int addBCD (int word, int bcd, int shft) {
    return word | (bcd<<shft);
  }

  protected static void udelay (int usec) {
    for (int i=0; i<usec; i++) {
      for (int j=0; j<16; j++);
    }
  }  

  protected static void awBus (int addr, int data) {
    saddr = addr;
    sdata = data;
  }
  protected static int arBus (int addr) {
    saddr = addr;
    saddr = addr; // make sure address is presented for ample clocks
    return sdata;
  }

  protected static void awMem (int addr, int data) {
    maddr = addr;
    mdata = data;
  }
  protected static int arMem (int addr) {
    maddr = addr;
    return mdata;
  }

}
