package nxm.ice.lib;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;

import nxm.ice.lib.MDevIce;
import nxm.ice.lib.DevIce.DmaMap;
import nxm.sys.lib.Convert;
import nxm.sys.lib.IOResource;
import nxm.sys.lib.Native;
import nxm.sys.lib.Parser;
import nxm.sys.lib.Shell;
import nxm.sys.lib.Table;

/**
  Resource for interacting with a ICE File System

  @author Jeff Schoen
  @version $Id: FileResource.java,v 1.25 2011/02/01 20:41:56 ntn Exp $

  Potential Bugs
    - No known bugs (for now)

 */
public class IceResource extends IOResource {
  private final static int RAID0 = 1;
  private final static int RAMD0 = 2;

  // native function definitions
  private native long alloc       ();
  private native int setConfig    (long handle, String conf);
  private native int setDirection (long handle, int dir);
  private native int init         (long handle);
  private native long writeDet    (long handle, String fname, long off);
  private native long updateDet   (long handle, String fname, long off);
  private native long finalizeDet (long handle, String fname, long wrote);
  private native int nativeRead   (long handle, long buf, int len);
  private native int nativeReadB  (long handle, byte[] buf1, int n1);
  private native int nativeWrite  (long handle, long buf, int len);
  private native int nativeWriteB (long handle, byte[] buf1, int n1);  
  private native int seek         (long handle, long seek);
  private native int setOffset    (long handle, long off);
  private native int free         (long handle);
  private native int launchReads  (long handle, long size);
  private native int joinReads    (long handle);
  private native int joinWrites   (long handle);
  private native static long malloc_buf ( long size );
  private native static int free_buf ( long addr );

  private static boolean useNative;

  /** native function handle */
  private long h;
  /** location of the det files */
  private String rdpath;
  /** device name */
  private String devn;
  /** configuration file */
  private String config;
  /** total size of this file system */
  private long rdsize;
  /** ram offset */
  private long rdaddr;
  /** det offset */
  private long paddr;
  /** the user suggested extension size for writes */
  private long extend;
  /** resource type (i.e. RAID0 or RAMD0) */
  private int res;
  /** current detached header file */
  private File file;
  /** stripe size of the RAID */
  private int stripe_size;
  /** number of disks in the RAID */
  private int ndevs;
  /** number of bytes that the last optimal checker validated */
  private int last_bytes;
  /** number of bytes that the last sub-optimal (RAID) read/write processed */
  private int curr_bytes;
  /** number of bytes that each sub-optimal (RAID) read/write process should request */
  private int rwsize;
  /** current index that the sub-optimal (RAID) read/write is working on */
  private int bind;
  /** number of bytes that remain after a seek is requested */
  private int seek_rem;
  /** array of offsets that are in the detached header file */
  private long[] offs;
  /** array of sizes that are in the detached header file */ 
  private long[] sizes;
  /** current detached header file index */
  private int det_ind;
  /** track if we have optimal transfer lengths */
  private boolean optimal;
  /** track if this is the first read */
  private boolean first_read;
  /** byte array for sub-optimal read/writes */
  private byte[] b;
  /** handle to the pic card */
  private MDevIce pic;
  /** Direct Memory Access map */
  private DmaMap map;
  /** current DMA map offset */
  private long mapOff;
  /** transfer length size */
  private int xfer;
  /** total length of the file (length holds current det entry len)*/
  private long len_tot;
  /** write length */
  private long wlength;
  /** read length */
  private long rlength;
  /** using native implementation */
  private boolean isNative;

  /** load the native libraries */
  public static boolean load() {
    if (!useNative) useNative=Shell.loadLibrary("nxm.ice.lib.IceResource"); 
    return useNative;
  }

  public IceResource() {
    super();
    length = rlength = wlength = len_tot = offset = extend = rdsize = seek_rem = 0;
    rwsize = stripe_size = ndevs = 1;
    config = devn = rdpath = null;
    optimal = first_read = true;
    sizes = offs = null;
    b = null;
    paddr = bind = curr_bytes = res = last_bytes = xfer = -1;
    isOpen = isNative = false;
    load();
  }

  public boolean exists () { return file.exists(); }
  public boolean isFile () { return true; }
  public boolean delete () { return file.delete(); }

  public void setURL (String str) {
    super.setURL(str);
    if (str.startsWith("ice")) 
      str = str.substring(4);
    parseAux();
    parseConf();
    parseDevFile();
    file = new File(str);
    setExtend(extend);
  }

  private boolean parseDevFile () {
    // choose the implementation
    if (res==RAID0) {
      try {
        FileReader fr = new FileReader(devn);
        BufferedReader br = new BufferedReader(fr);
        String line = br.readLine();
        String[] fields = line.split(" ");
        // the full device size is in the 2nd field
        rdsize = (Long.parseLong(fields[1]) * 512);
        // number of devices is in the 4th field
        ndevs = Integer.parseInt(fields[3]);
        // stripe size is the 5th field
        stripe_size = Integer.parseInt(fields[4]) * 512;
        rwsize = stripe_size * ndevs;
        br.close();
        fr.close();
        return true;
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    else if ( res==RAMD0) {
      try {
        int beg, end;
        FileReader fr = new FileReader(devn);
        BufferedReader br = new BufferedReader(fr);
        // read the start
        String line = br.readLine();
        beg = line.indexOf(":");
        end = line.indexOf(" Mby");
        if ( beg<0 || end<0) {
          M.error("Trouble parsing RAM properties");
          return false;
        }
        rdaddr = Convert.s2xl(line.substring(beg+2, end)+"M");
        // now read the the size
        line = br.readLine();
        beg = line.indexOf(":");
        end = line.indexOf(" Mby");
        if ( beg<0 || end<0) {
          M.error("Trouble parsing RAM properties");
          return false;
        }
        rdsize = Convert.s2xl(line.substring(beg+2, end)+"M");
        br.close();
        fr.close();
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    return false;
  }

  /** 
   Setter for extend size.
   @param ext suggested size for extension
   */
  public void setExtend (long ext) {
    extend = ext; 
    if (res==RAMD0) {
      long extent = 0x01000000-1;
      extend = extend & ~extent;
    }      
  }

  public void setLength ( long len ) {
    length = len;
    rlength = len;
    wlength = rwsize*(long)Math.ceil(1.0*len/rwsize);
  }


  public boolean open () {
    if ( urlstr.endsWith("/") )
      return false;
    isOpen = false;
    h = alloc();
    if ( h <= 0L )
      M.error("Problem performing alloc()");
    if ( setConfig(h, config) < 0 )
      M.error("Problem setting config file");
    // input
    if (dir<=0) {
      paddr = readDet();
      setLength(sizes[det_ind]);
    }
    // output
    else {
      paddr = writeDet();     
    }
    if (res==RAID0) {
      if ( dir == IOResource.DIR_INOUT ) {
        // INOUT for headermod
        isOpen = false;
        return true;
      }
      if ( setDirection(h, dir) < 0 )
        M.error("Problem setting direction");
      if ( init(h) < 0 )
        M.error("Problem initializing");
      if ( setOffset(h, paddr) < 0 )
        M.error("Problem setting offset");
      isOpen = true;
    }
    else if (res==RAMD0) {
      pic = new MDevIce(M,"ICEPIC,DEVNO=0,IOM=NONE,PM=NONE");
      if (rdaddr==0 && rdsize!=0) {
        pic.open();
        rdaddr = pic.phys2long( pic.getKeyL(0, MDevIce.KEY_QALLOC) );
        pic.close();
      }
      map = pic.getMap();
      map.mode  = DmaMap.RAMD;
      map.bytes = length;
      map.paddr = getPhysicalAddress();
      map.dev   = pic;
      map.devn  = null;
      map.setPaged(Shell.ostag.startsWith("lin"));
      isOpen = true;
    }
    return isOpen;
  }

  private long readDet () {
    // make sure the detached header exists and is the correct size
    if ( !file.exists() || (file.length()%16)!=0  ) 
      M.error("Detached header file is invalid");
    int len = (int) (file.length()/16);
    det_ind = 0;
    offs = new long[len];
    sizes = new long[len];
    byte[] buf = new byte[16];
    // open, read and close the detached header
    int fd = Native.fopen(file.toString(), DIR_INPUT);
    len_tot = 0;
    for (int m=0; m<len; m++) {
      Native.freadb(fd, buf, 0, 16);
      offs[m] = Convert.unpackX(buf,0);
      sizes[m] = Convert.unpackX(buf,8);
      len_tot += sizes[m];
    }    
    Native.fclose(fd);
    return offs[det_ind];
  }

  private long writeDet () {
    if (res==RAMD0) {
      int PGSZ=8192;
      // calculate the size
      setLength((length+PGSZ-1)/PGSZ*PGSZ);
    }
    long off = writeDet(h, file.toString(), wlength);
    if ( off < 0 )
      M.error("Could not allocate "+wlength+" bytes");
    return off;
  }

  private long updateDet () {
    long rv = updateDet(h, file.toString(), extend);
    if ( rv < 0 )
      M.error("Could not allocate an additional "+extend+" bytes");
    setLength( wlength + extend );
    last_bytes = -1;
    return rv;
  }

  private void remap (long len) {
    map = null;
    map = pic.getMap();
    map.mode  = DmaMap.RAMD;
    map.bytes = len;
    map.paddr = getPhysicalAddress();
    map.dev   = pic;
    map.devn  = null;
    map.setPaged(Shell.ostag.startsWith("lin"));
    mapOff = 0;
  } 

  private long getPhysicalAddress() {
    return rdaddr+paddr;
  }

  public boolean close () {
    if ( isOpen ) {
      if ( dir == DIR_OUTPUT ) {
        // write the last segment (if necessary)
        if ( bind > 0 && !optimal ) {
          offset += nativeWriteB(h, b, bind);
        }
        // finalize detached header size size
        setLength(offset);
        finalizeDet(h, file.toString(), offset);
      }
      if (map!=null) map.close(DmaMap.NOCHK);
    }
    // close the native
    free(h);    
    isOpen = false;
    return true;
  }

  public int read ( byte[] buf, int boff, int bin ) {
    if (res==RAID0) {
      if (first_read) {
        first_read = false;
        if ( launchReads(h, sizes[det_ind]) < 0 )
          M.error("Problem launching threads");
      }
      // if we are stripe aligned then we can just write directly
      if ( isOptimal(bin) ) {
        curr_bytes = processRead(buf, bin);
        offset += curr_bytes;
        return curr_bytes;
      }
      else {
        // copy a piece of the buffer
        int cw = (bind+bin<curr_bytes) ? bin : curr_bytes-bind;
        System.arraycopy(b, bind, buf, 0, cw);
        bind += cw;
        offset += cw;
        // check if we need to perform additional read(s)
        while ( cw < bin ) {
          curr_bytes = processRead(b, rwsize);
          bind = (curr_bytes+cw<bin) ? curr_bytes : bin-cw;
          System.arraycopy(b, 0, buf, cw, bind);
          cw += bind;
          offset += bind;
        }
        return cw;
      }
    }
    else if (res==RAMD0) {
      long vaddr;
      int bp = 0;
      int bytes;
      while ( bp < bin ) {
        bytes = (int) Math.min(bin-bp, rlength-offset);
        // look for a re-mapping
        if ( bytes==0 ) {
          // extend and re-map
          det_ind++;
          paddr = offs[det_ind];
          setLength(rlength+sizes[det_ind]);
          remap(sizes[det_ind]);
          // get new size and address
          bytes = (int) Math.min(bin-bp, rlength-offset);
        }
        vaddr = map.getVirtualAddress(mapOff, bytes);
        Native.p2ja (vaddr, 0, buf, boff+bp, bytes);
        bp += bytes;
        offset += bytes;
        mapOff += bytes;
      } 
      return bp;
    }
    return -1;
  }

  public int read ( long lbuf, int boff, int bin ) {
    if (res==RAID0) {
      if ( xfer < 0 ) {
        xfer = bin;
        isNative = true; 
        // make sure this xfer length is acceptable
        if ( (xfer%stripe_size)!=0 && xfer<len_tot )
          M.error("Read transfer lengths must be evenly disivible by " + stripe_size);
        else if ( xfer > extend )
          M.error("Transfer length ("+xfer+") cannot be greater than extend size ("+extend+")");
        else if ( seek_rem != 0 )
          M.error("Cannot perform a seek not aligned to "+stripe_size+" in native read");
        optimal = true;
        if ( launchReads(h, sizes[det_ind]) < 0 )
          M.error("Problem launching threads");
      }
      // determine how many bytes we can safely process 
      int bytes = (int) Math.min(bin, rlength-offset);
      if ( bytes==0 && (det_ind+1)<offs.length ) {
        // join the reads
        if ( joinReads(h) < 0 )
          M.error("Problem joining threads");
        // increment det index
        det_ind++;
        // extend length
        setLength( rlength + sizes[det_ind] );
        // perform native seek as we just changed the det index
        if ( setOffset(h, offs[det_ind]) < 0 )
          M.error("Problem setting offset");
        if ( launchReads(h, sizes[det_ind]) < 0 )
          M.error("Problem launching threads");
        // calculate how many bytes we can safely process 
        bytes = (int) Math.min(bin, rlength-offset);
      }
      // process the data
      int rv = nativeRead(h, lbuf, bytes);
      if ( rv != bytes ) 
        M.error("I/O error");
      // update offset and return
      offset += rv;
      return rv;
    }
    else if (res==RAMD0) {
      long vaddr;
      int bp = 0;
      int bytes;
      while ( bp < bin ) {
        bytes = (int) Math.min(bin-bp, length-offset);
        // look for a re-mapping
        if ( bytes==0 ) {
          // extend and re-map
          det_ind++;
          paddr = offs[det_ind];
          setLength(length+sizes[det_ind]);
          remap(sizes[det_ind]);
          // get new size and address
          bytes = (int) Math.min(bin-bp, length-offset);
        }
        vaddr = map.getVirtualAddress(mapOff, bytes);
        Native.p2p (vaddr, 0, lbuf, boff+bp, bytes);
        bp += bytes;
        offset += bytes;
        mapOff += bytes;
      } 
      return bp;
    }
    return -1;
  }

  public int write ( long lbuf, int boff, int bin ) {
    if ( res==RAID0 ) {      
      if ( xfer < 0 ) {
        xfer = bin;
        isNative = true;
        // make sure buffer size is appropriate
        if ( xfer%stripe_size != 0 && xfer<length)
          M.error("Write transfer length (" + xfer + ") must be an even multiple of " + stripe_size);
        else if ( xfer > extend ) 
          M.error("Transfer length ("+xfer+") cannot be greater than extend size ("+extend+")");
        else if ( seek_rem != 0 )
          M.error("Cannot perform a seek not aligned to "+stripe_size+" in native write");
        optimal = true;
        // align the user specified extend size
        extend = bin * Math.round ( (double)extend/bin );
      }
      // determine how many bytes we can safely process
      int bytes = (int) Math.min(bin, wlength-offset);
      // check if we need to extend our file
      if ( bytes == 0 ) {
        // wait on the write threads
        if ( joinWrites(h) < 0 )
          M.error("Trouble joining threads");
        long seek = updateDet();
        // call native seek
        if ( setOffset(h, seek) < 0 )
          M.error("Problem setting offset");
        bytes = (int) Math.min(bin, wlength-offset);
      }
      else if ( bytes!=xfer && (bytes%stripe_size)!=0 && (bytes+offset)!=length )
        M.error("IceResource alignment problem");
      // process the data
      int rv = nativeWrite(h, lbuf, bytes);
      if ( rv != bytes ) 
        M.error("I/O problem");
      offset += rv;
      return rv;
    }
    else if (res==RAMD0) {      
      long vaddr;
      int bp = 0;
      int bytes;
      while ( bp < bin ) {
        bytes = (int) Math.min(bin-bp, wlength-offset);
        // look for a re-mapping
        if ( bytes==0 ) {
          // extend and re-map
          paddr = updateDet();
          remap(extend);
          // get new size and address
          bytes = (int) Math.min(bin-bp, wlength-offset);
        }
        vaddr = map.getVirtualAddress(mapOff, bytes);
        Native.p2p(lbuf, boff+bp, vaddr, 0, bytes);
        bp += bytes;
        offset += bytes;
        mapOff += bytes;
      }
      return bp;
    }
    return -1;
  }

  /*
   */
  public int write ( byte[] buf, int boff, int bin ) {
    if (res==RAID0) {
      // if we are stripe aligned then we can just write directly
      if ( isOptimal(bin) ) {
        curr_bytes = processWrite(buf, bin);
        return curr_bytes;
      }
      else {
        int cw = ((bind+bin)<rwsize) ? bin : rwsize-bind;
        System.arraycopy(buf, 0, b, bind, cw);
        bind += cw;
        // check if we need to perform additional read(s)
        while ( cw < bin ) {
          curr_bytes = processWrite(b, rwsize);
          // finish off this buffer
          bind = ((cw+rwsize)<bin) ? rwsize : bin-cw; 
          System.arraycopy(buf, cw, b, 0, bind);
          cw += bind;
        }
        return cw;
      }
    }
    else if (res==RAMD0) {
      long vaddr;
      int bp = 0;
      int bytes;
      while ( bp < bin ) {
        bytes = (int) Math.min(bin-bp, wlength-offset);
        // look for a re-mapping
        if ( bytes==0 ) {
          // extend and re-map
          paddr = updateDet();
          remap(extend);
          // get new size and address
          bytes = (int) Math.min(bin-bp, wlength-offset);
        }
        vaddr = map.getVirtualAddress(mapOff, bytes);
        Native.ja2p (buf, boff+bp, vaddr, 0, bytes);
        bp += bytes;
        offset += bytes;
        mapOff += bytes;
      } 
      return bp;
    }
    return -1;
  }

  /** seek to the specified offset in file */
  public boolean seek ( long soff ) {
    if ( soff == offset )
      return true;
    if (res==RAID0) {
      if ( dir==DIR_INPUT ) {
        if ( soff == 0 ) {
          // assume they are looping and perform a soft reset
          if ( len_tot%stripe_size != 0 && isNative ) {
            // total length must be evenly divisble if native
            M.error("File size must be evenly divisible by " + stripe_size + " for looped playback");
          }
          // perform a soft reset
          wlength = length = offset = seek_rem = 0;
          //seek_rem = 0;
          optimal = true;
          sizes = offs = null;
          b = null;
          paddr = bind = curr_bytes = last_bytes = -1;
          // close this 
          close();
          // re-open it
          if ( !open () ) 
            return false;
          // re-launch threads
          return ( launchReads(h, sizes[det_ind]) >= 0 ) ? true : false;
        }
        // figure out which entry into the det the desired seek runs into
        if ( soff >= length ) {
          // increment the index until the length is greater than the desired off
          while ( soff>=length && det_ind<sizes.length ) {
            offset += sizes[det_ind];
            det_ind++;
            setLength( length+sizes[det_ind] );
          }
          // set new offset
          if ( setOffset(h, offs[det_ind]) < 0 )
            M.error("Problem setting offset");
        }
        // determine how much seeking is required into this det index
        long delta = soff-offset;
        //seek_rem = icefs.seek(del);
        seek_rem = seek(h, delta);
        // alter the size so when read threads are launched the correct data
        // is read into the queue's
        sizes[det_ind] -= (delta-seek_rem);
        // finally set the offset to what it is
        offset = soff - seek_rem;
        return true;
      }
      else {
        offset = soff;
        return true;
      }
    }
    else if (res==RAMD0) {
      if ( dir==DIR_INPUT ) {
        if ( soff >= length ) {
          // increment the index until the length is greater than the desired off
          while ( soff>=length && det_ind<sizes.length ) {
            det_ind++;
            setLength( length+sizes[det_ind] );
          }
        }
      }
      // adjust the address based on what remains
      paddr = offs[det_ind] + (soff - (length-sizes[det_ind]));
      remap(sizes[det_ind]);
      offset = soff;
      return true; 
    }
    return false;
  }

  public Table getStatFS () {
    File tmp;
    String[] lf = M.io.listFiles(rdpath,"*","det");
    int files = lf.length;
    long total = 0;
    byte[] buf = new byte[16];
    int fd, ret;
    long psizei;
    for (int i=0; i<files; i++) {
      tmp = new File(rdpath+lf[i]);
      // make sure that the size is valid
      if ( tmp.length() % 16 == 0 ) {
        fd = Native.fopen (tmp.toString(),-1);
        for ( ret=Native.freadb(fd,buf,0,16); ret==16; ret=Native.freadb(fd,buf,0,16) ) {
          psizei = Convert.unpackX(buf,8);
          total += psizei;
        }
        Native.fclose(fd);
      }
    }
    long tmby = (rdsize/1000000);
    long fmby = tmby - (total/1000000);
    int used = (int) (100-100*fmby/tmby);
    String stats = "{PATH="+rdpath+",FILES="+files+",TMBY="+tmby+
        ",FMBY="+fmby+",USED="+used+"}";
    return new Table(stats);
  }

  private boolean parseAux () {
    // get the path
    String path = urlstr.substring(0, urlstr.lastIndexOf("/"));
    Table auxtbl = (Table) M.results.get("AUX");
    String auxstr = null;
    for (Table.Iterator ti=auxtbl.iterator(); auxstr==null && ti.getNext(); ) {
      String s = ti.value.toString();
      if (s.startsWith(path))
        auxstr = s;
    }
    if (auxstr==null) 
      M.error("AUX entry "+path+" is not found");
    Parser p = new Parser(auxstr);
    config = p.get(2);
    return true;
  }

  private boolean parseConf () {
    res = -1;
    extend = -1;
    rdpath = devn = "";
    int bit_mask = 0;
    try {
      FileReader fr = new FileReader(config);
      BufferedReader br = new BufferedReader(fr);
      String cur;
      while ( (cur = br.readLine()) != null ) {       
        if ( cur.toUpperCase().startsWith("TYPE") ) {
          res = cur.contains("RAID") ? RAID0 : cur.contains("RAMD") ? RAMD0 : -1;
        }
        else if ( cur.toUpperCase().startsWith("NMPATH") ) {
          rdpath = cur.substring(7) + "/";
          bit_mask |= 0x1;
        }
        else if ( cur.toUpperCase().startsWith("DMSETUP") ) {
          devn = cur.substring(8);
          bit_mask |= 0x2;
        }
        else if ( cur.toUpperCase().startsWith("EXTEND") ) {
          extend = Convert.s2xl(cur.substring(7));
          bit_mask |= 0x4;
        }       
      }
      // make sure everything is set!
      if ( res == -1 || bit_mask != 0x7 )
        M.error("Incorrect config file. Refer to nxm.data.icefs1.conf for an example config file");

      // make sure the Ram Disk extend is at least 16M
      if ( res==RAMD0 && extend<16777216 )
        extend = 16777216;

      br.close();
      fr.close();
      return true;
    } catch ( Exception e ) {
      return false;
    } 
  }

  /** Return a printable string containing information about this object. */
  public String toString () { return "IceResource : " + getURL(); }

  private boolean isOptimal (int bytes) {
    // if the bytes haven't changed or if we were ever sub-optimal return
    if ( !optimal || bytes == last_bytes )
      return optimal;
    last_bytes = bytes;
    optimal = false;
    if ( bytes%rwsize==0 || (bytes+offset)==getTotalLen() ) {
      optimal = true;
      return true;
    }
    M.warning("/TL=" + bytes + " is sub-optimal (multiples of "+stripe_size+" are optimal)");
    curr_bytes = rwsize;
    if ( dir==DIR_INPUT ) 
      bind = rwsize;
    else 
      bind = 0;
    b = new byte[rwsize];
    return false;
  }

  private int processRead ( byte[] buf, int bin ) {
    // calculate the number of bytes 
    int bytes = (int) Math.min(bin, rlength-offset);
    // check for an detached index increment
    if ( bytes==0 ) {
      // join the reads
      if ( joinReads(h) < 0 )
        M.error("Problem joining threads");
      // increment det index
      det_ind++;
      last_bytes = -1;
      if ( det_ind < sizes.length ) {
        setLength( rlength + sizes[det_ind] );
        if ( setOffset(h, offs[det_ind]) < 0 )
          M.error("Problem setting offset");
        if ( launchReads(h, sizes[det_ind]) < 0 )
          M.error("Problem launching threads");
        bytes = (int) Math.min(bin, rlength-offset);
      }
    }
    // read the data
    int rv = nativeReadB(h, buf, bytes);
    // check if we need to finish our seek
    if ( seek_rem > 0 ) {
      if (seek_rem >= rv ) {
        offset += rv;
        seek_rem -= rv;
        rv = 0;
      }
      else {
        // finish our seek
        System.arraycopy(buf, seek_rem, buf, 0, (rv-seek_rem));
        offset += seek_rem;
        rv -= seek_rem;
        seek_rem = 0;
      }
    }
    return rv;
  }

  private int processWrite (byte[] buf, int bin) {
    // calculate the number of bytes 
    int bytes = (int) Math.min(bin, wlength-offset);
    // check for an extension
    if ( bytes==0 ) {
      // align the extend size
      extend = bin * Math.round ( (double)extend/bin );
      if ( joinWrites(h) < 0 )
        M.error("Trouble joining threads");
      long seek = updateDet();
      last_bytes = -1;
      if ( setOffset(h, seek) < 0 )
        M.error("Problem setting offset");
      bytes = (int) Math.min(bin, wlength-offset);
    }
    // write the data
    int rv = nativeWriteB(h, buf, bytes);
    offset += rv;
    return rv;
  }

  /**
    Getter for size of the array
    @return total array size in bytes
   */
  public long getDeviceSize() { return rdsize; }

  /** 
    Getter for AUX directory
    @return full path to directory of this AUX
   */
  public String getPath () { return rdpath; }

  /** 
   Stripe size getter 
   * @return size of the stripe in bytes
   */
  public int getStripeSize () { return stripe_size; }

  /** 
   Getter for number of disks
    @return number of disk in the array
   */
  public int getNumDevs () { return ndevs; }

  /** 
   Getter for total file size
   @return size of file in bytes
   */
  public long getTotalLen() {
    if ( sizes==null ) {
      return length;
    }
    long accum = sizes[0];
    for (int m=1; m<sizes.length; m++)
      accum += sizes[m];
    return accum;
  }

  /** 
    Allocates a block of memory 
    @param size number of bytes to allocate
    @return pointer to the beginning of the allocated block of memory
   */
  public static long malloc ( long size ) {
    if (load()) return malloc_buf(size);
    System.out.printf("Problem opening library!!!!!!!!!!!\n");
    return -1;
  }

  /**
   Frees a block memory
    @param ptr pointer to the starting address to free
    @return 0 for success
   */
  public static int free_mem ( long ptr ) {
    return free_buf(ptr);
  }

}
