package nxm.ice.lib;

import nxm.sys.inc.PacketHandler;
import nxm.sys.inc.DataTypes;
import nxm.sys.lib.*;

/**
  Handler for VITA-49 Radio Transport packets.

  <pre>

  VRT Packet Word Placement

    0 HDR Header (Type:4,C:1,T:1,RR:2,TSI:2,TSF:2,Seq:4,Size:16)
    1 SID Stream ID 
    2 OUID Organizational Unique ID
    3 CID  Information Class : Packet Class
    4 TSIS TimeStamp - Integer Seconds
  6:5 TSFS TimeStamp - Fractional Seconds
    7 TLR Trailer
    
    X Data Payload (X=360 qwords for 1500MTU)

    Packet qword 7 moved after payload at transmit

  Header Types:
    0cxs c=Context x=Extension s=WithStreamID
    
  Packet Class (Pack:1,Type:2,Format:5,Resv:2,Size:6)

    Data Item Size = (bitsPerSample-1)

    Data Item Type
    00 Real
    01 Complex Cartesian
    10 Complex Polar

    Data Item Format
    u0000 Fixed Point - u=Unsigned|Signed
    u0eee VRT float w/ e-bit exponent
    01110 IEEE Single Precision Floating Point
    01111 IEEE Double Precision Floating Point

  Information Class


  TimeStamp
    TSI 0=None 1=UTC 2=GPS 3=Other
    TSF 0=None 1=Count 2=pSec 3=FreeRunningCount
  
  </pre>

  @author Jeff Schoen
  @version $Id: VRTPacket.java,v 1.28 2012/10/15 20:32:56 jgs Exp $

*/
public class VRTPacket implements PacketHandler,Cloneable {

  private static int O_HDR=0, O_SID=4, O_OUID=8, O_CID=12, O_TSIS=16, O_TSFS=20, O_TAIL=28;

  public static int OUID_ICE = 0x00104D77;

  public static int TSI_NONE=0, TSI_UTC=1, TSI_GPS=2, TSI_OTHER=3;
  public static int TSF_NONE=0, TSF_CNT=1, TSF_PSEC=2, TSF_FRC=3;

  public static String flagsList = "TRIM,BIGE";
  public static int FLG_NONE=0x0,FLG_TRIM=0x1,FLG_BIGE=0x2;

  public  byte[] buf = new byte[32];

  private int headerOffset=0, headerLength=32, trailerOffset=28, trailerLength=4, fixedBytes=1440;
  private int hdr, cid, size, bits, fbits, count, ipoff, rhdrs, whdrs, flags;
  private int dataLength, totalLength, fdataLength;
  private String format;

  byte rep = DataTypes.EEEI;
  private boolean isInput=false;
  private double npktoff;

  public VRTPacket() {
    this("SB",0);
  }

  public VRTPacket (String format, int flags) {
    if ((flags&FLG_BIGE)!=0) setRep(DataTypes.IEEE);
    // normal 1440 byte data packet w/ streamID, orgID, classID, TimeCode, and Trailer fields / SB data
    setInternals(0x1CF00170,0x7); 
    setCount(0);
    setSID(0);
    setOUID(OUID_ICE);
    setTC(0.0,0.0);
    setFormat(format);
    setFlags(flags);
    setBytes(fixedBytes);
  }

  /** init configuration from first packet header */
  private void initFrom (DataFile df) {
    setRep((byte)df.getDataRep().charAt(0));
    if (df.hp!=null) {
      df.hp.seek((long)df.getDataStart());
      df.hp.read (this.buf,headerOffset,headerLength);
      df.hp.seek((long)df.getDataStart());
    } else {
      df.io.seek((long)df.getDataStart());
      df.io.read (this.buf,headerOffset,headerLength);
      df.io.seek((long)df.getDataStart());
    }
    setInternals( Convert.unpackL(buf,O_HDR,rep), Convert.unpackL(buf,O_CID,rep) );
  }

  /** sets the data payload representation */
  public void setFlags (int flags) {
    this.flags = flags;
    if ((flags&FLG_TRIM)!=0) setInternals(hdr&~0x04000000,cid);
  }

  /** sets the local parameters to match header */
  private void setInternals (int hdr, int cid) {
    this.hdr = hdr;
    this.cid = cid;
    bits = (cid&0x3F)+1;
    if ((hdr&0x04000000)!=0) {	// has trailer
      headerLength = 32;
      trailerLength = 4;
    } else {			// no trailer
      headerLength = 28;
      trailerLength = 0;
    }
    totalLength = (hdr&0xFFFF)<<2;
    dataLength = totalLength-headerLength;
    fdataLength = fbits*dataLength/bits;
    size = dataLength*8/bits;
  }

  /** sets the data payload representation */
  public void setRep (byte rep) {
    this.rep = rep;
  }
  /** get representation of data payload */
  public byte getRep() {
    return rep;
  }

  /** sets the data payload size int elements */
  public void setSize (int elem) {
    this.size = elem;
    setBytes(elem*bits/8);
  }
  /** get Size of data payload in elements */
  public int getSize() {
    return size;
  }

  /** sets the data payload size in bytes */
  public void setBytes (int bytes) {
    size = bytes*8/bits;
    dataLength = bytes;
    fdataLength = fbits*dataLength/bits;
    totalLength = headerLength+dataLength;
    int pktsize=(totalLength+3)>>2;
    hdr = (hdr&0xFFFF0000) | pktsize;
    Convert.packL(buf,O_HDR,hdr,rep);
  }
  /** get Size of data payload in bytes */
  public int getBytes() {
    return dataLength;
  }

  /** sets Stream ID */
  public void setSID (int sid) {
    Convert.packL(buf,O_SID,sid,rep);
  }
  /** gets Stream ID */
  public int getSID() {
    return Convert.unpackL(buf,O_SID,rep);
  }

  /** sets Organizational Unique ID */
  public void setOUID (int ouid) {
    Convert.packL(buf,O_OUID,ouid,rep);
  }
  /** gets Organizational Unique ID */
  public int getOUID() {
    return Convert.unpackL(buf,O_OUID,rep);
  }

  /** sets Information Class ID */
  public void setICID (int icid) {
    setCID( (cid&0xFFFF) | (icid<<16) );
  }
  /** gets Information Class ID */
  public int getICID() {
    return (cid>>16)&0xFFFF;
  }

  /** sets Class ID */
  public void setCID (int cid) {
    this.cid = cid;
    Convert.packL(buf,O_CID,cid,rep);
  }
  /** gets Class ID */
  public int getCID() {
    return cid;
  }

  /** packet number since start of xfer */
  public void setCount (int count) {
    this.count = count;
    hdr = (hdr&0xFFF0FFFF) | ((count&0xF)<<16);
    Convert.packL(buf,O_HDR,hdr,rep);
  }
  /** get the current packet count */
  public int getCount() {
    if (isInput) {
      int hdr = Convert.unpackL(buf,O_HDR,rep);
      count = (hdr>>16)&0xF;
    }
    return count;
  }
  /** up the count by one */
  public void upCount() {
    count++;
    setCount(count);
  }

  /** set format/bits from Midas format digraph */
  public void setFormat (String format) {
    if (format.equals(this.format)) return;
    this.format = format;
    if (!DataFile.checkFormat(format,"S#,C#")) System.out.println("Unallowed format="+format);
    char mode=format.charAt(0), type=format.charAt(1);
    int bps = Data.getBPS(type);
    fbits = bits = (bps<0)? -bps : bps*8;
    int fb = (bits-1);
    int tb = 0x00; // assume fixed point
    if (type=='F') tb = 0x0E;
    if (type=='D') tb = 0x0F;
    if (mode=='C') tb |= 0x20;
    cid = (cid&0xFFFF0000) | (tb<<8) | (fb&0x3F);
    Convert.packL(buf,O_CID,cid,rep);
  }
  /** get the current Format */
  public String getFormat() {
    return format;
  }

  /** set number of bits per sample */
  public void setBits (int bits) {
    this.bits = bits;
    int fb = (bits-1);
    cid = (cid&0xFFFFFF00) | (fb&0x3F);
    Convert.packL(buf,O_CID,cid,rep);
  }
  /** get the number of bits per sample */
  public int getBits() {
    return bits;
  }

  /** set timecode */
  public void setTC (double wsec, double fsec) {
    int  isec = (int)wsec;
    long psec = (long)(fsec*1e12+.5);
    if (psec<0) { isec--; psec+=1e12; }
    Convert.packL(buf,O_TSIS,isec,rep); 
    Convert.packX(buf,O_TSFS,psec,rep); 
    hdr = (hdr&0xFF0FFFFF) | (TSI_UTC<<22) | (TSF_PSEC<<20);
    Convert.packL(buf,O_HDR,hdr,rep);
  }

  /** get timecode from ICE/VITA formatted header */
  public Time getTC() {
    int  isec = Convert.unpackL(buf,O_TSIS,rep); 
    long psec = Convert.unpackX(buf,O_TSFS,rep); 
    double wsec = isec;
    double fsec = psec*1e-12;
    return new Time (wsec,fsec);
  }

/*
  BE Packing Modes
     |       0       |       1       |       2       |       3       |
 16  |F E D C B A 9 8 7 6 5 4 3 2 1 0|F E D C B A 9 8 7 6 5 4 3 2 1 0|
 12  |F E D C B A 9 8 7 6 5 4|F E D C B A 9 8 7 6 5 4|F E D C B A 9 8
  8  |F E D C B A 9 8|F E D C B A 9 8|F E D C B A 9 8|F E D C B A 9 8|
  6  |7 6 5 4 3 2|7 6 5 4 3 2|7 6 5 4 3 2|7 6 5 4 3 2|7 6 5 4 3 2|7 6 
  4  |7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|


  LE Packing Modes
     |       3       |       2       |       1       |       0       |
 16  |F E D C B A 9 8 7 6 5 4 3 2 1 0|F E D C B A 9 8 7 6 5 4 3 2 1 0|
 12   B A 9 8 7 6 5 4|F E D C B A 9 8 7 6 5 4|F E D C B A 9 8 7 6 5 4|
  8  |F E D C B A 9 8|F E D C B A 9 8|F E D C B A 9 8|F E D C B A 9 8|
  6   3 2|7 6 5 4 3 2|7 6 5 4 3 2|7 6 5 4 3 2|7 6 5 4 3 2|7 6 5 4 3 2|
  4  |7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|7 6 5 4|

*/

  /** pack to different word sizes */
  private void pack (byte[] buf, int boff, int n, int ibits, int obits) {
    int i,j,iadr=0,oadr=0,ibit,obit,ibyt,obyt,diff=ibits-obits; // > 0
    if (rep==DataTypes.IEEE) {
      for (i=0; i<n; i++) {
        for (j=0, iadr=i*ibits, oadr=i*obits; j<obits; j++,iadr++,oadr++) {
          ibyt=iadr>>3; ibit=(~iadr)&0x7; 
          obyt=oadr>>3; obit=(~oadr)&0x7;
          if ((buf[ibyt]&(1<<ibit))==0) buf[obyt]&=~(1<<obit); else buf[obyt]|=(1<<obit); 
        }
      }
    } else {
      for (i=0; i<n; i++) {
        for (j=0, iadr=i*ibits+diff, oadr=i*obits; j<obits; j++,iadr++,oadr++) {
          ibyt=iadr>>3; ibit=iadr&0x7; 
          obyt=oadr>>3; obit=oadr&0x7;
          if ((buf[ibyt]&(1<<ibit))==0) buf[obyt]&=~(1<<obit); else buf[obyt]|=(1<<obit); 
        }
      }
    }
  }

  /** unpack to different word sizes */
  private void unpack (byte[] buf, int boff, int n, int ibits, int obits) {
    int i,j,iadr,oadr,ibit,obit,ibyt,obyt,diff=ibits-obits; // < 0
    if (rep==DataTypes.IEEE) {
      for (i=n; i>0; i--) {
        for (j=0, iadr=i*ibits-1-diff, oadr=i*obits-1; j<obits; j++,iadr--,oadr--) {
          ibyt=iadr>>3; ibit=(~iadr)&0x7;
          obyt=oadr>>3; obit=(~oadr)&0x7;
          if (j+diff<0 || (buf[ibyt]&(1<<ibit))==0) buf[obyt]&=~(1<<obit); else buf[obyt]|=(1<<obit); 
        }
      }
    } else {
      for (i=n; i>0; i--) {
        for (j=obits-1, iadr=i*ibits-1, oadr=i*obits-1; j>=0; j--,iadr--,oadr--) {
          ibyt=iadr>>3; ibit=iadr&0x7;
          obyt=oadr>>3; obit=oadr&0x7;
          if (j+diff<0 || (buf[ibyt]&(1<<ibit))==0) buf[obyt]&=~(1<<obit); else buf[obyt]|=(1<<obit); 
        }
      }
    }
  }

  /** {@inheritDoc} */
  public String getConfiguration (DataFile df) {
    String config = "VRT";
    if (df.hp!=null) {
      config += "/DET";
      if (df.hp.aux!=null && df.hp.aux.length()>0) config += "="+df.hp.aux;
    }
    return config;
  }

  /** {@inheritDoc} */
  public void setFileName (DataFile df, FileName filename) {
    DataFile hp = new DataFile();
    hp.init(null,filename,"3000","NH",df.getPacketHandlerFlags());
    String srec="HDR|SL,SID|SL,OUID|SL,PIC|SL,TCIS|SL,TCFS|SX,TLR|SL";
    hp.setSubRecords(srec);
    if (rep==DataTypes.IEEE) hp.setDataRep("IEEE");
    if (rep==DataTypes.EEEI) hp.setDataRep("EEEI");
    df.hp=hp;
  }

  /** {@inheritDoc} */
  public double naturalDataOffset (DataFile df, double boffset) {
    int bpp = fdataLength;
    double npkt = Math.floor(boffset/bpp);
    return npkt*bpp;
  }

  /** {@inheritDoc} */
  public double dataToPacketOffset (DataFile df, double boffset) {
    double npkt = Math.floor(boffset/fdataLength);
    double bpkthdr = npkt*headerLength;
    if (bits!=fbits) boffset = boffset*bits/fbits;
    ipoff = (int)(boffset-npkt*dataLength); 			// save offset into packet
    if (df.hp==null) {
      boffset += bpkthdr;
      if (ipoff>0) boffset += (headerLength-trailerLength);	// skip header
    } else {
      if (ipoff>0) npkt++;  					// skip header
      npktoff = npkt;
    }
    return boffset;
  }

  /** {@inheritDoc} */
  public double packetToDataOffset (DataFile df, double boffset) {
    int bpp = dataLength;
    if (df.hp==null) bpp += headerLength;
    double npkt = Math.floor(boffset/bpp);
    return npkt*fdataLength;
  }

  /** {@inheritDoc} */
  public void open (DataFile df) {
    isInput = df.isInput;
    setFormat(df.getFormat());
    if (df.hp!=null) df.hp.open();
    if (isInput) initFrom(df);
  }

  /** {@inheritDoc} */
  public void seek (DataFile df, double boffset) {
    boffset = dataToPacketOffset (df,boffset);
    if (df.hp!=null) df.hp.seek(npktoff);
    if (df.io!=null) df.io.seek((long)(df.getDataStart()+boffset));
  }

  /** {@inheritDoc} */
  public int read (DataFile df, byte[] buf, int boff, int bytes, long lbuf) {
    int nbp=getBytes(), nb, fboff=boff;
    int ndo=(bits!=fbits)? bytes*bits/fbits : bytes;
    int nelem=bytes*8/fbits;
    for (rhdrs=0; ndo>0; ndo-=nb) {
      if (ipoff==0) { 						 		// past packet header
        if (df.hp!=null) df.hp.read (this.buf,headerOffset,headerLength);	// read packet header
        else             df.io.read (this.buf,headerOffset,headerLength-trailerLength);	// read packet header less trailer
        rhdrs++;
      }
      nb = Math.min(ndo,nbp-ipoff);
      if (buf!=null) nb = df.io.read(buf,boff,nb);
      else           nb = df.io.read(lbuf,boff,nb);
      if (nb<0) return nb;
      boff+=nb; ipoff+=nb; if (ipoff==nbp) ipoff=0; // signal next packet header
      if (ipoff==0 && df.hp==null) df.io.read (this.buf,trailerOffset,trailerLength);	// read packet trailer
    }
    if (bits!=fbits) unpack (buf,fboff, nelem, bits,fbits);
    return bytes;
  }

  /** {@inheritDoc} */
  public int write (DataFile df, byte[] buf, int boff, int bytes, long lbuf) {
    int nbp=getBytes(), nb;
    int ndo=(bits!=fbits)? bytes*bits/fbits : bytes;
    int nelem=bytes*8/fbits;
    if (bits!=fbits) pack (buf,boff, nelem, fbits,bits);
    for (whdrs=0; ndo>0; ndo-=nb) {
      if (ipoff==0) { 								// past packet header
        if (df.hp!=null) df.hp.write (this.buf,headerOffset,headerLength);	// write packet header
        else             df.io.write (this.buf,headerOffset,headerLength-trailerLength);	// write packet header
        whdrs++;
      }
      nb = Math.min(bytes,nbp-ipoff);
      if (buf!=null) nb = df.io.write(buf,boff,nb);
      else           nb = df.io.write(lbuf,boff,nb);
      if (nb<0) return nb;
      boff+=nb; ipoff+=nb; if (ipoff==nbp) ipoff=0; 			// signal next packet header
      if (ipoff==0 && df.hp==null) df.io.write (this.buf,trailerOffset,trailerLength);	// write packet trailer
    }
    return bytes;
  }

  /** {@inheritDoc} */
  public void close (DataFile df) {
    if (df.hp!=null) df.hp.close();
  }

  /** {@inheritDoc} */
  public boolean hasHeader() {
    return (rhdrs>0);
  }

  /** {@inheritDoc} */
  public String listHeader() {
    Time t = getTC();
    String ts = (t!=null)? t.toString(12) : null;
    String list = "VRT-Packet ID="+Convert.l2x(getOUID())+" Stream="+getSID()+
	" Count="+getCount()+" Size="+getSize()+" Bits="+getBits()+" TC="+ts;
    return list;
  }

  /** {@inheritDoc} */
  public PacketHandler cloneOf() {
    PacketHandler ph=null;
    try { ph = (PacketHandler)this.clone(); }
    catch (Exception e) {}
    return ph;
  }

  /** copy bytes from a user buffer into the 64by header buffer */
  public void setBuffer (byte[] buffer, int boff) {
    System.arraycopy(buffer,boff, buf,headerOffset, headerLength);
  }

  /** copy bytes from the 64by header buffer into a user buffer */
  public void getBuffer (byte[] buffer, int boff) {
    System.arraycopy(buf,headerOffset, buffer,boff, headerLength);
  }

}
