/****************************************************************************/
/*                                                                          */
/*   A2DM14 module interface routines                                       */
/*                                                                          */
/****************************************************************************/
/* A2DM14 Defines */
#define A2DM14_BASE       0x0000  /* Base configuration reg bit[4:0]=(complex,testclk,testramp,tcxo_pd,enable) */
#define A2DM14_FOFF       0x0004  /* Base configuration complex output freq offset */
#define A2DM14_STATE      0x0010  /* 4-byte Configuration state register (See A2DM14_STATE_* for mapping) */
#define A2DM14_SLC        0x0100  /* Silicon Labs chip register bit[6:0]=(INC,DEC,CS,LOL,C2B,C1B,NRS) pins */
#define A2DM14_SLC_SPI    0xC100  /* Silicon Labs chip (SPI mode=2) */
#define A2DM14_ADC        0x0200  /* A to D chip register bit[0:0]=(MODE) pins */
#define A2DM14_ADC_SPI    0x8200  /* A to D chip (SPI mode=1) */
#define A2DM14_PGA        0x0300  /* Prog. Gain Amp register bit[7:0]=(NPD,GS,B[5:0]) pins */
#define A2DM14_FTR        0x1000  /* Fine Time Register */

#define A2DM14_STATE_TRUE      0x01
#define A2DM14_STATE_FALSE     0x00
#define A2DM14_STATE_ON        0x01
#define A2DM14_STATE_OFF       0x00

typedef struct {
    int_u4 chksum;    /* 32 bit checksum over packet ensure packet integrity */
    int_4  refclk;    /* Reference clock rate in Hz                          */
    real_8 samprate;  /* A/D sample rate in Hz                               */
    int_4  foffset;   /* complex rotation in output mode                     */
    int_1  enable;    /* Module enable/disable flag {TRUE,FALSE}             */
    int_1  bits;      /* Module output bits {-16,-8,8,16}                    */
    int_1  pga;       /* Gain of programmable gain amplifier {-100,20}       */
    int_1  tcxo;      /* TCXO power state {ON,OFF}                           */
    int_1  test;      /* A2D test waveform                                   */
    int_1  opts;      /* Options mask                                        */
    int_1  dual;      /* DDR output mode                                     */
    int_1  sysreg;    /* System Register                                     */
} A2DM14_State_Obj;

#define A2DM14_State_Obj_Bytes sizeof(A2DM14_State_Obj)


int_4 a2dm14_get_state (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST)
{ 
  int_4 status = iom_get_state (p,mport,A2DM14_STATE,(void*)ADST,sizeof(*ADST));

  /* If debug output is on, display unfiltered state info */
  if (p->verbose == 3) {
    printf("ADST.chksum:    %8.8X  \n",ADST->chksum);
    printf("ADST.refclk  :  %i     \n",ADST->refclk);
    printf("ADST.samprate:  %17.6f \n",ADST->samprate);
    printf("ADST.foffset:   %8.8x  \n",ADST->foffset);
    printf("ADST.enable:    %2.2X  \n",ADST->enable);
    printf("ADST.bits:      %i     \n",ADST->bits);   
    printf("ADST.pga:       %i     \n",ADST->pga);   
    printf("ADST.tcxo:      %2.2X  \n",ADST->tcxo);  
    printf("ADST.test:      %2.2X  \n",ADST->test);  
    printf("ADST.opts:      %2.2X  \n",ADST->opts);  
    printf(" \n");
  }

  return status;

}

int_4 a2dm14_set_state (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST) 
{ 
  return iom_set_state (p,mport,A2DM14_STATE,(void*)ADST,sizeof(*ADST));
}

int_4 a2dm14_init_state (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST)
{
  /* Zero the state buffer */
  vfill (0, (int_4*)ADST, sizeof(*ADST));

  /* Default the fields    */
  ADST->chksum    = 0xFFFFFFFF;
  ADST->refclk    = 10000000;
  ADST->samprate  = 100000000.0;
  ADST->foffset   = 0x80000000;
  ADST->enable    = A2DM14_STATE_FALSE;
  ADST->bits      = 16;
  ADST->pga       = -100;
  ADST->tcxo      = A2DM14_STATE_ON;
  ADST->opts      = 0;

  /* push state info onto module */
  return a2dm14_set_state(p,mport,ADST);

}   

 
int_4 a2dm14_cfg_clocks (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST)
{
  Si5326_Reg_Obj Si5326_Reg;
  int_u1 regadd, regval;
  int_4  i,ii;

  /* Si5326 Ref-Clock and A/D clock configuration    */
  Si5326_Reg.debuglevel = -1;
  Si5326_Reg.fref   = (double)ADST->refclk;
  Si5326_Reg.dclk1  = ADST->samprate;   
  Si5326_Reg.dclk2  = ADST->samprate;  
  Si5326_Reg.dwght1 = 1.0;
  Si5326_Reg.dwght2 = 1.0;   
  ii = si5326_regcfg_init(&Si5326_Reg,SI5326_REGCFG_MT_A2DM14);
  ii = si5326_regcfg_bwsel(&Si5326_Reg,7);

  if (ADST->opts & RFOPT_XADCLK) {
    /* Configure Si5326 for bypass */
    si5326_regcfg_clkin(&Si5326_Reg,SI5326_REGCFG_CLKIN_1);
    si5326_regcfg_clkin(&Si5326_Reg,SI5326_REGCFG_CLKIN_BYP);
  }
  else {
    /* Configure Si5326 to generate desired A/D clock rate */
    if (ADST->opts & RFOPT_XREF) {
      si5326_regcfg_clkin(&Si5326_Reg,SI5326_REGCFG_CLKIN_1);
    }
    else {
      si5326_regcfg_clkin(&Si5326_Reg,SI5326_REGCFG_CLKIN_2);
    }
    ii = si5326_regcfg_dividers(&Si5326_Reg);
    if (ii != 0) {
      ADST->samprate = Si5326_Reg.fclk2;
      printf("A2DM14 Warning: A/D sample rate not possible. Using %f\n", Si5326_Reg.fclk2);
    }
  }
  si5326_regcfg_debug(&Si5326_Reg);

  /* Reset the Si5326 */
  pic_jpmemwr (p,mport,A2DM14_SLC,0x00,1);
  udelay(10000);
  pic_jpmemwr (p,mport,A2DM14_SLC,0x01,1);

  if (ADST->opts & RFOPT_XADCLK) {
    pic_jpmemwr(p,mport,(A2DM14_SLC_SPI|0),0x22,1); 
  }
  else

  /* Now setup the Si5326 chip registers */
  for(i=0; i<Si5326_Reg.numreg; i++) {
    udelay(100);
    regadd = ((Si5326_Reg.regval[i] & 0x0000FF00)>>8);
    regval =  (Si5326_Reg.regval[i] & 0x000000FF);
    pic_jpmemwr(p,mport,(A2DM14_SLC_SPI|regadd),regval,1); 
    /* ICAL could take up to 53msec worst case */
    if (Si5326_Reg.regval[i] == 0x00008840) udelay(53000);  
  }

  udelay(10000); /* Reset DCM */
  pic_jpmemwr (p,mport,A2DM14_SLC,0x01,1);

  return (0);
}

int_4 a2dm14_cfg_pga (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST)
{ 
  /* Gain Block chip register bit[7:0]=(NPD,N/A,B[5:0]) pins - GS strobe now pulsed on reg write */
  int_u1 regval = 0x00;
  real_8 didx; 
  int_4 gain = ADST->pga;

  if (gain <= -100) {
    /* Turn off input amplifier output */
    pic_jpmemwr (p,mport,A2DM14_PGA,0x00,1);
    ADST->pga = -100;
    vprint("A2DM14: Turn PGA870 OFF \n");
  }
  else if ((gain >= -11) && (gain <= 20)) {
    /* Convert gain to integer index for PGA870 gain */
    didx = (real_8) gain;
    didx = 2.0*(didx + 11.5);
    didx = (didx >= 0.0) ? (didx) : (0.0);
    didx = (didx <= 63.0) ? (didx) : (63.0);
    regval = (int_u1) didx;
    regval = regval | 0x80;
    pic_jpmemwr (p,mport,A2DM14_PGA,regval,1);
    vprint("A2DM14: PGA870 gain set %i  %f  0x%2.2X \n",gain,didx,regval);
    ADST->pga = gain;
  }
  else {
    /* Invalid gain requested - return error status */
    vprint("A2DM14: Invalid gain request %i \n",gain);
    return (-1);
  }
  return (0);
}
  
int_4 a2dm14_cfg_ads61b49 (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST)
{
  int_u1 regadd, regval;
  double rate;

  rate = ADST->samprate;
  if (ADST->bits<0) rate *= 2;

  /* Reset the device */
  regadd = 0x00; regval=0x80;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Select high/low speed mode */
  regadd = 0x20; regval = (rate >= 75000000.0) ? 0x00 : 0x04;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Remove standby */
  regadd = 0x3F; regval=0x00;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Select LVDS data output */
  regadd = 0x41; regval=0x80;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Select clock edge position */
  regadd = 0x44; regval=0xD8;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Select output data format (2s compliment) */
  regadd = 0x50; regval=0x04;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Select output data content */
  regadd = 0x51; regval=0x00;            /* Upper 8-bits custom pat */
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);
  regadd = 0x52; regval=0x00;            /* Lower 6-bits custom pat */
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);
  /* 0x00=Normal, 0x01=AllZeros 0x02=AllOnes, 0x04=Ramp, 0x05=CustomPat */
  regadd = 0x62; regval=ADST->test;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* ADC Fine Gain (0-6dB) and DC offset correction time constant */
  regadd = 0x55; regval=0x08;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* DC offset pedestal */                      
  regadd = 0x63; regval=0x00;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  /* Enable / disable DC offset correction */
  regadd = 0x53; regval=(ADST->opts&RFOPT_DCS)?0x40:0x00;
  pic_jpmemwr(p,mport,(A2DM14_ADC_SPI|regadd),regval,1); 
  udelay(1000);

  return (0);
}

int_4 a2dm14_implement (PICSTRUCT *p, int_4 mport, A2DM14_State_Obj *ADST, int_4 flags)
{ 
  A2DM14_State_Obj ADSTCur;     
  int_4 sysreg; 

  /* consistency and range checks */
  ADST->samprate = max(1000000.0,min(300000000.0,ADST->samprate));
  if (ADST->opts&RFOPT_XADCLK) ADST->opts &= ~RFOPT_XREF;
  ADST->tcxo = (ADST->opts&(RFOPT_XREF|RFOPT_XADCLK))? 0 : 1;

  /* system register */
  sysreg = 0;
  if (ADST->bits<0)  sysreg |= 0x10; /* complex output */
  if (findintflag("DCMTEST",p->config)>0) sysreg |= 0x08; /* DCM version of half speed clock */
  if (ADST->test!=0) sysreg |= 0x04; /* internal test ramp */
  if (ADST->tcxo==0) sysreg |= 0x02; /* power down TCXO */
  if (findintflag("MTGO",p->config)>0) sysreg |= 0x20;
  if (ADST->dual!=0) sysreg |= 0x40; /* DDR output mode */
  if (findintflag("LSBT",p->config)>0) sysreg |= 0x80; /* LSB Test of BUFIO2 cell */
  pic_jpmemwr (p,mport,A2DM14_BASE,sysreg,1);
  pic_jpmemwr (p,mport,A2DM14_FOFF,ADST->foffset,4);

  /* A/D clock configuration    */   
  a2dm14_cfg_clocks(p,mport,ADST);

  /* programmable gain configuration     */   
  a2dm14_cfg_pga(p,mport,ADST);

  /* ADS61B49 configuration     */   
  a2dm14_cfg_ads61b49(p,mport,ADST);

  if (ADST->enable)  sysreg |= 0x01; /* enable transport */
  pic_jpmemwr (p,mport,A2DM14_BASE,sysreg,1);
  ADST->sysreg = sysreg;

  /* update state */
  a2dm14_set_state(p,mport,ADST);

  return (0);
}
 
/*---------------------------------------------------------------------------*/
/* A2DM14 - Public Interfaces for Module Initialization, Setup, and Adjust   */

int_4 a2dm14_setup (PICSTRUCT *p, int_4 mport, int_4 dir, int_4 bits, real_8 rate, int_4 gain, int_4 flags)
{ 
  A2DM14_State_Obj ADST; int_4 i; double d;

  /* Initialize state information on module */
  a2dm14_init_state(p,mport,&ADST);

  ADST.bits = bits; 
  ADST.pga  = gain; 
  ADST.test = findintflagdef("A2DTEST",p->config,0);
  ADST.samprate = (bits<0)? rate*2 : rate;
  ADST.opts = findmaskflagdef("A2DOPTS",p->config,RFOPTLIST,0);
  ADST.dual = (flags&FLG_DUAL) || (bits<0);

  /* adjust center freq of hilbert output ? */
  d = finddblflag("MFREQ",p->config);
  if (d != -1.0) { 
    d /= (-ADST.samprate/4); 
    if (d> 1.0) d-=2.0;
    if (d<-1.0) d+=2.0;
    ADST.foffset = R2L(d);
   }

  /* setup base register modes - disable module output */
  pic_jpmemwr (p,mport,A2DM14_BASE,0,1);

  /* Determine if external or internal reference is to be used */
  i = findintflag("PREFX",p->config);
  if (i > 0) ADST.opts |= RFOPT_XREF;
  if (i > 100) ADST.refclk = i; /* in Hz */
  else if (i > 1) ADST.refclk = i*1000000; /* in MHz */

  /* Determine if external clock is to be applied */
  if (findflag("MUXCLK=N",p->config) >= 0) ADST.opts |= RFOPT_XADCLK;
  if (findflag("MUXCLK=A",p->config) >= 0) {
    if (mport==1) ADST.opts |= RFOPT_XADCLK;
    else printf ("Cannot use MUXCLK=A with A2DM14 on Module=2\n");
  }
  if (findflag("MUXCLK=B",p->config) >= 0) {
    if (mport==2) ADST.opts |= RFOPT_XADCLK;
    else printf ("Cannot use MUXCLK=B with A2DM14 on Module=1\n");
  }

  /* implement with enable */
  ADST.enable = 0;
  a2dm14_implement(p,mport,&ADST,-1);

  return (0);
}

int_4 a2dm14_enable (PICSTRUCT *p, int_4 mport, int_4 ena) {
  A2DM14_State_Obj ADST; 
  a2dm14_get_state(p,mport,&ADST);
  ADST.enable = ena;
  if (ena) ADST.sysreg |= 0x01;
  else     ADST.sysreg &= 0xFE;
  pic_jpmemwr (p,mport,A2DM14_BASE,ADST.sysreg,1);
  a2dm14_set_state(p,mport,&ADST);
  return 0;
}

int_4 a2dm14_set_gain (PICSTRUCT *p, int_4 mport, int_4 gain)
{ 
  int_4 stat;
  A2DM14_State_Obj ADST;     
  /* Gain Block chip register bit[7:0]=(NPD,GS,B[5:0]) pins */
  a2dm14_get_state(p,mport,&ADST);
  ADST.pga = gain;
  stat = a2dm14_cfg_pga(p,mport,&ADST);
  a2dm14_set_state(p,mport,&ADST);
  return stat;
}
  
int_4 a2dm14_get_gain (PICSTRUCT *p, int_4 mport)
{ 
  A2DM14_State_Obj ADST;     
  a2dm14_get_state(p,mport,&ADST);
  return (int_4)ADST.pga;
}

int_4 a2dm14_set_opts (PICSTRUCT *p, int_4 mport, int_4 flags)
{
  A2DM14_State_Obj ADST;     
  a2dm14_get_state(p,mport,&ADST);
  /* Make certain options arent already enabled */
  if (!(flags & RFOPT_ENABLE)) return (0);
  a2dm14_implement(p,mport,&ADST,0);
  return 0;
}

int_4 a2dm14_get_opts (PICSTRUCT *p, int_4 mport)
{
  A2DM14_State_Obj ADST;     
  a2dm14_get_state(p,mport,&ADST);
  return (int_4)ADST.opts;
}

