/****************************************************************************/
/*                                                                          */
/*   iomlib_a2dm1x.h: includes for A2DM1X library interfaces                */
/*                                                                          */
/****************************************************************************/
#ifndef _IOMLIB_A2DM1X_H
#define _IOMLIB_A2DM1X_H

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>   
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "icelib.h" 

/* A2DM1X module control register */
#define A2DM1X_SYSREG       0x0000 /* 8-bit system register interface        */
#define A2DM1X_SYSREG_EXTREF  0x80 /*   bit[7] = external reference          */
#define A2DM1X_SYSREG_INTREF  0x40 /*   bit[6] = internal reference          */
#define A2DM1X_SYSREG_MTGO    0x20 /*   bit[5] = module triggered go         */
#define A2DM1X_SYSREG_CMPLX   0x10 /*   bit[4] = complex output data (I/Q)   */
#define A2DM1X_SYSREG_TESTCLK 0x08 /*   bit[3] = test clock                  */
#define A2DM1X_SYSREG_TESTRMP 0x04 /*   bit[2] = test ramp                   */
#define A2DM1X_SYSREG_PDTCXO  0x02 /*   bit[1] = TXCO power down             */
#define A2DM1X_SYSREG_ENABLE  0x01 /*   bit[0] = enable / disable            */

/* Base configuration complex tuner output freq offset */
#define A2DM1X_FOFF         0x0008  

/* Address to begin storing A2DM1X Object (A2DM1X_State_Obj) on module */
#define A2DM1X_STATE        0x0010  

/* Temp Sensor base address - Placeholder for now...*/
#define A2DM1X_TEMP         0x0100    

/* ADF4350 clock chip base addresses for pins and registers */
#define A2DM1X_CLK_PINS     0x0200  /* 8-bit register interface              */
                                    /*   bit[0]=(s_ce) - chip enable         */
                                    /*   bit[7:4]=(X,X,s_mux,s_ld} read only */
#define A2DM1X_CLK_RG_BASE  0x0201  /* Write 4 bytes to load 32-bit register */ 

/* ADC12D A/D converter chip base addresses for pins and registers */
#define A2DM1X_ADC_PINS     0x0320 /* 8-bit register interface               */
                                   /*   bit[1:0]=(adc_cal,adc_ndm)           */
#define A2DM1X_ADC_RG_BASE  0x8300 /* Base address for SPI read/write to ADC */
#define A2DM1X_ADC_RG_WROFF 0x0040 /* Write offset for SPI registers         */
#define A2DM1X_ADC_RG_RDOFF 0x00C0 /* Read offset for SPI registers          */

/* PE4302 attenuator control register base addresses */
#define A2DM1X_ATN1_RG_BASE 0x0400 
#define A2DM1X_ATN2_RG_BASE 0x0500
#define A2DM1X_ATN3_RG_BASE 0x0600 
#define A2DM1X_ATN4_RG_BASE 0x0700

/* Dual-site module - All comms go to module port 2 on A2DM1X */
#define A2DM1X_PORT_DSM     2

/* A2DM1X Object */
#define A2DM1X_ID        0x46FC /* 16-bit unique module identifier */
#define A2DM1X_RG_MAXCNT 1

typedef struct {
    int_u4   chksum;     /* 32 bit checksum over packet ensures integrity  */
    int_u2   id;         /* Unique 16-bit module ID                        */
    int_u1   mtrev;      /* Module type / revision                         */
    int_1    debuglevel; /* Module debug level {0,127}                     */
    int_u1   sysreg;     /* System Register                                */
    int_1    opts;       /* Options Register                               */
    int_1    bits;       /* Module output bits {-16,-8,8,16}               */
    int_1    ch1pga;     /* Chan1 programmable gain amplifier {-100,28}    */
    int_1    ch2pga;     /* Chan2 programmable gain amplifier {-100,28}    */
    int_1    test;       /* A2D test waveform                              */
    int_1    ovsr;       /* Oversample ratio {1,127}                       */
    int_1    pad[3];     /* Pad structure size to multiple of 4 bytes      */
    int_4    foffset;    /* complex rotation in output mode                */
    int_4    refclk;     /* Reference clock rate in Hz                     */
    real_8   samprate;   /* A/D sample rate in Hz                          */
#ifdef BOGUS
    uint8_t  pinval;                   /* A2DM1X Pin I/O Register     */
    uint16_t addr[PTCFG_ADDR_MAX];     /* Based and offset addresses  */
    uint16_t rgadd[A2DM1X_RG_MAXCNT];  /* A2DM1X Register Address     */
    uint8_t  rgval[A2DM1X_RG_MAXCNT];  /* A2DM1X Register Value       */
#endif
    ADF4350_Reg_Obj SYN;  /* Clock synthesizer part */
#ifdef BOGUS
    ADC12D_Reg_Obj  ADC;  /* A/D converter part     */
#endif
    PE4302_Reg_Obj  ATN1; /* Analog attenuator I/O Module 1, 1st stage   */
    PE4302_Reg_Obj  ATN2; /* Analog attenuator I/O Module 1, 2nd stage   */
    PE4302_Reg_Obj  ATN3; /* Analog attenuator I/O Module 2, 1st stage   */
    PE4302_Reg_Obj  ATN4; /* Analog attenuator I/O Module 2, 2nd stage   */
} A2DM1X_State_Obj;

/*                                                             */
/* !!!IMPORTANT!!! - sizeof(A2DM1X_State_Obj) MUST BE AN EVEN  */
/*                   MULTIPLE OF 4 BYTES FOR iom_set_state     */
/*                   TO WORK PROPERLY                          */
/*                                                             */
#define A2DM1X_State_Obj_Bytes sizeof(A2DM1X_State_Obj)
#define A2DM1X_STATE_MAX_BYTES 256 /* Max possible bytes in A2DM1X_State_Obj */

/* A2DM1X Register ID's */
#define A2DM1X_RG_PIN         4   /* User Control of Device Pins via FPGA */

/* A2DM1X Register ID's */
#define A2DM1X_RG_0h_SYSREG   0
#define A2DM1X_RG_1h_RESV     1
#define A2DM1X_RG_2h_RESV     2
#define A2DM1X_RG_3h_RESV     3

/* A2DM1X Pin Keys      */
#define A2DM1X_RG_PIN_NULL 0x0001  /* Non-Demux Mode {0=Demux, 1=No Demux  */

/* A2DM1X Register Keys */
#define A2DM1X_RG_0h_EXTREF  0x80 /*   bit[7] = external reference          */
#define A2DM1X_RG_0h_INTREF  0x40 /*   bit[6] = internal reference          */
#define A2DM1X_RG_0h_MTGO    0x20 /*   bit[5] = module triggered go         */
#define A2DM1X_RG_0h_CMPLX   0x10 /*   bit[4] = complex output data (I/Q)   */
#define A2DM1X_RG_0h_TESTCLK 0x08 /*   bit[3] = test clock                  */
#define A2DM1X_RG_0h_TESTRMP 0x04 /*   bit[2] = test ramp                   */
#define A2DM1X_RG_0h_PDTCXO  0x02 /*   bit[1] = TXCO power down             */
#define A2DM1X_RG_0h_ENABLE  0x01 /*   bit[0] = enable / disable            */

#endif /* _IOMLIB_A2DM1X_H */
