/****************************************************************************/
/*                                                                          */
/*   PUBLIC interfaces to part / IC configuration libraries                 */
/*                                                                          */
/*   Each part must define the following methods (at a minimum):            */
/*     _pt_get()   -> used to get register and pin status from memory       */
/*     _pt_gethw() -> used to get register and pin status from the hardware */
/*     _pt_set()   -> used to set register and pin status to memory         */
/*     _pt_sethw() -> used to set register and pin contents of the hardware */
/*     _pt_addr()  -> used to set base and offset addresses of HW registers */
/*     _pt_init()  -> initializes content of hardware object                */
/*     _pt_debug() -> dumps contents of the hardware object                 */
/*                                                                          */
/****************************************************************************/

#include "part_public.h"

/*---------------------------------------------------------------------------*/
/* Declare part ID's as a public static array terminated by PTCFG_ID_INVALID */
/*---------------------------------------------------------------------------*/
static uint32_t ptcfg_id_list[] =
      { NULLPART_ID,        ADF4350_ID,        ADC12D_ID,      
        PE4302_ID,          PTCFG_ID_INVALID };

/* Declare work functions for parts as public static array */
static uint32_t (*ptcfg_fcn_list_get[])() =
      { nullpart_pt_get,    adf4350_pt_get,    adc12d_pt_get,
        pe4302_pt_get     };

static uint32_t (*ptcfg_fcn_list_gethw[])() =
      { nullpart_pt_gethw,  adf4350_pt_gethw,  adc12d_pt_gethw,
        pe4302_pt_gethw   };

static int32_t (*ptcfg_fcn_list_set[])() =
      { nullpart_pt_set,    adf4350_pt_set,    adc12d_pt_set,
        pe4302_pt_set     };

static int32_t (*ptcfg_fcn_list_sethw[])() =
      { nullpart_pt_sethw,  adf4350_pt_sethw,  adc12d_pt_sethw,
        pe4302_pt_sethw   };

static int32_t (*ptcfg_fcn_list_addr[])() =
      { nullpart_pt_addr,   adf4350_pt_addr,   adc12d_pt_addr,
        pe4302_pt_addr     };

static int32_t (*ptcfg_fcn_list_init[])() =
      { nullpart_pt_init,   adf4350_pt_init,   adc12d_pt_init,
        pe4302_pt_init     };

static int32_t (*ptcfg_fcn_list_debug[])() =
      { nullpart_pt_debug,  adf4350_pt_debug,  adc12d_pt_debug,
        pe4302_pt_debug    };

/*---------------------------------------------------------------------------*/
/* Part configuration methods                                                */
/*---------------------------------------------------------------------------*/
int32_t ptcfg_get_id_index(uint32_t pt_id)
{
    int32_t idx = 0;
    int32_t rtn = -1;
    while (ptcfg_id_list[idx] != PTCFG_ID_INVALID) {
      if (pt_id == ptcfg_id_list[idx]) break;  
      idx++;
    }
    rtn = (ptcfg_id_list[idx] != PTCFG_ID_INVALID) ? (idx) : (0);
    return (rtn);
}


uint32_t ptcfg_get(uint32_t pt_id, void *pt_obj, uint32_t rg_id,     
                        uint32_t rg_key, int32_t *rtnstatus)
{
    int32_t  rtn=0;
    uint32_t rg_val;
    rg_val = ptcfg_fcn_list_get[ptcfg_get_id_index(pt_id)](pt_obj, 
               rg_id, rg_key);
    if (rtnstatus != (int32_t *)NULL) *rtnstatus = rtn;
    return (rg_val);
}

 
uint32_t ptcfg_gethw(PICSTRUCT *p, int32_t port, uint32_t pt_id, 
             void *pt_obj, uint32_t rg_id, uint32_t rg_key, int32_t *rtnstatus)
{
    int32_t  rtn=0;
    uint32_t rg_val;
    rg_val = ptcfg_fcn_list_gethw[ptcfg_get_id_index(pt_id)](p, port, 
               pt_obj, rg_id, rg_key);
    if (rtnstatus != (int32_t *)NULL) *rtnstatus = rtn;
    return (rg_val);
} 


int32_t ptcfg_set(uint32_t pt_id, void *pt_obj, uint32_t rg_id,     
                        uint32_t rg_key, uint32_t rg_val)
{
    int32_t rtn;
    rtn = ptcfg_fcn_list_set[ptcfg_get_id_index(pt_id)](pt_obj, 
            rg_id, rg_key, rg_val);
    return (rtn);
} 


int32_t ptcfg_sethw(PICSTRUCT *p, int32_t port, uint32_t pt_id, 
               void *pt_obj, uint32_t rg_id, uint32_t rg_key, uint32_t rg_val)
{
    int32_t rtn;
    rtn = ptcfg_fcn_list_sethw[ptcfg_get_id_index(pt_id)](p, port, 
            pt_obj, rg_id, rg_key, rg_val);
    return (rtn);
} 


int32_t ptcfg_addr(uint32_t pt_id, void *pt_obj, uint32_t addr_key,
                     uint32_t addr)
{
    int32_t rtn;
    rtn = ptcfg_fcn_list_addr[ptcfg_get_id_index(pt_id)](pt_obj, 
            addr_key, addr);
    return (rtn);
} 


int32_t ptcfg_init(uint32_t pt_id, void *pt_obj, int32_t module_type)
{
    int32_t rtn;
    rtn = ptcfg_fcn_list_init[ptcfg_get_id_index(pt_id)](pt_obj, module_type);
    return (rtn);
} 


int32_t ptcfg_debug(uint32_t pt_id, void *pt_obj)
{
    int32_t rtn;
    rtn = ptcfg_fcn_list_debug[ptcfg_get_id_index(pt_id)](pt_obj);
    return (rtn);
} 

