package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.dsp.prim.firkais;

/**
  Implements a Bank Of Tuners on a complex baseband input.

  @author Jeff Schoen
  @version $Id: icebot.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icebot extends Primitive implements Chainable {

  private int nout,nfft,gain,flags,nppt=8,nrft=4096,nrpt=64;
  private double irate,orate,sfreq,dfreq;
  private boolean real;
  private DataFile hi,ho,hf;
  private Data dbi,dbo,dbpp,dbrf,dbpf;
  private Core bot;

  public int open() {

    // open CORE library
    bot = Core.forName("BoT",MA);
    String flagsList = bot.getS("FLAGSLIST");
    flags = MA.getOptionMask("/FLAGS",flagsList,0);
    real = bot.isSet(flagsList,"REAL",flags);

    // open input file
    hi = MA.getDataFile ("IN","1000","C#",0);
    hi.open();

    // get parameters
    nout  = MA.getL("NCHN");
    sfreq = MA.getD("FREQ");
    dfreq = MA.getD("DFREQ");
    orate = MA.getD("ORATE");
    gain  = MA.getL("GAIN");

    irate = 1/hi.getXDelta();
    if (dfreq<0) dfreq = irate/nout;
    if (orate<0) orate = dfreq;
    nfft = (int)Math.round(irate/dfreq/2);
    nppt = MA.getL("/NPPT",nppt);
    nrft = MA.getL("/NRFT",nrft);
    nrpt = MA.getL("/NRPT",nrpt);

    // get filters
    dbpp = getFilter("FPPF",nppt*nfft);
    dbrf = getFilter("RFIR",nrft);

    // set CORE parameters
    bot.set("NFFT",nfft);
    bot.set("NOUT",nout);
    bot.set("NPPT",nppt);
    bot.set("NRPT",nrpt);
    bot.set("FLAGS",flags);
    bot.set("FOFF",sfreq/irate);
    bot.set("RATIO",orate/dfreq);
    bot.set("GAIN",gain);
    bot.set("FPPF",dbpp);
    bot.set("RFIR",dbrf);
    bot.open();

    // get implementation native format
    byte dtype = bot.getDataType();
    String oform = (real? "S":"C") + ((dtype==INT)? "I":"F");
    if (real) orate *= 2;

    // open output file
    ho = MA.getDataFile("OUT","2000",oform,0);
    ho.propagate(hi,"TIMECODE|KEYWORDS"); 
    ho.setSubSize (nout);
    ho.setXStart(0.0);
    ho.setXDelta(dfreq);
    ho.setXUnits(Units.FREQUENCY);
    ho.setYStart(hi.getXStart()+nppt/orate);
    ho.setYDelta(1.0/orate);
    ho.setYUnits(Units.TIME);
    ho.setSize(Math.floor(hi.getSize()*orate/irate));
    ho.open(ho.OUTPUT);
    ho.setDFS(0);

    // alloc processing buffers
    dbi = hi.getDataBuffer(nfft,dtype); 
    dbo = ho.getDataBuffer(nout);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = bot.process (dbi,n, dbo,nout);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    bot.close();
    ho.setSize(Math.floor(ho.seek()/nout));
    hi.close();
    ho.close();
    return (NORMAL);
  }

  private Data getFilter (String arg, int ntap) {

    // read filter from data file
    hf = MA.getDataFile ("/"+arg,"1000","SF",0);
    if (hf.getURL().length()>0) {
      hf.open(hf.INPUT); 
      int size = (int)hf.getSize();
      if (size!=ntap) M.error("Filter file="+hf.getURL()+" size="+size+" not equal to ntaps="+ntap); 
      Data db = hf.getDataBuffer(ntap,FLOAT);
      hf.read(db);
      hf.close();
      return db;
    }

    // generate filter internally using FIRKAIS method
    int nptap = ntap;
    boolean dbg = false;
    float[] fbo = new float[ntap*2];
    float[] taps = new float[ntap];
    float scale = 1;
         if (arg.equals("FPPF")) { nptap/=nppt; firkais.generate(1, 0.80/nptap,0.0,0.2/nptap, 70.0, fbo,ntap,dbg); }
    else if (arg.equals("RFIR")) { nptap/=nrpt; firkais.generate(1, 0.375/nptap,0.0,0.05/nptap, 80.0, fbo,ntap,dbg); scale=nptap; }
    for (int i=0; i<ntap; i++) taps[i]=fbo[i+i]*scale; // complex to real and scale
    if (arg.equals("FPPF")) for (int i=0; i<ntap; i++) taps[i]*=nptap;
    Data db = new Data(taps);

    // output filter for debug
    hf = MA.getDataFile ("/O"+arg,"1000","SF",0);
    if (hf.getURL().length()>0) {
      hf.open(hf.OUTPUT);
      hf.setXDelta(1.0/ntap);
      hf.setXStart(0.5*(1.0/ntap-1));
      hf.write(db);
      hf.close();
    }
    return db;
  }

  public Object getNextLink() { return bot; }
  public Object getPrevLink() { return null; }

}
