package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Implements a Tune/Filter/Decimate/Demod function similar to a Graychip DDC

  This primitive has multiple implementations: Java, Native-C, and FPGA

  @author Jeff Schoen
  @version $Id: icetfdd.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icetfdd extends Primitive implements Chainable {

  private int dec,gain,rmon,flags,ovsr,tl=4096;
  private double freq,ratio,rdec,tdec,fnyq;
  private DataFile hi,ho,hf;
  private Data dbi,dbo,dbf;
  private Core tfd;
  private float[] cfir,pfir,rfir;
  private boolean htf,resamp,pm;
  private String ofmt;

  public int open() {

    hi = MA.getDataFile ("IN","1000","S#,C#",0);
    hi.open();

    ovsr   = MA.getL("/OVSR",1);
    ofmt   = MA.getS("/OFMT","SI");
    ratio  = MA.getD("/RATIO",1.0);
    rmon   = MA.getL("/RMON",ratioMoN(ratio));

    dec    = MA.getL("DEC") * ovsr;
    freq   = MA.getD("FREQ") / ovsr;
    gain   = MA.getL("GAIN");

    fnyq   = 0.5/hi.getDelta();
    htf    = (dec==1);
    resamp = (ratio!=1.0);
    rdec   = dec/ratio;
    tdec   = (ofmt.charAt(0)=='C')? rdec*2/ovsr : rdec/ovsr;

    ho = MA.getDataFile ("OUT",hi,0);
    ho.setFormat(ofmt);
    ho.setDelta(hi.getDelta()*tdec);
    ho.open();

    tfd = Core.forName("TFDD",MA);

    flags = 0; String flagsList = tfd.getS("FLAGSLIST");
    if (hi.spa==2) flags = Parser.mask(flagsList,"+CXI",flags);
    if (ho.spa==2) flags = Parser.mask(flagsList,"+CXO",flags);
    if (htf)       flags = Parser.mask(flagsList,"+HTF",flags);
    if (resamp)    flags = Parser.mask(flagsList,"+RESAMP",flags);
    flags = MA.getOptionMask("/FLAGS",flagsList,flags);

    String cfn = MA.getCS("/CFN",htf?"cfir_80f":resamp?"dfir_25":"cfir_80");
    cfir = getFilterTaps("nxm.ice.dat."+cfn);

    String pfn = MA.getCS("/PFN",htf?"pfir_80f":"pfir_80");
    pfir = getFilterTaps("nxm.ice.dat."+pfn);

    tfd.set("FLAGS",flags);
    tfd.set("DECIMATION",dec);
    tfd.set("FREQUENCY",freq/fnyq);
    tfd.set("RATIO",ratio);
    tfd.set("RMON",rmon);
    tfd.set("OVSR",ovsr);
    tfd.set("GAIN",gain);
    tfd.set("CFIR",cfir);
    tfd.set("PFIR",pfir);

    if (resamp) {
      String rfn = MA.getCS("/RFN",pm?"res_8x256_90":"res_10x2048_90");
      rfir = getFilterTaps("nxm.ice.dat."+rfn);
      tfd.set("RFIR",rfir);
    }

    tfd.open();

    tl = MA.getL("/TL", tl);
    byte dtype = INT;
    dbi = hi.getDataBuffer(tl,dtype); 
    dbo = ho.getDataBuffer(tl,dtype);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = tfd.process (dbi,n, dbo,tl);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    tfd.close();
    hi.close();
    ho.close();
    return (NORMAL);
  }

  public Object getNextLink() { return tfd; }
  public Object getPrevLink() { return null; }

  private float[] getFilterTaps (String fname) {
    DataFile hf = new DataFile(this,fname);
    hf.open();
    int taps = (int)hf.size;
    double scale = 1.0;
    if (hf.getFormat().endsWith("I")) scale = 1.0/(64*1024);
    if (hf.getFormat().endsWith("L")) scale = 1.0/(32768*4);
    Data dbf = hf.getDataBuffer(taps,DataTypes.FLOAT);
    hf.read(dbf);
    hf.close();
    float[] coef = new float[taps];
    dbf.toArray(coef);
    float sum=0;
    for (int i=0; i<taps; i++) { coef[i]*=scale; sum += coef[i]; }
    float gain = sum;
    M.info("Read filter file="+fname+" taps="+taps+" gain="+gain);
    return coef;
  }

  private int ratioMoN (double ratio) {
    double tolr = 0.0000001;
    for (int m=1; m<0xFFFF; m++) {
      double d = m/ratio;
      int n = (int)(d+.5);
      if (Math.abs(d-n)<tolr) return (m<<16)|n;
    }
    return 0;
  }

}

